#!/usr/bin/perl
use strict;
use Time::HiRes qw( clock_gettime CLOCK_REALTIME );
use Image::Imlib2;
use POSIX qw(ceil);

# USAGE HINTS:
#    ./createUnifiedMakefile.pl > makefile
#    make clean
#    nice -n 10 longjob make -j4 all
#

hugeheading("PREDECLARING ALL");
print(".PHONY: preall all clean\n");
print("\n");
print("preall: all\n");
print("\n");

my $pwdParentDir  = join('/', removeLast(split/\//, `pwd`));
my $sourcedata    = "$pwdParentDir/sourcedata";

my @posTrainCfgs  = (grep { !shouldSkipTrain(%{$_}) }
                     grep { my %cfg = parseLstPath($sourcedata,$_); $_=\%cfg }
                     sort 
                     grep { s/\.\./$pwdParentDir/ } 
                     grep { !/old_/ }
                     grep { /w64x128c8x8/ } # temporary filter
                     split(/[\n\r]+/, 
                           `ls ../sourcedata/*/train/*/pos_*_*_*.lst`));

my @posTestCfgs   = (grep { my %cfg = parseLstPath($sourcedata,$_); $_=\%cfg }
                     sort 
                     grep { s/\.\./$pwdParentDir/ } 
                     grep { !/old_/ }
                     grep { /inria/ || /20080619-0710-westwall/ } # temporary filter
                     grep { /w64x128c8x8/ } # temporary filter
                     split(/[\n\r]+/, 
                           `ls ../sourcedata/*/test/*/pos_*_*_*.lst`));

my @negTestCfgs   = (map { my %cfg = parseLstPath($sourcedata, 
                                                  findBestNegLst(%{$_}), %{$_});
                           \%cfg }
                     @posTestCfgs);


my @fullTestLsts   = (sort 
                      grep { s/\.\./$pwdParentDir/ } 
                      grep { !/old_/ } 
                      grep { /eastwall/ || /westwall/ } # temporary filter
                      split(/[\n\r]+/, 
                            `find ../sourcedata/*/test/ -name full`));

my @perspTrainLsts = (sort 
                      grep { s/\.\./$pwdParentDir/ } 
                      grep { !/old_/ } 
                      grep { -d "$_/headtips" && -f "$_/calib.txt" }
                      split(/[\n\r]+/, 
                            `find ../sourcedata/*/train/ -name full`));

my @perspTestLsts  = (sort 
                      grep { s/\.\./$pwdParentDir/ } 
                      grep { !/old_/ } 
                      grep { -d "$_/headtips" && -f "$_/calib.txt" }
                      split(/[\n\r]+/, 
                            `find ../sourcedata/*/test/ -name full`));

my @allTargets;
my @impl = ("cpu", "gpu");
my (%implTargets, %implGeomTargets, %isImplGeomTargets, %isImplGeomErrTargets,
    %isImplGeomTrkTargets);

foreach my $h (sort(@posTrainCfgs)) {
    my %posTrainCfg = %{$h};
    my %negTrainCfg = parseLstPath($sourcedata,
                                   findBestNegLst(%posTrainCfg), %posTrainCfg);

    # Only test on data meant for the same HOG setup
    my @myPosTestCfgs = (grep { ${$_}{hogId} eq $posTrainCfg{hogId} } 
                         @posTestCfgs);
    my @myNegTestCfgs = (grep { ${$_}{hogId} eq $posTrainCfg{hogId} } 
                         @negTestCfgs);
    
    my @myFullTestCfgs   = (grep { scalar(%{$_}) }
                            map { my %cfg = parseLstPath($sourcedata,
                                                         $_,%posTrainCfg); 
                                  \%cfg }
                            map { ''.$_ }
                            @fullTestLsts);

    my @myPerspTrainCfgs = (grep { scalar(%{$_}) }
                            map { my %cfg = parseLstPath($sourcedata,
                                                         $_,%posTrainCfg); 
                                  \%cfg }
                            map { ''.$_ }
                            @perspTrainLsts);

    my @myPerspTestCfgs  = (grep { scalar(%{$_}) }
                            map { my %cfg = parseLstPath($sourcedata,
                                                         $_,%posTrainCfg); 
                                  \%cfg }
                            map { ''.$_ }
                            @perspTestLsts);

    for my $impl (@impl) {
        my @targets = ();
        writeExperimentTree($impl,
                            \%posTrainCfg,\%negTrainCfg,\@myPerspTrainCfgs,
                            \@myPosTestCfgs, \@myNegTestCfgs, 
                            \@myPerspTestCfgs, \@myFullTestCfgs, 
                            \@targets);
        my %t = %posTrainCfg;
        push(@{$implTargets{"$impl"}}, @targets);
        push(@{$implGeomTargets{"$t{hogId}_$t{windowset}_$impl"}}, @targets);
        push(@{$isImplGeomTargets{"$t{imageset}_$t{annoset}_$t{hogId}_$t{windowset}_$impl"}}, @targets);
        push(@{$isImplGeomErrTargets{"$t{imageset}_$t{annoset}_$t{hogId}_$t{windowset}_${impl}_errors"}},
             grep { /train_error/ || /test_error/ } @targets);
        push(@{$isImplGeomTrkTargets{"$t{imageset}_$t{annoset}_$t{hogId}_$t{windowset}_${impl}_trk"}},
             grep { /tracker/ } @targets);
        push(@allTargets, @targets);
    }
}

hugeheading("PHONY TARGETS");
print("ALLTARGETS := ".join(" ", @allTargets)."\n");
print("\n");
print("all: \$(ALLTARGETS)\n");
print("\n");
foreach my $h (\%implTargets, \%implGeomTargets, 
               \%isImplGeomTargets, \%isImplGeomErrTargets,
               \%isImplGeomTrkTargets) {
    my %targets = %{$h};
    foreach my $key (sort keys %targets) {
        print ".PHONY: $key\n";
        print "\n";
        print "$key: ".join(" ", reverse @{$targets{$key}})."\n";
        print "\n";
    }
}
print("clean: \n");
print("\trm -rf *~ \\#*\\#  \$(ALLTARGETS)\n");
print("\n");

###############################################################################
###############################################################################
####  Filtering out (temporarily) undesired datasets
###############################################################################
###############################################################################

# FILTERING
sub shouldSkipTrain(%) {
    my %cfg = @_;
    #return 1 if ($cfg{winw} != 64);
    #return 1 if ($cfg{winh} != 128);
    
    # only full-height windows supported right now
    if ($cfg{winw} == 64) {
        return 1 if ($cfg{winw}*2 != $cfg{winh}); 
    }

    # skip windowsets that exclude partially-occluded people
    return 1 if (!($cfg{windowset} eq 'all')); 

    return 0;
}

###############################################################################
###############################################################################
####  Filename munging
###############################################################################
###############################################################################

# Input: 
#   $sourcedata : base directory for input data (see doc/lstpathname.*)
#   $lst        : "relative path" for a .lst file (see doc/lstpathname.*) OR
#                 the path to a full/ imageset directory
#   %cfg        : default values
#
# Returns a hash containing the following fields:
#   sourcedata,pathname,relativepath,filename
#   dataset,traintest,imageset,posneg,annoset
#   hogId,winw,winh,cellw,cellh,windowset
#   winstridex,winstridey, outerpaddingx,outerpaddingy
#   numcellx,numcelly, orientbin, blockstridex,blockstridey
#
# And sometimes these fields too:
#   scaleprecision, rodrigues, translation, focalLength
#   calibfile              : file from from which the perspective calibration came
#   video                  : video from which frames were taken
#   imagewidth, imageheight: input images should be rescaled to this size
sub parseLstPath($$%) {
    my ($sourcedata,$lst,%cfg) = @_;

    # parsed fields
    (-d $sourcedata) || 
        die "sourcedata path '$sourcedata' must be a directory.";
    (substr($sourcedata,0,1) eq '/') || 
        die "sourcedata path '$sourcedata' must be an absolute path.";
    $cfg{sourcedata} = $sourcedata;

    # pre-scaled/cropped pos/neg lst vs mixed images directory
    if (-f $lst) {
        $cfg{isLst} = 1;
    } elsif (-d $lst) {
        $cfg{isLst} = 0;
    } else {
        die "'$lst' must be a file or directory that exists.";
    }
    (substr($lst,0,1) eq '/') || 
        die "'$lst' must be an absolute path.";

    # pathname parsing
    $cfg{pathname} = $lst;

    $cfg{relativepath} = removePrefix("$sourcedata/",$lst);
    
    my @pathparts = split('/',$cfg{relativepath});
    (@pathparts == 4) || 
        die("Unexpected number of parts (".scalar(@pathparts).") ".
            "to the training .lst file: '$lst'");
    
    $cfg{dataset}     = $pathparts[0];
    $cfg{traintest}   = $pathparts[1];
    $cfg{imageset}    = $pathparts[2];
    $cfg{filename}    = $pathparts[3];
    $cfg{basename}    = $cfg{filename};
    if ($cfg{isLst}) {
        ($cfg{basename} =~ s/\.lst$//) || 
            die("Filename does not end in .lst: '$lst'");
    }

    $cfg{dir}         = removeSuffix("/$cfg{filename}",$cfg{pathname});
    $cfg{reldir}      = removePrefix("$sourcedata/",$cfg{dir});

    $cfg{scaleprecision} = 0.1;
    $cfg{rodrigues}      = '-1,-1,-1';
    $cfg{translation}    = '-1,-1,-1';
    $cfg{focalLength}    = '-1,-1';

    # Customize the configuration for specific datasets
    %cfg = ($cfg{isLst}) ? 
        customizePosNegCfg($lst, %cfg) : 
        customizeFullImagesCfg(%cfg);
        
    return %cfg;
}

sub findBestNegLst(%) {
    my (%posCfg) = @_;

    ($posCfg{posneg} eq 'pos') || 
        die("'$posCfg{pathname}' is not a positive .lst file");

    opendir D,$posCfg{dir};
    my @allFiles = readdir(D);
    closedir(D);

    my @candidates;

    # "exact" match
    @candidates = grep { 
        /^neg_$posCfg{annoset}_$posCfg{hogId}_$posCfg{windowset}\.lst$/ 
        } @allFiles;
    if (@candidates > 1) {
        die("Unexpected multiple negative .lst best matches for ".
            "'$posCfg{pathname}'");
    } elsif (@candidates == 1) {
        return "$posCfg{dir}/$candidates[0]";
    } # else continue...5F

    # different windowset
    @candidates = grep { 
        /^neg_$posCfg{annoset}_$posCfg{hogId}/ && /\.lst$/
        } @allFiles;
    if (@candidates > 1) {
        die("We don't have a strategy for multiple non-exact negative .lst ".
            "files yet: '$posCfg{pathname}'");
    } elsif (@candidates == 1) {
        return "$posCfg{dir}/$candidates[0]";
    } # else continue...

    # neg_full
    @candidates = grep { /^neg_full\.lst$/ } @allFiles;
    if (@candidates > 1) {
        die("Unexpected multiple full negative .lst files for ".
            "'$posCfg{pathname}'");
    } elsif (@candidates == 1) {
        return "$posCfg{dir}/$candidates[0]";
    } # else continue...

    die "Could not find a neg*.lst compatible with '$posCfg{pathname}'";
}

# Given a base training directory and config info for a calib.txt
# file, return the filename of the output .lst file.
sub permissiveLst($%) {
    my ($d,%cfg) = @_;
    return "$d/$cfg{imageset}/permissive_list.txt";
}

sub findBestGroundTruthDir(%) {
    my (%cfg) = @_;
    return $cfg{pathname};
}

sub findBestMaskFile(%) {
    my (%cfg) = @_;
    my $d = findBestGroundTruthDir(%cfg);
    return maskFileFromCalibTxt("$d/calib.txt");
}

sub maskFileFromCalibTxt($) {
    my ($calibfile) = @_;
    open F, $calibfile;
    while (my $line = <F>) {
        if ($line =~ /calibrationimage\s+(.*)/) {
            my @parts = split(/\./, $1);
            my $mask = join('.', @parts[0..$#parts-1], 'mask', $parts[$#parts]);
            die "could not find mask file in $calibfile" unless (-f $mask);
            close F;
            return $mask;
        }
    }
    die "could not find calibration image data in $calibfile";
}

sub stillFrameFileFromCalibTxt($) {
    my ($calibfile) = @_;
    open F, $calibfile;
    while (my $line = <F>) {
        if ($line =~ /calibrationimage\s+(.*)/) {
            my $stillframe = $1;
            die "still frame file specified in $calibfile does not exist" unless (-f $stillframe);
            close F;
            return $stillframe;
        }
    }
    die "could not find still frame file in $calibfile";
}

###############################################################################
###############################################################################
####  Configuration customization
###############################################################################
###############################################################################

sub customizeFullImagesCfg(%) {
    my %cfg = @_;

    # determine extra params for unscaled, directories with
    # uncropped unlabeled pos/neg windows.

    ($cfg{dataset} eq "grandcentral") ||
        die "Hard-coded people sizes only made for grandcentral right ".
        "now: '$cfg{pathname}'";

    $cfg{nononmax} = 0;
   
    # Slider control customization
    if (-f "$cfg{dir}/full/calib.txt") {
        # Found perspective calibration data
        $cfg{slidertype}  = 2;
        $cfg{calibfile}   = "$cfg{dir}/full/calib.txt";
        open(F, $cfg{calibfile});
        while (<F>) {
            s/^\s+//;
            my @tokens = split /\s+/;
            if ($tokens[0] eq 'calibrationimage') {
                $cfg{calibrationimage} = $tokens[1];
                my @pathparts = split(/\//, $cfg{calibrationimage});
                my @fparts    = split(/\./, $pathparts[$#pathparts]);
                my $avi = join('/', @pathparts[0..$#pathparts-1], 
                               join('.', @fparts[0..$#fparts-2], 'avi'));
                if (-f $avi) {
                    $cfg{video} = $avi;
                } else {
                    my $mpg = join('/', @pathparts[0..$#pathparts-1], 
                                   join('.', @fparts[0..$#fparts-2], 'mpg'));
                    if (-f $mpg) {
                        $cfg{video} = $mpg;
                    }
                }
            } else {
                $cfg{imagewidth}  = int($tokens[1]) if ($tokens[0] eq 'imagewidth');
                $cfg{imageheight} = int($tokens[1]) if ($tokens[0] eq 'imageheight');
                $cfg{rodrigues}   = join(',', @tokens[1..3]) if ($tokens[0] eq 'om');
                $cfg{translation} = join(',', @tokens[1..3]) if ($tokens[0] eq 'T');
                $cfg{focalLength} = join(',', @tokens[1..2]) if ($tokens[0] eq 'f');
            }
        }
        close(F);

        $cfg{startscale}     = 0.9;
        $cfg{scaleratio}     = 1.05;
        $cfg{endscale}       = 1.1;
        $cfg{nonmaxsigma}    = '4,8,1.3';
        $cfg{scaleprecision} = 0.05;

        # Custom-tuned parameters for various HOG geometries
        if ($cfg{hogId} eq "w16x32c8x8") {
            $cfg{svmthreshold} = -1;
        } elsif ($cfg{hogId} eq "w16x32c2x2") {
            $cfg{svmthreshold} = -2.25;
        } elsif ($cfg{hogId} eq "w12x30c6x6") {
            $cfg{svmthreshold} = -0.9999;
        } else {
            $cfg{svmthreshold} = 0;
        }

        # Custom-tuned parameters for some videos (20061026-westwall
        # has some calibration mistakes that goof up the scaling)
        if ($cfg{imageset} =~ /20061026-westwall/) {
            $cfg{startscale}  = 0.6;
            $cfg{endscale}    = 0.8;
            $cfg{nonmaxsigma} = '4,8,1.3';
        }

    } else {
        # No perspective calibration data found: default to a
        # pyramidal slider
        $cfg{slidertype} = 1; 
        
        if ($cfg{imageset} eq "1159407741750") { # 2 bins
            $cfg{minHumanHeight} =  80;
            $cfg{maxHumanHeight} = 180;
        } elsif ($cfg{imageset} =~ /20061026-westwall/) { # great hall (std def)
            $cfg{minHumanHeight} =  30;
            $cfg{maxHumanHeight} =  48;
        } elsif ($cfg{imageset} =~ /metlife/) { # escalator
            $cfg{minHumanHeight} =  90;
            $cfg{maxHumanHeight} = 140;
        } elsif (($cfg{imageset} =~ /westwall/) || 
                 ($cfg{imageset} =~ /eastwall/)) { # great hall, high def
            $cfg{minHumanHeight} =  50;
            $cfg{maxHumanHeight} =  120;
        } elsif (($cfg{imageset} == "1161257763406-7.30am") ||
                 ($cfg{imageset} == "1157719641656") ||
                 ($cfg{imageset} == "1157719419546") ||
                 ($cfg{imageset} == "1161259437437.8.03am")){ # pre-visit ones to skip for now
            return;
        } else {
            die "Unknown grandcentral sequence: '$cfg{pathname}'";
        }
        
        my $h2w = (($cfg{winsizex} - $cfg{innerpaddingx}*2) / 
                   ($cfg{winsizey} - $cfg{innerpaddingy}*2)); # aspect ratio
        $cfg{startscale} = ($cfg{minHumanHeight} * $h2w / 
                            ($cfg{winw} - 2*$cfg{innerpaddingx}));
        $cfg{endscale} = ($cfg{maxHumanHeight} * $h2w / 
                          ($cfg{winw} - 2*$cfg{innerpaddingx}));
        my $maxMinRatio = ($cfg{endscale}/$cfg{startscale});
        my $minMaxDiff  = ($cfg{endscale} - $cfg{startscale});
        my $pxDiffPerScale = 2;
        $cfg{scaleratio} =  $maxMinRatio ** ($pxDiffPerScale/$minMaxDiff);
    }

    # Detector parameter tuning
    if (($cfg{imageset} eq 'ngct01') || ($cfg{imageset} =~ /metlife/)) {
        $cfg{svmthreshold} = -1.75;
        $cfg{nononmax}     = 0;
        $cfg{nonmaxsigma}  = "2,3,1.3";
        $cfg{threshold}    = $cfg{svmthreshold};
        $cfg{showscore}    = 0;
    } elsif ($cfg{imageset} =~ /20061026-westwall/) {
        $cfg{svmthreshold}  = -0.5;
        $cfg{threshold}     = $cfg{svmthreshold};
        $cfg{showscore}     = 0; # scene is too dense to see anything if scores are shown
    }

    return %cfg;
}

sub customizePosNegCfg($%) {
    my ($lst,%cfg) = @_;

    # parsing of pos/neg filename to get train/test params

    # $cfg{basename} examples:
    #   pos_20080211_w16x32c8x8_all.lst
    #   neg_full.lst

    my @fnparts = split('_', $cfg{basename});
    if (@fnparts == 2) {
        $cfg{posneg}  = $fnparts[0];
        $cfg{annoset} = $fnparts[1];
        checkFieldExists($lst,'hogId',%cfg);
        checkFieldExists($lst,'windowset',%cfg);
        checkFieldExists($lst,'winw',%cfg);
        checkFieldExists($lst,'winh',%cfg);
        checkFieldExists($lst,'cellw',%cfg);
        checkFieldExists($lst,'cellh',%cfg);
        checkFieldExists($lst,'innerpaddingx',%cfg);
        checkFieldExists($lst,'innerpaddingy',%cfg);
        checkFieldExists($lst,'numcellx',%cfg);
        checkFieldExists($lst,'numcelly',%cfg);
        checkFieldExists($lst,'blockstridex',%cfg);
        checkFieldExists($lst,'blockstridey',%cfg);
        checkFieldExists($lst,'orientbin',%cfg);
        checkFieldExists($lst,'outerpaddingx',%cfg);
        checkFieldExists($lst,'outerpaddingy',%cfg);
        checkFieldExists($lst,'winstridex',%cfg);
        checkFieldExists($lst,'winstridey',%cfg);

    } elsif (@fnparts == 4) {    
        $cfg{posneg}    = $fnparts[0];
        $cfg{annoset}   = $fnparts[1];
        $cfg{hogId}     = $fnparts[2];
        $cfg{windowset} = $fnparts[3];

        ($cfg{hogId} =~ /^w([0-9]+)x([0-9]+)c([0-9]+)x([0-9]+)$/) || 
            die("Could not parse HOG ID '$cfg{hogId}' in '$lst'");
        $cfg{winw}  = $1;
        $cfg{winh}  = $2;
        $cfg{cellw} = $3;
        $cfg{cellh} = $4;

        # Read one image to get size information
        open LST,"$cfg{pathname}";
        my $firstImg;
        ($firstImg = <LST>) || die "Could not read '$cfg{pathname}'";
        close LST;
        chomp $firstImg;
        my $image = Image::Imlib2->load("$cfg{dir}/$firstImg");
        my $w = $cfg{firstimgw} = $image->width();
        my $h = $cfg{firstimgh} = $image->height();
        
        # implicit fields
        ($cfg{cellw} == $cfg{cellh}) || 
            die("The cell width ($cfg{cellw}) and heights ($cfg{cellh}) ".
                "must be the same for the currently-recognized descriptor ".
                "geometries: '$lst'.");
        if ($cfg{cellw} == 2) {
            $cfg{innerpaddingx} = 4;
            $cfg{innerpaddingy} = 4;
            $cfg{numcellx}      = $cfg{numcelly}      =  5;
            $cfg{blockstridex}  = $cfg{blockstridey}  =  6;
            $cfg{orientbin} = 5; 
            
        } elsif ($cfg{cellw} == 4) {
            $cfg{innerpaddingx} = 4;
            $cfg{innerpaddingy} = 4;
            $cfg{numcellx}      = $cfg{numcelly}      =  3;
            $cfg{blockstridex}  = $cfg{blockstridey}  =  4;
            $cfg{orientbin} = 5; 
            
        } elsif ($cfg{cellw} == 6) {
            $cfg{innerpaddingx} = 2;
            $cfg{innerpaddingy} = 3;
            $cfg{numcellx}      = $cfg{numcelly}      =  2;
            $cfg{blockstridex}  = $cfg{blockstridey}  =  6;
            $cfg{orientbin} = 5; 
            
        } elsif ($cfg{cellw} == 8) {
            if ($cfg{winw} == 64) {
                $cfg{innerpaddingx} = 16;
                $cfg{innerpaddingy} = 16;
            } elsif ($cfg{winw} == 16) {
                $cfg{innerpaddingx} = 4;
                $cfg{innerpaddingy} = 4;
            } else {
                die "Unknown window width: $cfg{pathname}";
            }
            $cfg{numcellx}      = $cfg{numcelly}      =  2;
            $cfg{blockstridex}  = $cfg{blockstridey}  =  8;
            $cfg{orientbin} = 9; 
            
        } else {
            die("Unsupported cellsize ($cfg{cellw}) in '$lst'");
        }

        # compute correct outerpadding (turns into --topleft arg)
        $cfg{outerpaddingx} = ($w - $cfg{winw}) / 2;
        $cfg{outerpaddingy} = ($h - $cfg{winh}) / 2;
        ($cfg{outerpaddingx} == int($cfg{outerpaddingx})) || 
            die("odd-sized image chips not allowed");
        ($cfg{outerpaddingy} == int($cfg{outerpaddingy})) || 
            die("odd-sized image chips not allowed");
        
        $cfg{winstridex} = $cfg{blockstridex};
        $cfg{winstridey} = $cfg{blockstridey};
    } else {
        die("Unsupported number of parts to the .lst filename ".
            "'$cfg{basename}' for '$lst'");
    }
    
    # Slider controls
    if ($cfg{posneg} eq "pos") {
        $cfg{slidertype} = 0;
    } elsif ($cfg{posneg} eq "neg") {
        $cfg{slidertype} = ($cfg{annoset} eq 'full') ? 1 : 0;
    } else {
        die "Can't tell if '$cfg{pathname}' lists positive or negative ".
            "examples.";
    }
    
    $cfg{startscale} = 1;
    $cfg{endscale}   = 0; 
    $cfg{scaleratio} = 1.2;

    if ($cfg{slidertype}==1) {
        $cfg{finestridex} = $cfg{finestridey} = 1;
    }

    return %cfg;
}

###############################################################################
###############################################################################
####  Argument assembly for learcode
###############################################################################
###############################################################################

#--------------------------------------------------------------------------------
# Options multitree:
#
#             oShared           oSvmLearn
#             /    \
#          oRhog   oDump4svmlearn
#         /     \
#     oTest      oClassifyHard
#       |  \
#     oTrk  oTrkFiltering

sub oShared() {
    return "--verbose=6 ";
}

# Options for dump4svmlearn
sub oDump4svmlearn() {
    return oShared().
        "--maxpos=-1 ".         # maximum positive vectors, -1 implies no limit
        "--maxneg=-1 ".         # maximum negative vectors, -1 implies no limit
        #"--format=BiSVMLight ". # Options: BiSVMLight  SVMLight  BiMatlab  Matlab
        "";
}

# svm_learn options
sub oSvmLearn() {
    my $svmPosToNegCostFactor  = 3;
    my $svmAreInputFilesBinary = 1;
    my $svmMode                = 'c'; # c=classification, r=regression, p=preference ranking
    my $svmKernelType          = 0;   # 0=linear, 1=poly, 2=rbf, 3=sigmoid, 4=user defined

    return 
        "-j $svmPosToNegCostFactor ".
        "-B $svmAreInputFilesBinary ".
        "-z $svmMode ".
        "-v 1 ".
        "-t $svmKernelType ";
}

# Options for dump_rhog and classify_rhog
sub oRhog($%) {
    my ($impl,%cfg) = @_;
    my $iscuda = -1;
    $iscuda=0 if ($impl=~/cpu/i);
    $iscuda=1 if ($impl=~/gpu/i);
    $cfg{quantizepos} =  0 unless (defined $cfg{quantizepos});
    $cfg{finestridex} = -1 unless (defined $cfg{finestridex});
    $cfg{finestridey} = -1 unless (defined $cfg{finestridey});
    $cfg{imagewidth}  = -1 unless (defined $cfg{imagewidth});
    $cfg{imageheight} = -1 unless (defined $cfg{imageheight});
    die "unrecognized implementation specification: $impl" if ($iscuda==-1);
    return oShared().
        "--cuda=$iscuda ". 
        "--imageext=png ".
        "--initialsize=$cfg{imagewidth}x$cfg{imageheight} ".
        # Detections are made for detector windows that are (optionally) scanned
        # across the image at a pyramid of scales.
        "--window=$cfg{winw},$cfg{winh} ".                    # size of the detector window (in pixels)
        "--winstride=$cfg{winstridex},$cfg{winstridey} ".     # sample detector windows every <winstride> pixels
        "--finestride=$cfg{finestridex},$cfg{finestridey} ".  # sample HOG descriptors every <finestride> pixels when precomputing them with CUDA
        "--topleft=$cfg{outerpaddingx},$cfg{outerpaddingy} ". # shift the top left detector window by this amount (in pixels)
        "--quantizepos=$cfg{quantizepos} ".                   # quantize window locations to a finestride grid (only applies to CUDA)
        # Each descriptor window is broken down into a number of cells.  To provide 
        # some spatial invariance, sets of cells are grouped together in blocks.  
        # Blocks overlap each other.  The whole descriptor is a concatenation of block 
        # descriptors.
        "--cellsize=$cfg{cellw},$cfg{cellh} ".                  # size of each cell (in pixels)
        "--numcell=$cfg{numcellx},$cfg{numcelly} ".             # number of cells in each block
        "--orientbin=$cfg{orientbin} ".                         # number of orientation bins in each cell/block (a block's histogram is created by adding the histograms of each of the block's cells)
        "--descstride=$cfg{blockstridex},$cfg{blockstridey}  ". # how far away are the cell centers from each other (usually the same as cellsize)
        # Histogram parameters
        "--proc=rgb_sqrt ".   # image colorspace (preprocessing).  Options: rgb  lab  rgb_sqrt  lab_sqrt  rgb_log
        "--norm=l2hys ".      # histogram normalization method.  Options: nonorm  l1  l2  l1sqrt  l2hys  
        "--fullcirc=0  ".     # 0->use orientations in range (0-180), 1->use orientations in range (0-360)
        "--wtscale=2 ".       # magnitude weighting scale
        "--gscale=0 ".        # gradient scale
        "--maxvalue=0.2 ".    # chop feature vector values > maxvalue after normalizing, iff norm=l2hys
        "--epsilon=1 ".       # add epsilon to each bin while normalizing
        # Non-understood params: fullstride, alignin, alignmargin
        "";
}

# classify_rhog production of hard examples
sub oClassifyHard($%) {
    my ($impl,%cfg) = @_;
    my $iscuda = -1;
    $iscuda=0 if ($impl=~/cpu/i);
    $iscuda=1 if ($impl=~/gpu/i);
    return oRhog($impl,%cfg).
        "--cachesize=1024 ".        # descriptor cache size (in MB).  Note: the library is 32-bit.
        #
        "--margin=0 ".              # margin for normalized image (in pixels)
        "--avsize=0 ".              # compute image margin to add using average size (seems to override the margin parameter when avsize!=0)
        #
        "--svmthreshold=0 ".        # Initial pure SVM output threshold (high hysteresis threshold)
        "--threshold=0 ".           # Final classification threshold (low hysteresis threshold)
        "--score2prob=1,0 ".        # (A,B) params of a sigmoid function 1/(1+exp(A*score+B)) or soft clip log(1+exp(A*(score-B))).  Small A softens the curve, B shifts the 0.5 probability point.
        "--no_nonmax=1 ".           # Skip non-maximal suppression
        "--nonmaxsigma=8,16,1.3 ".  # (x,y,smooth scale) for non-maximal suppression
        "--softmax=0 ".             # score->prob transformation function: 0=hardclip, 1=sigmoid, 2=softclip, 3=none
        #
        "--startscale=$cfg{startscale} ". # smallest scale (smaller numbers = more zooming in)
        "--scaleratio=$cfg{scaleratio} ". # what it says
        "--endscale=$cfg{endscale} ".     # largest scale; 0=zoom out until there's just one window in the image
        "--slider=$cfg{slidertype} ".     # pyramid search
        #
        "--label=N ".               # Only negative examples are classified the first time
        "";
}

# Final classify_rhog results, shared parameters 
sub oTest($%) {
    my ($impl,%cfg) = @_;

    # optimize finestride
    if ($cfg{slidertype} == 0) { # centered
        # Pick a finer finestride if necessary to allow computing
        # descriptors for a centered window.  Since all images tend to
        # be the same size for a given test, we'll just look at the
        # first one's size.
        $cfg{finestridex} = gcd($cfg{winstridex}, $cfg{blockstridex}, 
                                int(($cfg{firstimgw} - $cfg{winsizex})/2));
        $cfg{finestridey} = gcd($cfg{winstridey}, $cfg{blockstridey}, 
                                int(($cfg{firstimgh} - $cfg{winsizey})/2));
    } elsif ($cfg{slidertype} == 1) { # pyramid
        $cfg{finestridex} = $cfg{finestridey} = 1;
    } elsif ($cfg{slidertype} == 2) { # perspective
        # do nothing: trust the caller
    } else {
        die "Unknown sliderType: $cfg{slidertype}";
    }

    # tweak classifier params if desired
    my $svmthresh   = getWithDefault('svmthreshold', 0,             %cfg);
    my $noNonMax    = getWithDefault('nononmax',     0,             %cfg);
    my $nonmaxsigma = getWithDefault('nonmaxsigma',  '8,16,1.3',    %cfg);
    my $finalthresh = getWithDefault('threshold',    $svmthresh,    %cfg);
    my $showscore   = getWithDefault('showscore',    1,             %cfg);

    # assemble args
    return oRhog($impl,%cfg).
        "--cachesize=1024 ".           # descriptor cache size (in MB).  Note: the library is 32-bit.
        #
        "--margin=0 ".                 # margin for normalized image (in pixels)
        "--avsize=0 ".                 # compute image margin to add using average size (seems to override the margin parameter when avsize!=0)
        # Final classifier
        "--svmthreshold=$svmthresh ".  # Initial pure SVM output threshold (high hysteresis threshold)
        "--threshold=$finalthresh ".   # Final classification threshold (low hysteresis threshold)
        "--score2prob=1,1 ".           # (A,B) params of a sigmoid function 1/(1+exp(A*score+B)) or soft clip log(1+exp(A*(score-B))).  Small A softens the curve, B shifts the 0.5 probability point.
        "--no_nonmax=$noNonMax ".      # Skip non-maximal suppression
        "--nonmaxsigma=$nonmaxsigma ". # (x,y,smooth scale) for non-maximal suppression
        "--softmax=0 ".                # score->prob transformation function: 0=hardclip, 1=sigmoid, 2=softclip, 3=none
        # Extra geometric parameters for perspective sliders.  Most parameters
        # come from the Matlab Camera Calibration Toolbox.
        "--borderpixels=$cfg{innerpaddingy} ".                  # needed to compute geometries when using perspective sliders
        "--rodrigues=$cfg{rodrigues} ".                         # rodrigues rotation vector
        "--translation=$cfg{translation} ".                     
        "--focalLength=$cfg{focalLength} ".
        "--scaleprecision=$cfg{scaleprecision} ".
        # Scale
        "--slider=$cfg{slidertype} ".       # 0=centered, 1=pyramid search, 2=persp.
        "--startscale=$cfg{startscale} ".   # smallest scale
        "--scaleratio=$cfg{scaleratio} ".   # what it says
        "--endscale=$cfg{endscale} ".       # largest scale; 0=full image
        # 
        "--showscore=$showscore ".     # show score when writing outimage
        "";
}

# use classify_rhog with a Kalman tracker
sub oTrk($%) {
    my ($impl,%cfg) = @_;

    my $huge = 1000*1000;
    return oTest($impl,%cfg).
        "--minvxy=9  --ivxy=25 --maxvxy=25 ".       # variance on position
        "--minvdxy=1 --ivdxy=4 --maxvdxy=4 ".       # variance on velocity
        "--minva=$huge --iva=$huge --maxva=$huge ". # variance on area (for now, basically ignore scale)
        "--matchdist=2.5 ".                         # max standard deviations to be a match
        "--deldist=$huge ".                         # don't delete just because there are no good match candidates in a single frame
        "";
}

# Use trk2trk to re-do the data association and add track filtering,
# because it's much faster to redo these stages with existing tracks
# than to re-run classify_rhog.
sub oTrkFiltering($%) {
    my ($impl,%cfg) = @_;
    my $huge = 1000*1000;

    # For now, we won't bother projecting down to the image plane.
    my $pxToTrackerCoords = 1;
    my $frToFilterCoords  = 1;

    my $maxCc        = 1024;
    my $fps          = 29.97;
    my $searchFrames = 5;
    my $searchTime   = $searchFrames * 1000 / $fps;
    my $maxTrk       = $searchFrames * $maxCc;

    return "".
        "    minva=$huge iva=$huge maxva=$huge ".
        "".
        "    minvxy=".sq(3*$pxToTrackerCoords).
        "    ivxy=".sq(5*$pxToTrackerCoords).
        "    maxvxy=".sq(5*$pxToTrackerCoords).
        "".
        "    minvdxy=".sq(sqrt(0.5)*$pxToTrackerCoords).
        "    ivdxy=".sq(1.5*$pxToTrackerCoords).
        "    maxvdxy=".sq(1.5*$pxToTrackerCoords).
        "".
        "    maxmatchdist=2.5 deldist=$huge ".
        "    maxcc=$maxCc maxtrk=$maxTrk maxUnObsMs=$searchTime ".
        "    minccarea=0 nlogpnew=80 ".
        "".
        "    filteringSetup='". # filtering happens after conversion back to feet
        (       )."filter=1".
        (       )." minDist=".(0.01*$frToFilterCoords).
        (       )." maxSpeed=".(0.05*$frToFilterCoords).
        (       )." aspectRatio=1".
        (       )." centroidSmoothingSigma=5".
        (       )." centroidSmoothingTaps=25".
        (       )." maxFrameGap=2".
        (       )." skippedStartObs=5".
        (       )." skippedEndObs=5".
        (       )." minFrames=50".
        (       )."'";
}


###############################################################################
###############################################################################
####  Rule generating functions
###############################################################################
###############################################################################

# Creates the makefile rules corresponding to one full set of
# experiments given a collection of training & test data and
# parameters such as the HOG geometry.  The rule set created
# corresponds to those outlined in ../doc/dataflow.*
sub writeExperimentTree($$$$$$$$$) {
    my ($impl,$hPosCfg,$hNegCfg,$hPerspTrainCfgs,
        $hPosTestCfgs,$hNegTestCfgs,$hPerspTestCfgs,$hFullImgCfgs,$hTargets) = @_;
    my %posCfg          = %{$hPosCfg};
    my %negCfg          = %{$hNegCfg};
    my @perspTrainCfgs  = @{$hPerspTrainCfgs};
    my @posTestCfgs     = @{$hPosTestCfgs};
    my @negTestCfgs     = @{$hNegTestCfgs};
    my @perspTestCfgs   = @{$hPerspTestCfgs};
    my @testFullImgCfgs = @{$hFullImgCfgs};

    my $classifier = join('/', ichomp(`pwd`), 
                          "$posCfg{imageset}_$posCfg{annoset}",
                          "$posCfg{hogId}_$posCfg{windowset}_$impl");  

    my $firstPassDir        = "$classifier/firstPass";
    my $firstPassPosRhog    =   "$firstPassDir/pos.RHOG";
    my $firstPassNegRhog    =   "$firstPassDir/neg.RHOG";
    my $firstPassSvmInput   =   "$firstPassDir/train.blt";
    my $firstPassModelFile  =   "$firstPassDir/model.alt";
    my $firstPassHardNegTxt =   "$firstPassDir/hard_neg.txt";

    my $hardDir             = "$classifier/hard";
    my $hardNegRhog         =   "$hardDir/hard_neg.RHOG";
    my $hardSvmInput        =   "$hardDir/train.blt";
    my $hardModelFile       =   "$hardDir/model.alt";
    my @hardPermTxts        =   map { permissiveLst($hardDir,%{$_}) } @perspTrainCfgs;
    my @hardPermMasks       =   map { findBestMaskFile(%{$_})       } @perspTrainCfgs;
    my @hardPermGtDirs      =   map { findBestGroundTruthDir(%{$_}) } @perspTrainCfgs;

    my $perspDir            = "$classifier/persp";
    my $perspPosTxt         =   "$perspDir/pos.txt";
    my $perspNegTxt         =   "$perspDir/neg.txt";
    my $perspFalsePosTxt    =   "$perspDir/false_pos.txt";
    my $perspFalseNegTxt    =   "$perspDir/false_neg.txt";
    my $perspPosRhog        =   "$perspDir/pos.RHOG";
    my $perspNegRhog        =   "$perspDir/neg.RHOG";
    my $perspFalsePosRhog   =   "$perspDir/false_pos.RHOG";
    my $perspFalseNegRhog   =   "$perspDir/false_neg.RHOG";
    my $perspSvmInput       =   "$perspDir/train.blt";
    my $perspModelFile      =   "$perspDir/model.alt";    

    # Consolidate perspective training configs because we'll be
    # operating on consolidated list.txt files that contain window
    # data from all of the configs.
    my %perspTrainCfg = %{$perspTrainCfgs[0]};
    for (my $i=1; $i<@perspTrainCfgs; $i++) {
        my %tmp = %{$perspTrainCfgs[$i]};
        for my $field ("hogId", "scaleprecision") {
            if ($tmp{$field} != $perspTrainCfg{$field}) {
                die "All perspective-trained configs must use the same $field";
            }
        }
    }
    # Limit memory consumption on the CUDA side for dump_rhog so that
    # we can handle 1920x1080 frames with ~2x upsampling on a GTX 260+
    # card (896MB of RAM).  Presumably a Tesla card with 4GB could
    # handle finestride=1 and thus we wouldn't need to quantize
    # positions.
    $perspTrainCfg{'quantizepos'} = 1;
    $perspTrainCfg{'finestridex'} = getMinWithDefault('finestridex', 3, %perspTrainCfg);
    $perspTrainCfg{'finestridey'} = getMinWithDefault('finestridey', 3, %perspTrainCfg);



    hugeheading("TRAIN AND TEST $classifier");

    `mkdir -p $classifier $firstPassDir $hardDir $perspDir`; # do it now



    bigheading("1ST LEARNING ITERATION");

    writeRule("get features on positive images",
              $firstPassPosRhog,
              "$posCfg{pathname}",
              "dump_rhog --samples=1 $posCfg{pathname} $firstPassPosRhog ".
              oRhog($impl,%posCfg),
              "$firstPassPosRhog.log", $impl, $hTargets);

    writeRule("get features on (sampled) negative images",
              $firstPassNegRhog,
              "$negCfg{pathname}",
              "dump_rhog --samples=10 $negCfg{pathname} $firstPassNegRhog ".
              oRhog($impl,%negCfg),
              "$firstPassNegRhog.log", $impl, $hTargets);
    
    writeRule("dump positive + (sampled) negative examples to svmdense format",
              $firstPassSvmInput,
              "$firstPassPosRhog $firstPassNegRhog",
              "dump4svmlearn ".
              "  --posfile $firstPassPosRhog ".
              "  --negfile $firstPassNegRhog ".
              "  --outfile $firstPassSvmInput ".
              "  ".oDump4svmlearn(),
              "$firstPassSvmInput.log", $impl, $hTargets);

    writeRule("SVM learning with positive + (sampled) negative examples",
              $firstPassModelFile,
              "$firstPassSvmInput",
              "svm_learn ".
              "  $firstPassSvmInput ".
              "  $firstPassModelFile".
              "  ".oSvmLearn(),
              "$firstPassModelFile.log", $impl, $hTargets);

    writeRule("Classify all negatives to find hard examples",
              $firstPassHardNegTxt, # $firstPassDir/hard_list.txt $firstPassDir/hard_hist.txt 
              "$negCfg{pathname} $firstPassModelFile",
              "classify_rhog ".
              "  --infile    $negCfg{pathname} ".
              "  --modelfile $firstPassModelFile ".
              "  --falsetxt  $firstPassHardNegTxt ".
              "  --outfile   $firstPassDir/hard_list.txt ".
              "  --outhist   $firstPassDir/hard_hist.txt ".
              "  ".oClassifyHard($impl,%negCfg),
              "$firstPassHardNegTxt.log", $impl, $hTargets);

    writeTests($impl, $hPosCfg,$hPerspTrainCfgs,
               $hPosTestCfgs,$hNegTestCfgs,$hPerspTestCfgs,$hFullImgCfgs, 
               "firstPass", $hTargets);
    


    bigheading("2ND LEARNING ITERATION (HARD EXAMPLES)");

    writeRule("Dump hard examples using the false positives from the classifier",
              $hardNegRhog,
              "$firstPassHardNegTxt ",
              "dump_rhog ".
              "  --infile  $firstPassHardNegTxt ".
              "  --outfile $hardNegRhog ".
              "  --samples 0 ".
              #"  --poscases 2416 --negcases 12180 --memorylimit 1700 ".
              "  --presetexamples=1 --hardscore=0 --label=N ".
              "  ".oRhog($impl,%negCfg),
              "$hardNegRhog.log", $impl, $hTargets);

    writeRule("dump positive + (sampled) negative + hard negative examples to svmdense format",
              $hardSvmInput,
              "$firstPassPosRhog $firstPassNegRhog $hardNegRhog",
              "dump4svmlearn ".
              "  --posfile $firstPassPosRhog ".
              "  --negfile $firstPassNegRhog ".
              "  --negfile $hardNegRhog ".
              "  --outfile $hardSvmInput ".
              "  ".oDump4svmlearn(),
              "$hardSvmInput.log", $impl, $hTargets);

    writeRule("SVM learning with positive + (sampled) negative + hard negative examples",
              $hardModelFile,
              $hardSvmInput,
              "svm_learn ".
              "  $hardSvmInput ".
              "  $hardModelFile".
              "  ".oSvmLearn(),
              "$hardModelFile.log", $impl, $hTargets);

    writeTests($impl, $hPosCfg,$hPerspTrainCfgs,
               $hPosTestCfgs,$hNegTestCfgs,$hPerspTestCfgs,$hFullImgCfgs, 
               "hard", $hTargets);



    bigheading("3RD LEARNING ITERATION (PERSPECTIVE DATA)");
    
    writeRule("Create window lists from perspective data",
              "$perspDir/detopt.csv $perspPosTxt $perspNegTxt $perspFalsePosTxt $perspFalseNegTxt",
              join(' ', @hardPermTxts, @hardPermMasks, @hardPermGtDirs),
              "detoptimizer ".
              "  --outdir=$perspDir ".
              "  --noparamsubdirs=true ".
              "  --csvout=true ".
              "  --postxtout=true ".
              "  --negtxtout=true ".
              "  --falsepostxtout=true ".
              "  --falsenegtxtout=true ".
              "  --imageout=false ". 
              "  --maxdatafrac=1 ".
              "  --minthresh=0 --maxthresh=0 --nthresh=1 ". 
              "  --minnonmaxsigma=4,8,100 --nnonmaxsigma=1 ".
              "  --posscore=-10 ".
              "  --negscore=10 ".
              "  --samples=5000 ".
              "  --startscale=$perspTrainCfg{startscale} ". # assume they're all the same
              "  --endscale=$perspTrainCfg{endscale} ".
              "  --scaleprecision=$perspTrainCfg{scaleprecision} ".
              "  --threads=3 ".
              join('', map { "  --txtfile=$_ "  } @hardPermTxts)." ".
              join('', map { "  --gtdir=$_ "    } @hardPermGtDirs)." ".
              join('', map { "  --maskfile=$_ " } @hardPermMasks)." ",
              "$perspDir/detopt.csv.log", $impl, $hTargets);  

    writeRule("Dump perspective training examples (positives)",
              $perspPosRhog,
              $perspPosTxt,
              "dump_rhog ".
              "  --infile   $perspPosTxt ".
              "  --outfile  $perspPosRhog ".
              "  --samples  0 ".
              # TODO: poscases, negcases, memorylimit?
              "  --presetexamples=true --hardscore=0 --label=P ".
              "  --suppressnonjittered=false ".
              "  ".oRhog($impl,%perspTrainCfg),
              "$perspPosRhog.log", $impl, $hTargets);

    writeRule("Dump perspective training examples (negatives)",
              $perspNegRhog,
              $perspNegTxt,
              "dump_rhog ".
              "  --infile   $perspNegTxt ".
              "  --outfile  $perspNegRhog ".
              "  --samples  0 ".
              # TODO: poscases, negcases, memorylimit?
              "  --presetexamples=true --hardscore=0 --label=N ".
              "  --suppressnonjittered=false ".
              "  ".oRhog($impl,%perspTrainCfg),
              "$perspNegRhog.log", $impl, $hTargets);

    writeRule("Dump perspective training examples (false positives)",
              $perspFalsePosRhog,
              $perspFalsePosTxt,
              "dump_rhog ".
              "  --infile   $perspFalsePosTxt ".
              "  --outfile  $perspFalsePosRhog ".
              "  --samples  0 ".
              # TODO: poscases, negcases, memorylimit?
              "  --presetexamples=true --hardscore=0 --label=N ".
              "  --suppressnonjittered=false ".
              "  ".oRhog($impl,%perspTrainCfg),
              "$perspFalsePosRhog.log", $impl, $hTargets);

    writeRule("Dump perspective training examples (false negatives)",
              $perspFalseNegRhog,
              $perspFalseNegTxt,
              "dump_rhog ".
              "  --infile   $perspFalseNegTxt ".
              "  --outfile  $perspFalseNegRhog ".
              "  --samples  0 ".
              # TODO: poscases, negcases, memorylimit?
              "  --presetexamples=true --hardscore=0 --label=P ".
              "  --suppressnonjittered=false ".
              "  ".oRhog($impl,%perspTrainCfg),
              "$perspFalseNegRhog.log", $impl, $hTargets);

    writeRule("dump positive + (sampled) negative + hard negative + perspective examples to svmdense format",
              $perspSvmInput,
              join(' ', $firstPassPosRhog, $firstPassNegRhog, $hardNegRhog, 
                   $perspPosRhog, $perspNegRhog, 
                   $perspFalsePosRhog, $perspFalseNegRhog),
              "dump4svmlearn ".
              "  --posfile $firstPassPosRhog ".
              "  --posfile $perspPosRhog ".
              "  --posfile $perspFalseNegRhog ".
              "  --negfile $firstPassNegRhog ".
              "  --negfile $hardNegRhog ".
              "  --negfile $perspNegRhog ".
              "  --negfile $perspFalsePosRhog ".
              "  --outfile $perspSvmInput ".
              "  ".oDump4svmlearn(),
              "$perspSvmInput.log", $impl, $hTargets);
              
    writeRule("SVM learning with positive + (sampled) negative + hard negative + perspective examples",
              $perspModelFile,
              $perspSvmInput,
              "svm_learn ".
              "  $perspSvmInput ".
              "  $perspModelFile".
              "  ".oSvmLearn(),
              "$perspModelFile.log", $impl, $hTargets);

    writeTests($impl, $hPosCfg,$hPerspTrainCfgs,
               $hPosTestCfgs,$hNegTestCfgs,$hPerspTestCfgs,$hFullImgCfgs, 
               "persp", $hTargets);
}

sub writeTests($$$$$$$$$) {
    my ($impl,$hPosCfg,$hPerspTrainCfgs,
        $hPosTestCfgs,$hNegTestCfgs,$hPerspTestCfgs,$hFullImgCfgs,
        $trainRegime, $hTargets) = @_;
    my %posCfg          = %{$hPosCfg};
    my @perspTrainCfgs  = @{$hPerspTrainCfgs};
    my @posTestCfgs     = @{$hPosTestCfgs};
    my @negTestCfgs     = @{$hNegTestCfgs};
    my @perspTestCfgs   = @{$hPerspTestCfgs};
    my @testFullImgCfgs = @{$hFullImgCfgs};

    writeInriaStyleTests($impl, \%posCfg, \@posTestCfgs, \@negTestCfgs, $trainRegime,$hTargets);
    writeNonmaxExpts($impl, \%posCfg, \@testFullImgCfgs, $trainRegime,$hTargets);
    writeDetOptimizerExpt($impl, \%posCfg, \@perspTrainCfgs, $trainRegime, "train_error",$hTargets);
    writeDetOptimizerExpt($impl, \%posCfg, \@perspTestCfgs,  $trainRegime, "test_error",$hTargets);
    writeTrackerExpts($impl, \%posCfg, \@testFullImgCfgs, $trainRegime,$hTargets);
}

sub writeInriaStyleTests($$$$$$) {
    my ($impl, $hPosTrainCfg, $hPosTestCfgs, $hNegTestCfgs, $trainRegime,$hTargets) = @_;
    my %posTrainCfg = %{$hPosTrainCfg};
    my @posTestCfgs = @{$hPosTestCfgs};
    my @negTestCfgs = @{$hNegTestCfgs};

    for (my $i=0; $i<@posTestCfgs; $i++) {
        my %posTestCfg = %{$posTestCfgs[$i]};
        my %negTestCfg = %{$negTestCfgs[$i]};

        ($posTrainCfg{hogId} eq $posTestCfg{hogId}) || 
            die("The feature geometry must be the same for the training ".
                "($posTrainCfg{pathname}) and test ($posTrainCfg{pathname}) ".
                "sets.");

        my $classifierName = "$posTrainCfg{imageset}_$posTrainCfg{annoset}/".
            "$posTrainCfg{hogId}_$posTrainCfg{windowset}_$impl/$trainRegime";
        my $classifier = ichomp(`pwd`)."/$classifierName";

        # Data sources
        my $modelFile = "$classifier/model.alt";

        # Additional output locations
        my $testName = "$posTestCfg{imageset}_$posTestCfg{annoset}_$posTestCfg{windowset}";
        my $testDir  = "$classifier/$testName";
        
        bigheading("TEST $classifierName ON THE $testName TESTSET");

        `mkdir -p $testDir`; # do it now
        
        writeRule("Test positive examples (no scalespace)",
                  "$testDir/pos_list.txt $testDir/pos_hist.txt",
                  "$posTestCfg{pathname} $modelFile",
                  "classify_rhog ".
                  "  --infile    $posTestCfg{pathname} ".
                  "  --modelfile $modelFile ".
                  "  --outfile   $testDir/pos_list.txt ".
                  "  --outhist   $testDir/pos_hist.txt ".
                  "  ".oTest($impl,%posTestCfg),
                  "$testDir/pos_list.log", $impl, $hTargets);
        
        writeRule("Test (sampled) negative examples over scalespace",
                  "$testDir/neg_list.txt $testDir/neg_hist.txt",
                  "$negTestCfg{pathname} $modelFile",
                  "classify_rhog ".
                  "  --infile    $negTestCfg{pathname} ".
                  "  --modelfile $modelFile ".
                  "  --outfile   $testDir/neg_list.txt ".
                  "  --outhist   $testDir/neg_hist.txt ".
                  "  ".oTest($impl,%negTestCfg),
                  "$testDir/neg_list.log", $impl, $hTargets);
    }
}

sub writeNonmaxExpts($$$$$) {
    my ($impl,$hPosTrainCfg,$hTestCfgs,$trainRegime,$hTargets) = @_;

    my %posTrainCfg = %{$hPosTrainCfg};
    my @testCfgs    = @{$hTestCfgs};

    for my $hTestCfg (@testCfgs) {
        my %testCfg = %{$hTestCfg};

        if ($testCfg{isLst}) {
            ($posTrainCfg{hogId} eq $testCfg{hogId}) || 
                die("The feature geometry must be the same for the training ".
                    "($posTrainCfg{pathname}) and test ($posTrainCfg{pathname}) ".
                    "sets.");
        }

        my $classifierName = "$posTrainCfg{imageset}_$posTrainCfg{annoset}/".
            "$posTrainCfg{hogId}_$posTrainCfg{windowset}_$impl/$trainRegime";
        my $classifier = ichomp(`pwd`)."/$classifierName";

        # Patch up test
        my %cfg = mergehash(%posTrainCfg,%testCfg);
        $cfg{outerpaddingx} = $cfg{outerpaddingy} = 0;

        # Data sources
        my $modelFile = "$classifier/model.alt";
        my $indir     = "$cfg{dir}/full";

        # Additional output locations
        my $testName = "$cfg{imageset}";
        my $testDir  = "$classifier/$testName";
        my $outfile  = "$testDir/nonmax_list.txt";
        my $outhist  = "$testDir/nonmax_hist.txt";
        my $vidfile  = "$testDir.avi";
        
        bigheading("FULL IMAGE EXPERIMENT USING $classifierName ON THE $cfg{imageset} IMAGESET");

        `mkdir -p $testDir`; # do it now
        
        writeRule("Detect pedestrians with non-maxima suppression enabled",
                  "$outfile $outhist",
                  "$modelFile",
                  "classify_rhog ".
                  "  --infile    $indir ".
                  "  --outfile   $outfile ".
                  "  --modelfile $modelFile ".
                  "  --outhist   $outhist ".
                  "  --outimage  $testDir ".
                  "  --txtimgnames=1 ".
                  "  ".oTest($impl,%cfg), 
                  "$outfile.log", $impl, $hTargets);

        #writeRule("Make a video",
        #          "$vidfile",
        #          "$outfile",
        #          "ffmpeg -y -an -b 2000k -i '$testDir/%07d.jpg' $vidfile",
        #          "$vidfile.log", $impl, $hTargets);
    }
}

sub writeTrackerExpts($$$$$) {
    my ($impl,$hPosTrainCfg,$hTestCfgs,$trainRegime,$hTargets) = @_;

    my %posTrainCfg = %{$hPosTrainCfg};
    my @testCfgs    = @{$hTestCfgs};

    for my $hTestCfg (@testCfgs) {
        my %testCfg = %{$hTestCfg};

        my $classifierName = "$posTrainCfg{imageset}_$posTrainCfg{annoset}/".
            "$posTrainCfg{hogId}_$posTrainCfg{windowset}_$impl/$trainRegime";
        my $classifier = ichomp(`pwd`)."/$classifierName";

        # Patch up test
        my %cfg = mergehash(%posTrainCfg,%testCfg);
        $cfg{outerpaddingx} = $cfg{outerpaddingy} = 0;

        next if (!defined($cfg{video})); # Only run the tracker on videos

        # firstPass and hard have high false positive rates and take
        # much longer to process.  For now, we'll stick with only the
        # highly-tuned version of the classifier.  Temporary
        # modifications of this script can be made to generate any
        # tracker results needed to study the effects of having a
        # better classifier.
        next unless ($trainRegime eq "persp"); 

        # Data sources
        my $video     = $cfg{video};
        my $modelFile = "$classifier/model.alt";
        my $maskfile;
        eval {
            $maskfile = maskFileFromCalibTxt($cfg{calibfile});
        }; if ($@) {
            next; # skip videos with missing mask files for now
        }

        # Additional output locations
        my $testName = "$cfg{imageset}";
        my $testDir  = "$classifier/$testName";
        my $trkfile  = "$testDir/tracker.trk";
        my $filtered = "$testDir/tracker.filtered.trk";
        
        bigheading("KALMAN TRACKER EXPERIMENT USING $classifierName ON $video");

        `mkdir -p $testDir`; # do it now
        
        writeRule("Detect pedestrians and run a Kalman tracker on the results",
                  "$trkfile",
                  "$video $maskfile $modelFile",
                  "classify_rhog ".
                  "  --infile    $video ".
                  "  --modelfile $modelFile ".
                  "  --maskfile  $maskfile ".
                  "  --trkfile   $trkfile ".
                  "  ".oTrk($impl,%cfg), 
                  "$trkfile.log", $impl, $hTargets);

        writeRule("Tracker post-processing",
                  "$filtered",
                  "$trkfile",
                  "/scratch/$ENV{USER}/tracker/trunk/applets/trk2trk.opt ".
                  "  outfile=$filtered ".
                  "  outFormat=Josh ".
                  "  frameHeight=$cfg{imageheight} ".
                  "  frameWidth=$cfg{imagewidth} ".
                  "  infile=$trkfile ".
                  "  ".oTrkFiltering($impl,%cfg),
                  "$filtered.log",$impl,$hTargets);

        foreach my $small (0,1) {
            my $rendered = $small ? 
                "$testDir/tracker.filtered.sd.avi" :
                "$testDir/tracker.filtered.hd.avi";
            my $W = $small ? 640 : $cfg{imagewidth};
            my $H = ceil($W * $cfg{imageheight}/$cfg{imagewidth} / 16) * 16;
            writeRule("Track rendering",
                      "$rendered",
                      "$filtered",
                      "/scratch/$ENV{USER}/tracker/trunk/applets/trk2vid.opt ".
                      "  trkfile=$filtered ".
                      "  inVid=$video ".
                      "  width=$W ".
                      "  height=$H ".
                      "  outclass=FFMPEGSink ".
                      "  outsetup='".
                      (  )."filename=$rendered".
                      (  )." width=$W".
                      (  )." height=$H".
                      (  )." codec=mpeg4".
                      (  )." bitrate=".sprintf("%d",4e6).
                      (  )."' ".
                      "  bboxThickness=0".
                      "  trajThickness=1".
                      "  markerRadius=3".
                      "  srcMarkers=false".
                      "  sinkMarkers=false",
                      "$rendered.log",$impl,$hTargets);
        }

        # Matlab commands to run the activityModeling experiments.
        foreach my $cellSize (2.5, 5,10) { #(2.5,5,10,20) {
            foreach my $nOri (8,16) { #(4,8,16) {
                next if ($nOri!=16 && $cellSize!=5);
                foreach my $stationarySpeedMph (0.01) { #(0, 0.01) {
                    my $stationarySpeed = $stationarySpeedMph * 5280/3600; # feet/sec
                    my $datassbase = sprintf("%s/datass-s%06.4f-c%04.1f-o%02d",
                                             $testDir, $stationarySpeed, $cellSize, $nOri);
                    mkdir($datassbase);
                    my $datassfile   = "$datassbase/datass.mat";
                    my $marginalfile = "$datassbase/datass.jpg";
                    # Quantize tracks
                    writeRule("Quantize tracks for activity modeling experiments",
                              "$datassfile $marginalfile",
                              "$filtered",
                              matlabWrap($datassfile,
                                         "genPdatass(".(
                                             "'$filtered',".
                                             "'$cfg{calibfile}',".
                                             "'$datassbase/datass',".
                                             "'cellsize',$cellSize,".
                                             "'nori',$nOri,".
                                             "'stationaryspeedmph',$stationarySpeedMph)")),
                              "$datassfile.log", $impl, $hTargets);

                    # Run the HDP
                    foreach my $dirichletPrior (0.001, 0.01, 0.1, 1) { 
                        foreach my $trackFrac (0.1) { #(0.1, 1) {
                            my $hdpdir = sprintf('%s/h%07.4f-t%4.2f',
                                                 $datassbase, $dirichletPrior, $trackFrac);
                            mkdir($hdpdir);
                            my $hdpfile = "$hdpdir/model.mat";
                            writeRule("Run the activity modeling experiments",
                                      "$hdpfile",
                                      "$datassfile",
                                      matlabWrap($hdpfile,
                                                 "runHdp(".(
                                                     "'$hdpdir',".
                                                     "'$datassfile',".
                                                     "'dirichletprior',$dirichletPrior,".
                                                     "'trackfrac',$trackFrac)")),
                                      "$hdpfile.log", $impl, $hTargets);
                        }
                    }
                }
            }
        }
    }
}

sub matlabWrap($$) {
    my ($outfile, $whatToDo) = @_;
    my $matlab = "/afs/csail.mit.edu/common/matlab/2009a/bin/matlab -nodisplay";
    return ("touch $outfile; ". # lets us use multiple machines almost-safely
            "  $matlab -r \"".(
                "cd ../matlab; tic; ".
                "try; ".
                "$whatToDo; ".
                "catch; ".
                "lasterror, ".
                "end; toc; ".
                "exit;")."\"");
}

sub writeDetOptimizerExpt($$$$$$) {
    my ($impl,$hPosCfg,$hCfgs,$trainRegime,$testRegime,$hTargets) = @_;
    my %posCfg = %{$hPosCfg};
    my @testCfgs   = @{$hCfgs};

    my $classifierName = "$posCfg{imageset}_$posCfg{annoset}/".
        "$posCfg{hogId}_$posCfg{windowset}_$impl/$trainRegime";
    my $classifier = ichomp(`pwd`)."/$classifierName";

    bigheading("DETECTOR OPTIMIZATION EXPERIMENT USING $classifierName");

    # Run classify_rhog with permissive settings for each config
    for my $hTestCfg (@testCfgs) {
        my %testCfg = %{$hTestCfg};

        if ($testCfg{isLst}) {
            ($posCfg{hogId} eq $testCfg{hogId}) || 
                die("The feature geometry must be the same for the training ".
                    "($posCfg{pathname}) and test ($posCfg{pathname}) ".
                    "sets.");
        }

        # Patch up test
        my %cfg = mergehash(%posCfg,%testCfg);
        $cfg{outerpaddingx} = $cfg{outerpaddingy} = 0;

        # Data sources
        my $modelFile = "$classifier/model.alt";
        my $indir     = "$cfg{dir}/full";

        # Additional output locations
        my $testName = "$cfg{imageset}";
        my $testDir  = "$classifier/$testName";
        my $outfile  = "$testDir/permissive_list.txt";
        my $outhist  = "$testDir/permissive_hist.txt";
        die "programmer error" unless ($outfile eq permissiveLst($classifier, %testCfg));

        my $svmThresh = getMinWithDefault('svmthreshold', -1.5, %cfg);

        `mkdir -p $testDir`;

        writeRule("Detect pedestrians with permissive settings",
                  "$outfile $outhist",
                  "$modelFile",
                  "classify_rhog ".
                  "  --infile    $indir ".
                  "  --outfile   $outfile ".
                  "  --modelfile $modelFile ".
                  "  --outhist   $outhist ".
                  "  --txtimgnames=1 ".
                  "  ".oTest($impl,%cfg,'nononmax',1,'svmthreshold',$svmThresh),
                  "$outfile.log", $impl, $hTargets);
    }

    # Now run detoptimizer on the .txt files
    my $outdir = "$classifier/$testRegime";
    my $detopt = "$outdir/detopt.csv";
    my @txtfiles  = map { permissiveLst($classifier,%{$_}) } @testCfgs;
    my @histfiles = map { my $t=$_; $t=~s/_list/_hist/; $t } @txtfiles;
    my @maskfiles = map { findBestMaskFile(%{$_})          } @testCfgs;
    my @gtdirs    = map { findBestGroundTruthDir(%{$_})    } @testCfgs;

    `mkdir -p $outdir`;
    
    writeRule("Compute precision-recall statistics",
              $detopt,
              join(' ', @txtfiles, @histfiles, @maskfiles, 
                   map { "$_/calib.txt" } @gtdirs),
              "detoptimizer ".
              "  --outdir=$outdir ".
              "  --csvout=true ".
              "  --postxtout=false ". 
              "  --negtxtout=false ".
              "  --falsepostxtout=false ".
              "  --falsenegtxtout=false ".
              "  --maskgt=1 ".
              "  --minnonmaxsigma=4,8,100 ".
              "  --maxnonmaxsigma=16,32,100 ".
              "  --nnonmaxsigma=4,4,1 ".
              "  --maxdatafrac=0.05 ".
              "  --minthresh=-5.0 ".
              "  --maxthresh=1.0 ".
              "  --nthresh=600 ".
              #"  --imageout=true ". 
              join('', map { "  --txtfile=$_ "  } @txtfiles)." ".
              join('', map { "  --histfile=$_ " } @histfiles)." ".
              join('', map { "  --gtdir=$_ "    } @gtdirs)." ".
              join('', map { "  --maskfile=$_ " } @maskfiles)." ",
              "$detopt.log", $impl, $hTargets);  
}

###############################################################################
###############################################################################
####  Utilities for writing makefile rules
###############################################################################
###############################################################################

sub flockWrap($$) {
    my ($cmd, $lockfile) = @_;
    if (length($lockfile) == 0) {
        return $cmd;
    } else {
        return "(flock 200 ; $cmd) 200>$lockfile";
    }
}

sub writeRule($$$$$$) {
    my ($comment,$targets,$prereqs,$cmd,$logfile,$impl,$hTargets) = @_;
    my ($target,@extraTargets) = split(' ',$targets);
    my ($prog,@args) = split(' ', $cmd);
    my @cudaEnabledProgs = ("dump_rhog", "classify_rhog");
    my @nonCudaProgs = ("dump4svmlearn", "svm_learn", "detoptimizer",
                        "ffmpeg", "render_rhog");

    $cmd =~ s/[ ]+/ /g;

    my $lockfile = "";
    if (($impl =~ /gpu/i) && (grep { /$prog/ } @cudaEnabledProgs)) {
        $lockfile = ".gpu-lock";
    }
    
    heading($comment);
    print("$target: $prereqs\n");
    if (length($logfile)) {
        my $cmdEscaped = $cmd; $cmdEscaped=~s/'/'\\''/g;
        my $timeWrappedCmd = "time nice -n +10 $cmd";
        print("\t\@date\n");
        print("\t\@echo \`hostname\`  >  $logfile\n");
        print("\t\@echo \`date\`      >> $logfile\n");
        print("\t\@echo '$cmdEscaped' >> $logfile\n");
        print("\t\@echo \`date\`      >> $logfile\n");
        print("\t\@echo               >> $logfile\n");
        print("\t((".flockWrap($timeWrappedCmd,
                               $lockfile).") 2>&1) >> $logfile\n");
    } else {
        print("\t".flockWrap($cmd,$lockfile)."\n\n");
    }
    print(join(' ', @extraTargets).": $target\n") if (@extraTargets);
    print "\n";

    push @$hTargets, $target,@extraTargets;
}

sub genheading($$$) {
    my ($str,$indent,$char) = @_;
    print("#".($char x $indent)." $str ".
          ($char x (100-length($str)-2-$indent-1))."\n");
}

sub hugeheading($) {
    my ($str) = @_;
    print(("#"x100)."\n");
    genheading($str,5,'#');
    print(("#"x100)."\n\n");
}

sub bigheading($) {
    my ($str) = @_;
    genheading($str,10,'=');
    print "\n";
}

sub heading($) {
    my ($str) = @_;
    genheading($str,15,'-');
    print "\n";
}

###############################################################################
###############################################################################
####  General utility functions
###############################################################################
###############################################################################
sub removeLast(@) {
    pop(@_);
    return @_;
}

sub removePrefix($$) {
    my ($prefix, $str) = @_;
    (substr($str,0,length($prefix)) eq $prefix) || 
        die "Missing prefix '$prefix' in '$str'.\n";
    return substr($str,length($prefix));
}

sub removeSuffix($$) {
    my ($suffix, $str) = @_;
    my $lenRemaining = length($str)-length($suffix);
    (substr($str,$lenRemaining) eq $suffix) || 
        die "Missing suffix '$suffix' in '$str'.\n";
    return substr($str,0,$lenRemaining);
}

sub ichomp($) {
    chomp $_[0];
    return $_[0];
}

# folds fields and values from the second arg into the first
sub mergehash(%%) {
    my %h = @_;
    return %h;
}

sub getWithDefault($$%) {
    my ($name,$default,%cfg) = @_;
    if (defined $cfg{$name}) {
        return ((length($cfg{$name})==0) ? $default : $cfg{$name});
    } else {
        return $default;
    }
}

sub getMinWithDefault($$%) {
    my ($name,$defaultOrMin,%cfg) = @_;
    if (defined $cfg{$name}) {
        if (length($cfg{$name})==0) {
            return $defaultOrMin;
        } else {
            return min($defaultOrMin, $cfg{$name});
        }
    } else {
        return $defaultOrMin;
    }
}

sub gcd(@) {
    my $a = shift @_;
    while (@_) {
        $a = gcdEuclid($a, shift @_);
    }
    return $a; 
}

# http://www.itl.nist.gov/div897/sqg/dads/HTML/euclidalgo.html
sub gcdEuclid($$) {
    my ($a, $b) = @_;
    if ($b==0) {
        return $a; 
    } else {
        return gcdEuclid($b, $a % $b);
    }
}

sub checkFieldExists($$%) {
    my ($lst, $fn, %cfg) = @_;
    if (!defined $cfg{$fn}) {
        die "Field $fn does not exist in the configuration information for ".
            "'$lst'.";
    }
}

sub printCfg(%) {
    my (%cfg) = @_;
    my $longest = 0;
    foreach my $key (keys %cfg) {
        $longest = max($longest, length($key));
    }
    foreach my $key (sort keys %cfg) {
        print STDERR "  ".lpad($key,$longest).": $cfg{$key}\n";
    }
}

sub lpad($$) {
    my ($s,$l) = @_;
    if (length($s) < $l) {
        return (' 'x($l-length($s))).$s;
    } else {
        return $s;
    }
}

sub min(@) {
    my $v = shift @_;
    while (my $t = shift @_) {
        $v = $t if ($t < $v);
    }
    return $v;
}

sub max(@) {
    my $v = shift @_;
    while (my $t = shift @_) {
        $v = $t if ($t > $v);
    }
    return $v;
}

sub sq($) {
    return $_[0]*$_[0];
}

