/*
 * ============================================================================
 *
 *       Filename:  classify_rhog.cpp
 *
 *    Description:  Provides functionality to detect objects which are
 *    described by dense RHOG descriptors and learned via linear SVMs.
 *
 *    Can optionally perform non-maximum suppression 
 *
 *        Version:  1.0
 *        Created:  Monday 11 June 2007 03:41:37  IST IST
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Navneet Dalal
 *        Company:  Lear / INRIA.
 *
 * ============================================================================
 */

#include <memory>
#include "windetectmain.h"
#include <lear/image/inimageiter.h>

using namespace std;
using namespace lear;

int main(int argc, char** argv) {
  using namespace std;

  // ------------------------------------------------------------------------
  // Parse command line

  WinDetectClassifyMain windetectmain;
  RHOGDenseMain rhogdensemain;

  WinDetectClassify windetect;
  {  // cmdline
    lear::Cmdline cmdline;

    cmdline.commandName("classify_rhog");
    cmdline.version("0.0.1", "Author: Navneet Dalal "
                    "(mailto: Navneet.Dalal@inrialpes.fr)");
    cmdline.brief(
      "Compute HOG features and classify objects in image(s)");

    cmdline.description(
      "For all windows at all scales in the image, compute HOG features and "
      "classify each window as positive or negative class");

    // set command line options. The statement below will fill up usage, 
    // and code for reading various command line options
    windetectmain.setCommonMainParam(cmdline, &windetect);
    windetectmain.setClassifyParam(cmdline,&windetect);
    rhogdensemain.setRHOGDenseParam(cmdline);

    int status = cmdline.parse(argc, argv);
    if (status != cmdline.ok) return status;
  }

  if (!windetectmain.imageext.empty() && windetectmain.imageext[0] != '.')
    windetectmain.imageext = '.' + windetectmain.imageext;

  // now set options read from command line 
  windetectmain.fill(&windetect);

  std::vector<RHOGDenseParam const *> desc = rhogdensemain.fill();
  windetect.init(desc); 
  for (unsigned i= 0; i< desc.size(); ++i) delete desc[i];

  // ------------------------------------------------------------------------
  // Do the detection

#ifndef _DEBUG
  try {
#endif
    // The classifier uses a linear decision boundary, typically one learned
    // by an SVM.  This is a boundary on the HOG feature vector.
    auto_ptr<lear::LinearClassify> classifier;
    if (windetectmain.modelfile == "defaultperson") {
      classifier.reset(new lear::LinearClassify());
    } else {
      classifier.reset(new lear::LinearClassify(windetectmain.modelfile, 
                                                windetect.verbose));
    }

    // Find images on which we'll run the detector.
    auto_ptr<InputImageIterator> 
      infiles(createInputImageIterator(windetectmain.infile, 
                                       windetectmain.imageext));

    // Does the actual work
    windetect.runImageSlider(*classifier, *infiles, 
                             windetectmain.maskfile,
                             windetectmain.outfile, 
                             windetectmain.writeListFilename,
                             windetectmain.outimage, windetectmain.outhist, 
                             windetectmain.falsetxt, windetectmain.testlocs);
#ifndef _DEBUG
  } catch(lear::Exception& e) {
    cerr << "Caught "<< e.what() << endl;
    return 1;
  } catch(std::exception& e) {
    cerr << "Caught "<< (e.what() ? e.what() : "UNKNOWN ERROR") << endl;
    return 1;
  } catch(...) {
    cerr << "Caught unknown exception" << endl;
    return 1;
  }
#endif
  return 0;
}
