#include <stdio.h>
#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <set>
#include <math.h>

#include "lib/cfg.h"

#include <boost/shared_ptr.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/filesystem/operations.hpp>

#include <lear/util/util.h>
#include <lear/util/fileutil.h>
#include <lear/image/imageio.h>
#include <lear/image/inimageiter.h>
#include <lear/image/rescale.h>

#include <lear/cmdline.h>
#include <lear/vectoropt.h>
#include <lear/tinyvecopt.h>
#include <lear/util/customoption.h>

#include <lear/classifier/resultholder.h>
#include <lear/classifier/markimage.h>
#include <lear/classifier/ms_processresult.h>
#include <lear/classifier/hard_ppresult.h>
#include <lear/classifier/hist_processresult.h>

#include <lear/numericutil/rng.h>

#include <lear/cvision/perspective.h>

#include <lear/cvision/iprocessor.h>
#include <lear/interface/windetect.h>

#ifdef WIN32
#define snprintf _snprintf
#define isnan    _isnan
#endif

namespace fs=boost::filesystem;
using boost::shared_ptr;
using lear::numericutil::RandomGenerator;
using namespace std;
using namespace lear;

typedef WinDetect::RealType            RealType;
typedef IProcessor::GrayImage          GrayImage;   // Array<float,2>
typedef blitz::TinyVector<RealType,3>  NonmaxType;
typedef blitz::TinyVector<RealType,2>  SigmaType;
typedef std::list<DetectInfo>          DetList;
typedef blitz::TinyVector<int,2>       IndexType;

struct InData {
  string txtfile;  // input image names & detection results
  string gtdir;    // calibration data & ground truth points
  string maskfile; // mask
  string histfile; // histogram of all (prefiltered) detection scores
};

static float const maskThreshold = 10.f;

// Configuration info for the histogram files (since this data is implicit)
static float  const minThHist = lear::Hist_ProcessResult::DefaultMinTh;
static float  const extThHist = lear::Hist_ProcessResult::DefaultExtTh;
static float  const maxThHist = minThHist + maxThHist;
static float  const bwThHist  = lear::Hist_ProcessResult::DefaultBandwidth;
static size_t const nThHist   = size_t(ceil(extThHist / bwThHist));

///////////////////////////////////////////////////////////////////////////////
// Global vars initialized by parseCmdline

int verbose;

vector<InData> inData;
string         outdir;
bool csvout, imageout, postxtout, negtxtout, falsepostxtout, falsenegtxtout;

bool noparamsubdirs;

blitz::TinyVector<int,2> winsize;
int                      borderpixels;
float                    pedestrianHeight;

float maxdatafrac;
float minthresh, maxthresh; 
int nthresh;
blitz::TinyVector<float,3> minnonmaxsigma, maxnonmaxsigma; 
blitz::TinyVector<int,3> nnonmaxsigma;

float minOverlapFrac;

bool resizemask, maskgt;

size_t negsamples, maxretries;
float  posscore, negscore;
float  startscale, endscale, scaleprecision;

int borderThickness;

#ifdef HAVE_OPENMP
int nThreads;
#endif

///////////////////////////////////////////////////////////////////////////////
// Command line parsing

void parseCmdline(int argc, char **argv) {
  typedef lear::TinyVecOpt<int,2>         IndexOpt;
  typedef lear::VectorOpt<std::string>    MultStrOpt;
  typedef lear::TinyVecOpt<RealType,3>    NonmaxOptType;
  typedef lear::TinyVecOpt<int,3>         NonmaxCountOptType;

  MultStrOpt inTxtFiles_;
  MultStrOpt gtDirs_; // must contain calib.txt and headtips/
  MultStrOpt maskFiles_;
  MultStrOpt histFiles_;

  IndexOpt winsize_;

  NonmaxOptType minnonmaxsigma_, maxnonmaxsigma_;
  NonmaxCountOptType nnonmaxsigma_;

  IndexOpt::seperator(",x");

  lear::Cmdline cmdline;
  cmdline.commandName("detoptimizer");
  cmdline.version("0.0.2", "Author: Gerald Dalley "
                  "(mailto: dalleyg@ieee.org)");
  cmdline.brief("Assists the user in optimizing the final detector parameters");
  cmdline.description("Reads a list.txt file "
                      "(from \"classify_rhog --outfile [TXTFILE]\") and "
                      "performs thresholding and non-maxima suppression, "
                      "exhaustively searching over a range of nonmaxsigma and "
                      "detector thresholds.  The results are written to a "
                      ".csv file.  Typically, the input file should have been "
                      "produced without any non-maxima suppression and with a "
                      "threshold less than or equal to --minthresh.");

  cmdline.addOption()
    ("verbose,v",option<int>(&verbose)
     ->defaultValue(7)->minValue(0)->maxValue(9),
     "Verbosity level")
    ("txtfile", option<MultStrOpt>(&inTxtFiles_), 
     "Detector results file (.txt) from classify_rhog.  classify_rhog must "
     "have used --txtimgnames=1 and it should have --no_nonmax=0.  Multiple "
     "input files are allowed, as long as there are the same number of gtdir "
     "and maskfile entries.")
    ("gtdir,g",option<MultStrOpt>(&gtDirs_),
     "Directory containing ground truth data.  It must have a calib.txt file "
     "(see test/input.txt for an example), and it must have a subdirectory "
     "called headtips/ that contains black and white images.  White dots "
     "correspond to the locations of head tips.  The image filenames must "
     "correspond to the filenames found in the .txt file specified by "
     "--txtfile.") 
    ("maskfile,m",option<MultStrOpt>(&maskFiles_),
     "Detections that land on black pixels in this image are ignored") 
    ("histfile,h",option<MultStrOpt>(&histFiles_),
     "Score histogram file (produced by classify_rhog at the same time as "
     "the corresponding --txtfile.  This is required when --maxdatafrac is "
     "not 1.")

    ("outdir", option<string>(&outdir)->defaultValue("detopt"),
     "Base directory for all output that is written to files.  Actual "
     "results will be written in this directory and subdirectories.")
    ("csvout", bool_option(&csvout)->defaultValue(false),
     "If true, statistics from the detection tests will be written to "
     "[outdir]/detopt.csv.")
    ("imageout", bool_option(&imageout)->defaultValue(false),
     "If true, a rendering of the detector and ground truth windows for each "
     "test image will be written displaying true positives, false positives, "
     "and false negatives.  The results will be written to subdirectories of "
     "[outdir].")
    ("postxtout", bool_option(&postxtout)->defaultValue(false),
     "If true, image file + detector window information for all ground truth "
     "object locations will be written to a file suitable for usage with "
     "dump_rhog.  This file is primarily for adding training data to help "
     "improve the detection rate.  The file will be [outdir]/pos.txt.")
    ("negtxtout", bool_option(&negtxtout)->defaultValue(false),
     "If true, image file + detector window information for randomly-sampled "
     "windows will be written to [outdir]/neg.txt.  Sampled "
     "windows that match any ground truth objects are omitted (see "
     "--minoverlapfrac).")
    ("falsepostxtout", bool_option(&falsepostxtout)->defaultValue(false),
     "If true, image file + detector window information for false positive "
     "detections will be written to \"false_pos.txt\" files in "
     "subdirectories of [outdir].")
    ("falsenegtxtout", bool_option(&falsenegtxtout)->defaultValue(false),
     "If true, image file + detector window information for false negative "
     "detections will be written to \"false_neg.txt\" files in "
     "subdirectories of [outdir].")
    ("noparamsubdirs", bool_option(&noparamsubdirs)->defaultValue(false),
     "Places output files directly in [outdir] rather than in subdirectories "
     "corresponding to different parameter settings.  This option may only "
     "be enabled when [nnonmaxsigma] and [nthresh] are both 1.")

    ("winsize", option<IndexOpt>(&winsize_)
     ->defaultValue(IndexOpt(64,128))->minValue(IndexOpt(1,1)),
     "Size of the detector window, in pixels")
    ("borderpixels", option<int>(&borderpixels)
     ->defaultValue(16)->minValue(1),
     "The pixel height of a canonical person is winsize[1]-2*borderpixels")
    ("pheight,p", option<RealType>(&pedestrianHeight)
     ->defaultValue((5 + 9.f/12))->minValue(0),
     "Height of a standard person in whatever units are used in the calib.txt "
     "files")

    ("maxdatafrac",option(&maxdatafrac)
     ->defaultValue(1)->minValue(0),
     "Maximum fraction of the input data that will be considered.  When "
     "discarding a fraction of the input data, the detections with the "
     "lowest scores are thrown out.  Use this option to bypass the "
     "processing of small thresholds (because they are time consuming).")

    ("minthresh",option<float>(&minthresh)
     ->defaultValue(-1.5f),
     "Minimum detector threshold to consider in the optimization")
    ("maxthresh",option<float>(&maxthresh)
     ->defaultValue(0.2f),
     "Maximum detector threshold to consider in the optimization")
    ("nthresh",option<int>(&nthresh)
     ->defaultValue(18)->minValue(1),
     "Number of evenly-spaced detector thresholds to test between the min and "
     "max thresholds")

    ("minnonmaxsigma", option<NonmaxOptType>(&minnonmaxsigma_)
     ->defaultValue(NonmaxOptType(2,4,100))->minValue(NonmaxOptType(0,0,1)),
     "Minimum value for smoothing sigma used for non-maxima suppression "
     "(x,y,scale)")
    ("maxnonmaxsigma", option<NonmaxOptType>(&maxnonmaxsigma_)
     ->defaultValue(NonmaxOptType(16,32,100))->minValue(NonmaxOptType(0,0,1)),
     "Maximum value for smoothing sigma used for non-maxima suppression "
     "(x,y,scale)")
    ("nnonmaxsigma", option<NonmaxCountOptType>(&nnonmaxsigma_)
     ->defaultValue(NonmaxCountOptType(8,8,1))
     ->minValue(NonmaxCountOptType(0,0,1)),
     "Number of evenly-spaced smoothing sigma values to test between the min "
     "and max nonmaxsigma values.")

    ("minoverlapfrac,f", option<float>(&minOverlapFrac)
     ->defaultValue(0.5f)->minValue(0.0f)->maxValue(1.0f),
     "A ground truth and a test bounding box are considered to be matches if "
     "the ratio of the area of intersection of their bounding boxes over the "
     "area of the bounding box of the convex hull of their bounding boxes is "
     "greater than this fraction, as defined by the PASCAL Visual Object "
     "Challenge (http://pascallin.ecs.soton.ac.uk/challenges/VOC/voc2008/).")

    ("samples",option<size_t>(&negsamples)
     ->defaultValue(0)->minValue(0),
     "Number of randomly sampled window locations to write (see "
     "--negtxtout).")
    ("retries",option<size_t>(&maxretries)
     ->defaultValue(1000)->minValue(0),
     "Maximum number of failed sampling attempts allowed when using "
     "--negtxtout.  This option is meant to help avoid very long runtimes "
     "for images that a filled with ground truth objects and where it is "
     "hard to find a non-matching window.")
    ("posscore",option<float>(&posscore)
     ->defaultValue(-10),
     "Score to use when outputting ground truth positive windows using "
     "--postxtout.  Generally you will want to pick a value that is smaller "
     "than the --hardscore you'll be using when using dump_rhog.")
    ("negscore",option<float>(&negscore)
     ->defaultValue(10),
     "Score to use when outputting sampled negative windows using "
     "--negtxtout and/or --falsenegtxtout.  Generally you will want to "
     "pick a value that is larger than the --hardscore value you'll be "
     "using when using dump_rhog.")
    ("startscale",option<float>(&startscale)
     ->defaultValue(1)->minValue(1/32768.0f),
     "Minimum scaling of a canonical person to use when sampling negative "
     "windows using --negtxtout and --samples.  Smaller values result in "
     "image windows that look at a larger pixel area.")
    ("endscale",option<float>(&endscale)
     ->defaultValue(1)->minValue(1/32768.0f),
     "Maximum scaling of a canonical person (see --startscale).  Must be "
     "at least as large as the startscale.")
    ("scaleprecision", option<RealType>(&scaleprecision)
     ->defaultValue(0.1f)->minValue(0),
     "Precision to use when quantizing window scales in the file created "
     "when --negtxtout is enabled.  Coarser (larger) precision values "
     "result in more accurate window sizes, but can force more image "
     "resizes on programs like a CUDA-enabled dump_rhog that need to "
     "extract features from those windows.  If a scale of 1.2452 is "
     "computed and scaleprecision=0.01, then a scale of 1.25 will "
     "actually be used.")

    ("resizemask", bool_option(&resizemask)->defaultValue(true),
     "Indicates whether to resize the mask image to match the headtip images, "
     "when necessary.  This is required sometimes when aspect ratio changes "
     "are needed, but the mask did not have the aspect ratio corrected.  If "
     "this option is disabled, different sizes for a headtips image and its "
     "corresponding mask image are considered to be errors.")
    ("maskgt", bool_option(&maskgt)->defaultValue(true),
     "Use the mask to filter out ground truth examples as well as detections.")

#ifdef HAVE_OPENMP
    ("threads", option<int>(&nThreads)
     ->defaultValue(0)->minValue(0),
     "Maximum number of OpenMP threads that the optimizer may use.  "
     "0=use default (typically the number of CPU cores). When the "
     "number of threads is not 1, the ordering of the output will "
     "be unpredictable.")
#endif

    ("borderthickness", option<int>(&borderThickness)
     ->defaultValue(2)->minValue(0),
     "Thickness of rendered detection borders (ignored if imageout "
     "is false).")
    ;

  int status = cmdline.parse(argc, argv);
  if (status != cmdline.ok) exit(status);


  if (inTxtFiles_.size() == 0) {
    ThrowException("At least one input .txt file must be specified.");
  }
  if (inTxtFiles_.size() != gtDirs_.size()) {
    ThrowException("There must be the same number of input .txt files as "
                   "ground truth data directories.");
  }
  if (maskFiles_.size() && (inTxtFiles_.size() != maskFiles_.size())) {
    ThrowException("There must be the same number of input .txt files as "
                   "mask files, if any mask files are specified.");
  }
  if (maxdatafrac<1) {
    if (inTxtFiles_.size() != histFiles_.size()) {
      ThrowException("When maxdatafrac>1, there must be the same number of "
                     "input .txt files as histogram files.");
    }
  }

  inData.resize(inTxtFiles_.size());
  for (size_t i=0; i<inTxtFiles_.size(); i++) {
    inData[i].txtfile  = inTxtFiles_.at(i);
    inData[i].gtdir    = gtDirs_.at(i);
    inData[i].maskfile = maskFiles_.at(i);
    if (maxdatafrac<1) {
      inData[i].histfile = histFiles_.at(i);
    }
  }

  for (int i=0; i<2; i++) {
    winsize[i] = winsize_[i];
  }

  for (int i=0; i<3; i++) {
    minnonmaxsigma[i] = minnonmaxsigma_[i];
    maxnonmaxsigma[i] = maxnonmaxsigma_[i];
    nnonmaxsigma[i]   = nnonmaxsigma_[i];
  }    
}

///////////////////////////////////////////////////////////////////////////////

#ifdef __linux__
#include <time.h>
static int64_t nsTime() {
  struct timespec tp;
  clock_gettime(CLOCK_REALTIME, &tp);
  return tp.tv_sec*(int64_t)1000000000 + tp.tv_nsec;
}
#else
typedef __int64 int64_t;
static int64_t nsTime() {
  static bool haveFreq = false;
  __int64 freq;
  if (!haveFreq) {
    QueryPerformanceFrequency((LARGE_INTEGER*)&freq);
  }
  __int64 time;
  QueryPerformanceCounter((LARGE_INTEGER*)&time);
  return int64_t(double(time)*1e9/freq);
}
#endif

///////////////////////////////////////////////////////////////////////////////

void readGtData(vector<shared_ptr<vector<DetectedRegion> > > &groundTruthDetectionsUnmasked,
                vector<shared_ptr<vector<DetectedRegion> > > &groundTruthDetections,
                vector<shared_ptr<PerspectiveProjector> >    &projs,
                vector<string>                               &infiles,
                map<string, size_t> &idxGivenName,
                InData const &indata, GrayImage *mask)
{
  if (verbose >= 3) {
    printf("Reading ground truth data from '%s'...\n", indata.gtdir.c_str());
  }
  if (verbose >= 7) {
    printf("  mask size: %dx%d\n", mask->extent()[0], mask->extent()[1]);
  }

  // Get a sorted directory listing
  string const headtipsDir(indata.gtdir + "/headtips/");
  set<string> headtipFiles;
  namespace fs=boost::filesystem;
  fs::directory_iterator end;
  for (fs::directory_iterator htiter(headtipsDir); htiter!=end; ++htiter) {
#if BOOST_VERSION <= 103301
    string const headtipFile = htiter->native_file_string();
#else
    string const headtipFile = htiter->path().native_file_string();
#endif
    headtipFiles.insert(headtipFile);
  }

  // Read each headtip file to get the ground truth pedestrian
  // locations.
  for (set<string>::const_iterator htiter=headtipFiles.begin();
       htiter!=headtipFiles.end(); ++htiter)
  {
    string const &headtipFile = *htiter;

    string const fname  = lear::basename(headtipFile, "");
    if (verbose >= 4) {
      printf("  extracting ground truth data for '%s' ('%s')\n", 
             headtipFile.c_str(), fname.c_str());
    }

    // Get the head tips
    GrayImage headTips;
    try {
      lear::ImageIO::read(headtipFile, headTips);
    } catch (lear::Exception const &) {
      // silently skip non-image files
      continue;
    }
    if (verbose >= 7) {
      printf("    headTips size: %dx%d\n", 
             headTips.extent()[0], headTips.extent()[1]);
    }

    if (resizemask && blitz::sum(mask->extent() != headTips.extent())) {
      // This is a bit of a hack at the moment: it basically assumes
      // that all of the headTip images (in a directory) are the same
      // size.  That's true for the current datasets, but in general
      // it might not be true.  When false, the big issue comes in
      // when the modified mask is used by readDetections.
      GrayImage tmp;
      tmp.resize(headTips.extent());
      tmp = lear::rescale(*mask, headTips.extent());
      mask->reference(tmp);
    }

    // Turn them into DetectedRegion instances
    shared_ptr<vector<DetectedRegion> > gtu(new vector<DetectedRegion>());
    shared_ptr<vector<DetectedRegion> > gt(new vector<DetectedRegion>());
    shared_ptr<PerspectiveProjector> proj
      (new PerspectiveProjector(indata.gtdir + "/calib.txt", 
                                winsize, borderpixels, pedestrianHeight));
    proj->detectedRegionsFromHeadTips(*gtu, headTips);
    proj->detectedRegionsFromHeadTips(*gt, headTips, *mask);
    projs.push_back(proj);

    // Store the results
    idxGivenName[fname] = groundTruthDetections.size();
    groundTruthDetectionsUnmasked.push_back(gtu);
    groundTruthDetections.push_back(gt);
    infiles.push_back(indata.gtdir + "/" + fname); // for now, assumes the extention is the same

    if (verbose >= 6) {
      printf("    found non-masked %ld detections\n", gt->size());
    }
  }
  if (verbose >= 5) {
    printf("...done reading ground truth data from '%s'\n", indata.gtdir.c_str());
  }
}


void readDetections(vector<double> &hist,
                    vector<shared_ptr<vector<DetectedRegion> > > &baseDetections,
                    map<string, size_t> const &idxGivenName,
                    InData const &indata, GrayImage const *mask)
{
  // Read the histogram
  if (indata.histfile.size()) {
    std::ifstream h(indata.histfile.c_str());
    for (size_t i=0; i<nThHist; i++) {
      double v;
      h >> v;
      assert(h.good());
      hist[i] += v;
    }
    double tmp;
    h >> tmp;
    assert(h.eof());
  }

  // Start reading the detections
  if (verbose >= 3) {
    printf("Reading detections from '%s'...\n", indata.txtfile.c_str());
  }
  std::ifstream from(indata.txtfile.c_str());
  if (!from) {
    ThrowException("Could not open '"<<indata.txtfile<<"'");
  }

  // Create a placeholder container for each image
  {
    size_t maxIdx = 0;
    for (map<string,size_t>::const_iterator i=idxGivenName.begin(); 
         i!=idxGivenName.end(); ++i) {
      if (i->second > maxIdx) maxIdx = i->second;
    }
    for (size_t i=0; i<=maxIdx; i++) {
      baseDetections.push_back
        (shared_ptr<vector<DetectedRegion> >(new vector<DetectedRegion>));
    }
  }

  while (from) {
    // Read the detection (code adapted from windetect.cpp)
    std::string file;
    from >> file;    
    blitz::TinyVector<int,2> lbound, extent;
    RealType scale, score;
    string lineEnd;
    from >> lbound[0] >> lbound[1] >> extent[0] >> extent[1] >> scale >> score >> lineEnd;
    if (!from) break;
    if (file.empty()) continue;
    if (file == "DIVIDER") continue;

    // Skip any images with no ground truth
    map<string,size_t>::const_iterator const idxIter = 
      idxGivenName.find(lear::basename(file,""));
    if (idxIter == idxGivenName.end()) {
      if (verbose >= 9) {
        printf("Skipping detection in '%s': no ground truth data\n", 
               lear::basename(file,"").c_str());
      }
      continue;
    }
    if (verbose >= 8) printf("Found detection in '%s'\n", lear::basename(file,"").c_str());

    // Add the detection
    size_t const i = idxIter->second;
    if ((*mask)(lbound[0],lbound[1]) > maskThreshold) {
      baseDetections[i]->push_back(DetectedRegion(score, scale, 
                                                  lbound[0], lbound[1],
                                                  extent[0], extent[1]));
    }
  }

  if (verbose >= 5) {
    printf("...done reading detections from '%s'\n\n", indata.txtfile.c_str());
  }
}

float computeEffectiveMinThreshold(vector<double> const &hist)
{
  if (maxdatafrac >= 1) return minthresh;

  // Build a histogram of detection scores
  double n=0;
  double m=0;
  for (size_t i=0; i<hist.size(); i++) {
    n += hist[i];
    m += hist[i] * (i*bwThHist + minThHist);
  }
  if (verbose >= 5) {
    printf("mean score = %f\n", m/n);
  }

  // Find the threshold that will reject n*(1-maxdatafrac) of the detection
  double const target = n * (1-maxdatafrac);
  double       s      = 0;
  for (size_t i=0; i<hist.size(); i++) {
    s += hist[i];
    if (s >= target) {
      return float(MIN(maxthresh, MAX(minthresh, i*bwThHist+minThHist)));
    }
  }

  // Handle degenerate cases
  return maxthresh;
}

void readInputData(vector<shared_ptr<vector<DetectedRegion> > > &groundTruthDetectionsUnmasked,
                   vector<shared_ptr<vector<DetectedRegion> > > &groundTruthDetections,
                   vector<shared_ptr<vector<DetectedRegion> > > &baseDetections,
                   vector<shared_ptr<PerspectiveProjector> >    &projs,
                   vector<string>                               &infiles,
                   float                                        &effectiveMinThresh)
{
  groundTruthDetections.clear();
  baseDetections.clear();
  infiles.clear();

  vector<double> hist(nThHist, 0);

  for (size_t i=0; i<inData.size(); i++) {
    // Read the mask file
    GrayImage mask;
    lear::ImageIO::read(inData[i].maskfile, mask);

    // Grab the calibration & ground truth data so we can later filter
    // the detections.
    map<string, size_t> idxGivenName;
    readGtData(groundTruthDetectionsUnmasked, groundTruthDetections, 
               projs, infiles, idxGivenName, inData[i], &mask);

    // Read the detections, filtering out any from files with no
    // ground-truth data
    readDetections(hist, baseDetections, idxGivenName, inData[i], &mask);
  }

  effectiveMinThresh = computeEffectiveMinThreshold(hist);
  if (verbose >= 5) {
    printf("minthresh = %f, but the effective minthresh = %f\n",
           minthresh, effectiveMinThresh);
  }
}

///////////////////////////////////////////////////////////////////////////////

static size_t const NOMATCH = (size_t)-1;

size_t findBestMatch(vector<DetectedRegion> const &gt, 
                     DetectedRegion const &d,
                     vector<size_t> *matchingDetIdxFromGt = NULL)
{
  float  bestOverlapFrac = -numeric_limits<float>::infinity();
  size_t bestMatch       = NOMATCH;
  for (size_t gti=0; gti<gt.size(); gti++) {
    if ((matchingDetIdxFromGt==NULL) || 
        ((*matchingDetIdxFromGt)[gti] == NOMATCH)) {
      DetectedRegion const &g = gt[gti];
      
      float const bix1 = max(d.x,            g.x);
      float const biy1 = max(d.y,            g.y);
      float const bix2 = min(d.x + d.width , g.x + g.width );
      float const biy2 = min(d.y + d.height, g.y + g.height);
      float const intersectArea = max(0.f, bix2-bix1+1) * max(0.f, biy2-biy1+1);
      
      float const bux1 = min(d.x,            g.x);
      float const buy1 = min(d.y,            g.y);
      float const bux2 = max(d.x + d.width , g.x + g.width );
      float const buy2 = max(d.y + d.height, g.y + g.height);
      float const unionArea = max(0.f, bux2-bux1+1) * max(0.f, buy2-buy1+1);
      
      float const overlapFrac = intersectArea / unionArea;
      
      if ((intersectArea > 0) && (overlapFrac > bestOverlapFrac)) { // greedy
        bestOverlapFrac = overlapFrac;
        bestMatch       = gti;
      }
    }
  }

  if (bestOverlapFrac > minOverlapFrac) {
    return bestMatch;
  } else {
    return NOMATCH;
  }
}

// TODO: consolidate this code with testcudapipeline.cpp
//
// Uses the PASCAL Visual Object Challenge protocol to evaluate
// detector mistakes: http://pascallin.ecs.soton.ac.uk/challenges/VOC/
//
// We match ground truth objects to detections by checking the
// intersection area / union area of their bounding boxes.  
//
//   ground truth object matches a detection: true positive
//   ground truth object has no match:        false negative
//   detection has no match:                  false positive
//  
// If a ground truth object is matched multiple times, all matches
// beyond the first count as false positives.
void vocMistakes(int &tp, int &fp, int &fn, 
                 vector<size_t> &matchingDetIdxFromGt,
                 vector<DetectedRegion> const &gt, 
                 vector<DetectedRegion> const &det,
                 float minOverlapFrac)
{
  assert(minOverlapFrac > 0);
  matchingDetIdxFromGt.clear();
  matchingDetIdxFromGt.resize(gt.size(), NOMATCH);

  for (size_t deti=0; deti<det.size(); deti++) {
    DetectedRegion const &d = det[deti];
    size_t const bestMatch = findBestMatch(gt, d, &matchingDetIdxFromGt);
    if (bestMatch != NOMATCH) {
      matchingDetIdxFromGt[bestMatch] = deti;
      tp++;
    } else {
      fp++;
    }
  }

  // Account for ground truth objects that weren't matched
  for (size_t i=0; i<matchingDetIdxFromGt.size(); i++) {
    if (matchingDetIdxFromGt[i] == NOMATCH) fn++;
  }
}

void safeMkdir(string const &d) {
  fs::path p(d);
  if (fs::exists(p)) {
    if (!fs::is_directory(p)) {
      ThrowException("requested output directory "
                     "exists but is not a directory: " << d);
    }
  } else {
    fs::create_directories(d);
    if (!fs::exists(p)) {
      ThrowException("could not create output "
                     "directory: " << d);
    }
  }
}

//  [outdir]/params_[thresh]_[nmw]_[nmh]_[nms]/
string getTestOutDir(float svmThreshold, NonmaxType nonmaxSigma) {
  if (noparamsubdirs) {
    if (blitz::sum(nnonmaxsigma != 1)) {
      ThrowException("nnonmaxsigma must be 1 when noparamsubdirs is enabled");
    }
    if (nthresh != 1) {
      ThrowException("nthresh must be 1 when noparamsubdirs is enabled");
    }
    safeMkdir(outdir);
    return outdir;

  } else {
    char d[8192];  
    d[sizeof(d)-1] = '\0';
    if (snprintf(d, sizeof(d)-1, 
                 "%s/params_%07.1f_%05.2f_%05.2f_%06.2f",
                 outdir.c_str(), svmThreshold, 
                 nonmaxSigma(0),nonmaxSigma(1),nonmaxSigma(2)) < 0) {
      ThrowException("output filename is too long");
    }
    safeMkdir(d);
    return d;
  }
}

// Turns a DetectedRegion vector into a DetectInfo list.  This file
// uses a lot of vectors because we need random access.  The main lear
// library uses lists almost everywhere.
void dtlFromDrv(list<DetectInfo> &dtl, vector<DetectedRegion> const &drv) {
  dtl.clear();
  for (size_t i=0; i<drv.size(); i++) {
    DetectedRegion const &reg = drv[i];
    DetectInfo r(reg.score, reg.scale, reg.x,reg.y, reg.width,reg.height);
    dtl.push_back(r);
  }
}      

void dtlFromDrv(list<DetectInfo> &dtl, vector<DetectedRegion> const &drv,
                RealType score) {
  dtl.clear();
  for (size_t i=0; i<drv.size(); i++) {
    DetectedRegion const &reg = drv[i];
    DetectInfo r(score, reg.scale, reg.x,reg.y, reg.width,reg.height);
    dtl.push_back(r);
  }
}

void buildConfusionMatrixLists(DetList &gtTp, DetList &gtFn,
                               DetList &detTp, DetList &detFp,
                               vector<DetectedRegion> const &gt,
                               vector<DetectInfo> const &prunedDet,
                               vector<size_t> const &matchingDetIdxFromGt,
                               float posscore)
{
  vector<unsigned char> detMatched(prunedDet.size(), false);
  gtTp.clear();
  gtFn.clear();
  assert(matchingDetIdxFromGt.size() == gt.size());
  for (size_t j=0; j<matchingDetIdxFromGt.size(); j++) {
    DetectedRegion const &reg = gt[j];
    DetectInfo r(posscore, reg.scale, reg.x,reg.y, reg.width,reg.height);
    if (matchingDetIdxFromGt[j] == NOMATCH) {
      gtFn.push_back(r);
    } else {
      gtTp.push_back(r);
      detMatched[matchingDetIdxFromGt[j]] = true;
    }
  }

  detTp.clear();
  detFp.clear();
  for (size_t j=0; j<prunedDet.size(); j++) {
    if (detMatched[j]) {
      detTp.push_back(prunedDet[j]);
    } else {
      detFp.push_back(prunedDet[j]);
    }
  }
}

void writeImageResults(int tp, int fp, int fn,
                       DetList &gtTp, DetList &gtFn,
                       DetList &detTp, DetList &detFp,
                       float svmThreshold, NonmaxType nonmaxSigma, 
                       int borderThickness, 
                       int i, string const &infile) 
{
  typedef std::list<DetectInfo>::const_iterator   const_iterator;

  // Graphics config
  typedef MarkImage::Color  Color;
  typedef MarkImage::Colors Colors;
  vector<Colors> borderColors;
  borderColors.push_back(Colors(borderThickness, Color(  0,255,255))); // GT  TP (cyan  : good)
  borderColors.push_back(Colors(borderThickness, Color(192,192,  0))); // GT  FN (yellow: bad)
  borderColors.push_back(Colors(borderThickness, Color(  0,  0,255))); // DET TP (blue  : good)
  borderColors.push_back(Colors(borderThickness, Color(255,  0,  0))); // DET FP (red   : bad) 

  // Filename config
  string const testOutDir = getTestOutDir(svmThreshold, nonmaxSigma);
  char outfileBase[8192]; 
  outfileBase[sizeof(outfileBase)-1] = '\0';
  snprintf(outfileBase, sizeof(outfileBase)-1, 
           "%s/%06d", testOutDir.c_str(), int(i));

  // Save metadata
  {
    ofstream descr((outfileBase + string(".csv")).c_str());
    descr << "inputFile," << infile << endl;
    descr << "svmThreshold," << svmThreshold << endl;
    descr << "nonmaxSigma," << nonmaxSigma(0) << "," << nonmaxSigma(1) 
          << "," << nonmaxSigma(2) << endl;
    descr << "outputIndex," << i << endl;
    descr << "tp,"<<tp<<endl;
    descr << "fp,"<<fp<<endl;
    descr << "fn,"<<fn<<endl;
  }
    
  // Gather detection data
  vector<const_iterator> starts;
  vector<const_iterator> ends;

  starts.push_back(gtTp.begin());   ends.push_back(gtTp.end());
  starts.push_back(gtFn.begin());   ends.push_back(gtFn.end());
  starts.push_back(detTp.begin());  ends.push_back(detTp.end());
  starts.push_back(detFp.begin());  ends.push_back(detFp.end());

  // Actually write the outputs
  MarkImage markImage(outfileBase + string(".jpg"), false, true);
  if (!markImage.verifyImageOut(testOutDir, false)) {
    ThrowException("error constructing image marker object");
  }
  lear::IProcessor::RGBImage origimage;
  lear::ImageIO::read(infile, origimage);
  markImage.write(starts, ends, borderColors, origimage, infile);
}

static inline 
float nonan(float x) {
  return isnan(x) ? 0 : x;
}

void mistakesFromScores(int &tp, int &fp, int &fn,
                        int &mistakes, float &precision, float &recall,
                        vector<shared_ptr<vector<DetectedRegion> > > const &groundTruthDetections, 
                        vector<shared_ptr<vector<DetectedRegion> > > const &baseDetections,
                        vector<string>                               const &infiles,
                        float svmThreshold, NonmaxType nonmaxSigma, 
                        float minOverlapFrac, int borderThickness)
{
  tp = fp = fn = 0;

  static SigmaType const scoreToProb(1,1);
  static int       const whichTransferFunc(0); // 0=HardMax, 1=Sigmoid, 2=SoftMax
  float const finalThreshold = svmThreshold; // for now
  MeanShift_ProcessResult meanshift(winsize, svmThreshold, finalThreshold, 
                                    scoreToProb, nonmaxSigma, 
                                    whichTransferFunc);
  auto_ptr<Hard_PPResult> falsepostxt;
  if (falsepostxtout) {
    falsepostxt.reset(new Hard_PPResult(getTestOutDir(svmThreshold,nonmaxSigma)
                                        + "/false_pos.txt"));
  }
  auto_ptr<Hard_PPResult> falsenegtxt;
  if (falsenegtxtout) {
    falsenegtxt.reset(new Hard_PPResult(getTestOutDir(svmThreshold,nonmaxSigma)
                                        + "/false_neg.txt"));
  }

  lear::IProcessor::RGBImage dummy(0,0);
  vector<DetectedRegion> finalDetections;
  size_t const nImages = groundTruthDetections.size();
  for (size_t i=0; i<nImages; i++) {
    typedef std::list<DetectInfo>::const_iterator   const_iterator;

    // Run the detector for this image
    meanshift.clear();
    vector<DetectedRegion> const &gt  = *(groundTruthDetections[i]);
    vector<DetectedRegion> const &det = *(baseDetections[i]);
    vector<DetectInfo> prunedDet;
    for (size_t j=0; j<det.size(); j++) {
      DetectedRegion const &reg = det[j];
      if (reg.score >= svmThreshold) {
        DetectInfo r(reg.score, reg.scale, reg.x, reg.y,
                     reg.width, reg.height);
        //r.lbound -=toadd;
        meanshift(r);
      }
    }
    meanshift.doit();
    const_iterator s = meanshift.begin();
    const_iterator e = meanshift.end();
    finalDetections.clear();
    for (; s!=e; ++s) {
      DetectedRegion r(s->score, s->scale, 
                       s->lbound[0], s->lbound[1], s->extent[0], s->extent[1]);
      finalDetections.push_back(r);
      prunedDet.push_back(*s);
    }
    
    // Calculate errors
    vector<size_t> matchingDetIdxFromGt;
    int tpImg=0, fpImg=0, fnImg=0;
    vocMistakes(tpImg,fpImg,fnImg, matchingDetIdxFromGt, 
                gt, finalDetections, minOverlapFrac);
    tp+=tpImg; fp+=fpImg; fn+=fnImg;

    // Write results 
    if (imageout || falsepostxtout || falsenegtxtout) {
      DetList gtTp, gtFn, detTp, detFp;
      buildConfusionMatrixLists(gtTp,gtFn, detTp,detFp, 
                                gt, prunedDet, matchingDetIdxFromGt,
                                posscore);
      if (imageout) {
        writeImageResults(tpImg,fpImg,fnImg, gtTp,gtFn, detTp,detFp, 
                          svmThreshold,nonmaxSigma, borderThickness, 
                          i, infiles[i]);
      }
      if (falsepostxtout) {
        falsepostxt->write(detFp.begin(), detFp.end(), dummy, infiles[i]);
      }
      if (falsenegtxtout) {
        falsenegtxt->write(gtFn.begin(), gtFn.end(), dummy, infiles[i]);
      }
    }
  } // for (size_t i=0; i<nImages; i++) 

  mistakes  = fp + fn;
  precision = nonan(tp / (float)(tp+fp));
  recall    = nonan(tp / (float)(tp+fn));
}

inline float getStep(float minval, float maxval, int n) {
  return ((maxval <= minval) || (n <= 1)) ? 1 : (maxval - minval) / (n - 1);
}

// linear index to column-major subscripts.  Adapted from Matlab's
// ind2sub function.
template <typename T, int N>
inline blitz::TinyVector<T,N> ind2sub(blitz::TinyVector<T,N> sz, int ndx) {
  blitz::TinyVector<T,N> sub;
  for (int d=0; d<N; d++) {
    sub[d] = ndx % sz[d];
    ndx /= sz[d];
  }
  return sub;
}

FILE *openCsvAndWriteHeaders()
{
  if (verbose >= 1) {
    cout << "      svm       non-max sigma      true  false  false                                eval time " << endl;
    cout << "    thresh.  width height  scale   pos.   pos.   neg.   mistakes  precision recall      (ms)   " << endl;
    cout << "    -------  ----- ------ ------  ------ ------ ------  --------  --------- ------  -----------" << endl;
  }
  if (!csvout) {
    return NULL;
  } else {
    safeMkdir(outdir);
    string const csv = outdir + "/detopt.csv"; 
    FILE *f = fopen(csv.c_str(), "wt");
    if (f) {
      fprintf(f, 
              "svmThreshold, nonmaxWidth,nonmaxHeight,nonmaxScale,"
              "truePos,falsePos,falseNeg, mistakes,"
              "precision,recall, evalTime\n");
    }
    return f;
  }
}

void printTableRow(FILE *f, float svmThreshold, NonmaxType nonmaxSigma, 
                   int tp, int fp, int fn, int mistakes, 
                   float precision, float recall, double etimeMS) 
{
  // Print tabular output
  if (verbose >= 1) {
    printf("    % 7.3f  %5.2f %6.2f %6.2f  %6d %6d %6d  %8d  %8d%% %5d%% % 12.2f\n",
           svmThreshold, nonmaxSigma(0),nonmaxSigma(1),nonmaxSigma(2), 
           tp,fp,fn, mistakes, int(precision*100),int(recall*100), etimeMS);
    fflush(stdout);
  }
  if (f) {
    fprintf(f, "%f, %f,%f,%f, %d,%d,%d, %d, %f,%f, %f\n",
            svmThreshold, nonmaxSigma(0),nonmaxSigma(1),nonmaxSigma(2), 
            tp,fp,fn, mistakes, precision,recall,etimeMS);
    fflush(f);
  }
}

// Iterator object that allows us to do a better job optimize the
// workload for OpenMP.  The design is quick-and-dirty.
struct OptimizerIter {
  OptimizerIter() : 
    tExt(nnonmaxsigma[0], nnonmaxsigma[1], nnonmaxsigma[2], nthresh),
    ntests(blitz::product(tExt)) {}

  void update(int t) {
    blitz::TinyVector<int,4> const sub = ind2sub(tExt, t);
    assert(sub[0] >= 0 && sub[0] < nnonmaxsigma[0]);
    assert(sub[1] >= 0 && sub[1] < nnonmaxsigma[1]);
    assert(sub[2] >= 0 && sub[2] < nnonmaxsigma[2]);
    assert(sub[3] >= 0 && sub[3] < nthresh);
    float const nmWidth = 
      minnonmaxsigma[0] + 
      getStep(minnonmaxsigma[0],maxnonmaxsigma[0],nnonmaxsigma[0])*sub[0];
    float const nmHeight = 
      minnonmaxsigma[1] + 
      getStep(minnonmaxsigma[1],maxnonmaxsigma[1],nnonmaxsigma[1])*sub[1];
    float const nmScale = 
      minnonmaxsigma[2] + 
      getStep(minnonmaxsigma[2],maxnonmaxsigma[2],nnonmaxsigma[2])*sub[2];

    svmThreshold = (maxthresh - getStep(minthresh,maxthresh,nthresh)*sub[3]);
    nonmaxSigma  = NonmaxType(nmWidth, nmHeight, nmScale);
  }
  
  blitz::TinyVector<int,4> const tExt;
  int                      const ntests;

  float      svmThreshold;
  NonmaxType nonmaxSigma;
};     

void writePosTxt(vector<shared_ptr<vector<DetectedRegion> > > const &groundTruthDetections,
                 vector<string>                               const &infiles)
{
  safeMkdir(outdir);
  Hard_PPResult postxt(outdir + "/pos.txt");
  
  lear::IProcessor::RGBImage dummy(0,0);
  size_t const nImages = groundTruthDetections.size();
  for (size_t i=0; i<nImages; i++) {
    list<DetectInfo> gtl;
    dtlFromDrv(gtl, *groundTruthDetections[i], posscore);
    postxt.write(gtl.begin(), gtl.end(), dummy, infiles[i]);
  }
}

void sampleNegWindows(vector<shared_ptr<vector<DetectedRegion> > > const &groundTruthDetections,
                      vector<shared_ptr<PerspectiveProjector> >    const &projs,
                      vector<string>                               const &infiles)
{
  assert(groundTruthDetections.size() == projs.size());
  assert(groundTruthDetections.size() == infiles.size());
  assert(startscale <= endscale);

  safeMkdir(outdir);
  Hard_PPResult postxt(outdir + "/neg.txt");
  RandomGenerator rnd;
  for (size_t i=0; i<infiles.size(); i++) {
    // Gather samples
    map<RealType, DetList> negs;
    PerspectiveProjector const &proj = *projs[i];
    size_t nSampled = 0;
    size_t retries = 0;
    while (nSampled < negsamples && retries < maxretries) {
      IndexType const pos(rnd.uniform() * proj.imageExtent());
      float     const baseScale = (startscale==endscale) ? startscale :
        rnd.uniform() * (endscale-startscale) + startscale;
      DetectedRegion reg = 
        proj.detectedRegionFromHeadTip(pos(0), pos(1), baseScale);
      RealType const quantScale = (scaleprecision > 0) ?
        (int(reg.scale / scaleprecision + 0.5) * scaleprecision) : reg.scale;
      if (scaleprecision > 0) {
        reg.scale  = quantScale;
        reg.width  = int(winsize[0] * reg.scale);
        reg.height = int(winsize[1] * reg.scale);
      }

      if ((reg.x < 0) || (reg.x + reg.width  >= proj.imageExtent()(0)) ||
          (reg.y < 0) || (reg.y + reg.height >= proj.imageExtent()(1))) {
        // skip windows that overlap the image border
        retries++;
      } else if (findBestMatch(*groundTruthDetections[i], reg) != NOMATCH) {
        // the sampled window matches a real object
        retries++;
      } else {
        // it's a negative
        nSampled++;
        DetectInfo ri(negscore, reg.scale, reg.x,reg.y, reg.width,reg.height);
        negs[quantScale].push_back(ri);
      }      
    }
    
    // Write, with all samples from the same scale grouped together
    // (this significantly improves the performance of CUDA-enabled
    // dump_rhog when processing the file).
    {
      lear::IProcessor::RGBImage dummy(0,0);
      map<RealType, DetList>::const_iterator j = negs.begin();
      for (; j!=negs.end(); ++j) {
        postxt.write(j->second.begin(), j->second.end(), dummy, infiles[i]);
      }
    }
  }
}

void runOptimizer(vector<shared_ptr<vector<DetectedRegion> > > const &groundTruthDetectionsUnmasked,
                  vector<shared_ptr<vector<DetectedRegion> > > const &groundTruthDetections,
                  vector<shared_ptr<vector<DetectedRegion> > > const &baseDetections,
                  vector<shared_ptr<PerspectiveProjector> >    const &projs,
                  vector<string>                               const &infiles,
                  float                                               effectiveMinThresh,
                  int                                                 borderThickness)
{
  assert(groundTruthDetections.size() == baseDetections.size());
  if (verbose >= 2) {
    size_t ngt=0, nbd=0;
    for (size_t i=0; i<groundTruthDetectionsUnmasked.size(); i++) {
      ngt += groundTruthDetectionsUnmasked[i]->size();
      nbd += baseDetections[i]->size();
    }
    fflush(stdout);
    cout << "There are " << ngt << " ground truth objects and " << nbd
         << " detections (before non-maxima suppression and thresholding) "
         << "from " << groundTruthDetectionsUnmasked.size() << " images." 
         << endl << endl;
  }

  FILE *f = openCsvAndWriteHeaders();

  if (postxtout) {
    writePosTxt(groundTruthDetectionsUnmasked, infiles);
  }
  if (negtxtout) {
    sampleNegWindows(groundTruthDetectionsUnmasked, projs, infiles);
  }

  vector<shared_ptr<vector<DetectedRegion> > > const &gt = maskgt ?
    groundTruthDetections : groundTruthDetectionsUnmasked;

  fflush(stdout);
  if (f) fflush(f);

  OptimizerIter itShared;
#ifdef HAVE_OPENMP
#  pragma omp parallel for num_threads(nThreads) schedule(dynamic,1) 
#endif
  for (int t=0; t<itShared.ntests; t++) {
    OptimizerIter it;
    it.update(t);

    // Compute the mistakes and write image output
    if (it.svmThreshold >= effectiveMinThresh) {
      int tp,fp,fn, mistakes;
      float precision, recall;
      int64_t const start = nsTime();
      mistakesFromScores(tp, fp, fn, mistakes, precision, recall, 
                         gt, baseDetections, infiles,
                         it.svmThreshold, it.nonmaxSigma, minOverlapFrac,
                         borderThickness);
      double const etimeMS = (nsTime() - start)*1e-6;
      
      {
#ifdef HAVE_OPENMP
#       pragma omp critical 
#endif
        {
          printTableRow(f, it.svmThreshold, it.nonmaxSigma, 
                        tp,fp,fn, mistakes, precision,recall, etimeMS);
        }
      }
    }
  }
  if (f) fclose(f);
  if (verbose >= 1) printf("\n");
}

///////////////////////////////////////////////////////////////////////////////
// main

int main(int argc, char **argv) {
#ifndef _DEBUG
  try {
#endif
    vector<shared_ptr<vector<DetectedRegion> > > groundTruthDetectionsUnmasked; // gtd[imgIdx][detIdx]
    vector<shared_ptr<vector<DetectedRegion> > > groundTruthDetections; // gtd[imgIdx][detIdx]
    vector<shared_ptr<vector<DetectedRegion> > > baseDetections;
    vector<shared_ptr<PerspectiveProjector> >    projs;
    vector<string>                               infiles;
    float                                        effectiveMinThresh;
    parseCmdline(argc, argv);
    readInputData(groundTruthDetectionsUnmasked, groundTruthDetections, 
                  baseDetections, projs, infiles, effectiveMinThresh);
    runOptimizer(groundTruthDetectionsUnmasked, groundTruthDetections, 
                 baseDetections, projs, infiles, effectiveMinThresh, 
                 borderThickness);
#ifndef _DEBUG
  } catch(lear::Exception const &e) {
    cerr << "Caught " << e.what() << endl;
    return 1;
  } catch(std::exception const &e) {
    cerr << "Caught " << (e.what() ? e.what() : "UNKNOWN ERROR") << endl;
    return 1;
  } catch(...) {
    cerr << "Caught unknown exception" << endl;
    return 1;
  }
#endif
} 
