#include <stdio.h>
#include <iostream>
#include <fstream>
#include <string>
#include <vector>

#include "lib/cfg.h"

#include <boost/shared_ptr.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/filesystem/operations.hpp>

#include <lear/util/fileutil.h>
#include <lear/image/imageio.h>
#include <lear/image/inimageiter.h>
#include <lear/image/rescale.h>

#include <lear/cmdline.h>
#include <lear/vectoropt.h>
#include <lear/tinyvecopt.h>
#include <lear/util/customoption.h>

#include <lear/cvision/iprocessor.h>

namespace fs=boost::filesystem;
using namespace std;
using namespace lear;

typedef blitz::TinyVector<int,2>  IndexType;
typedef IProcessor::RealType      RealType;
typedef IProcessor::RGBImage      RGBImage; 

#ifndef MIN
#  define MIN(a,b) (((a)<(b)) ? (a) : (b))
#endif
#ifndef MAX
#  define MAX(a,b) (((a)>(b)) ? (a) : (b))
#endif

int       verbose;
string    infile;
string    outdir;
string    outfile;
string    outext;
bool      append;
bool      hflip, vflip;
bool      prunepartials;
IndexType winsize, outerpadding;

void parseCmdline(int argc, char **argv) {
  typedef lear::TinyVecOpt<int,2>         IndexOpt;

  IndexOpt winsize_, outerpadding_;

  IndexOpt::seperator(",x");

  lear::Cmdline cmdline;
  cmdline.commandName("extractwindows");
  cmdline.version("0.0.2", "Author: Gerald Dalley "
                  "(mailto: dalleyg@ieee.org)");
  cmdline.brief("Takes a window list file and generates image chips.");
  cmdline.description("Reads a list.txt file "
                      "(typically from \"classify_rhog --outfile [TXTFILE]\" "
                      "or \"detoptimizer\") and writes out image chips for "
                      "each window listed.");

  cmdline.addOption()
    ("verbose,v",option<int>(&verbose)
     ->defaultValue(7)->minValue(0)->maxValue(9),
     "Verbosity level")

    ("infile",option<string>(&infile),
     "A .txt file containing image names and detector windows.  One image "
     "chip will be created for each listed window.")
    ("outdir",option<string>(&outdir),
     "Directory where the image chips will be written.")
    ("outfile",option<string>(&outfile),
     "The name of a .lst file that will contain the name of each image "
     "chip that is written.  If no outfile is given, no .lst file will "
     "be created (but the image chips will still be written.")
    ("outext",option<string>(&outext)
     ->defaultValue("jpg"),
     "Filename extension for the image chips that are written (excluding "
     "the initial period \".\".  If no extension is given, then the input "
     "file's extension is used.")

    ("append",bool_option(&append)
     ->defaultValue(false),
     "Appends chip filenames to the [outfile] instead of replacing that file.")

    ("hflip",bool_option(&hflip)->defaultValue(false),
     "Horizontally flip the resulting image chip.")
    ("vflip",bool_option(&vflip)->defaultValue(false),
     "Vertically flip the resulting image chip.")

    ("prunepartials",bool_option(&prunepartials)
     ->defaultValue(true),
     "Don't generate image chips whose window overlaps the original image "
     "boundary.  If false, pixel replication is used for missing values")

    ("winsize",option<IndexOpt>(&winsize_)
     ->defaultValue(IndexOpt(64,128))->minValue(IndexOpt(1,1)),
     "Size of the detector window at its canonical scale, in pixels.")
    ("outerpadding",option<IndexOpt>(&outerpadding_)
     ->defaultValue(IndexOpt(16,16))->minValue(IndexOpt(0,0)),
     "Number of extra pixels outside a detector window to include when "
     "creating image chips.  All image chips will be of size "
     "[outerpadding]+[winsize]+[outerpadding].  If the image chip extends "
     "beyond the border of the input image, pixel replication is used.")
    ;

  int status = cmdline.parse(argc, argv);
  if (status != cmdline.ok) exit(status);

  if (infile.size() == 0) {
    ThrowException("No input file specified.");
  }
  if (outdir.size() == 0) {
    ThrowException("No output directory specified.");
  }

  for (int i=0; i<2; i++) {
    winsize[i]      = winsize_[i];
    outerpadding[i] = outerpadding_[i];
  }
}


void safeMkdir(string const &d) {
  fs::path p(d);
  if (fs::exists(p)) {
    if (!fs::is_directory(p)) {
      ThrowException("requested output directory "
                     "exists but is not a directory: " << d);
    }
  } else {
    fs::create_directories(d);
    if (!fs::exists(p)) {
      ThrowException("could not create output "
                     "directory: " << d);
    }
  }
}


#define BOUND(i,dim) \
  MAX(0, MIN(rescaled.extent()[dim]-1, i - outerpadding[dim] + start[dim]))
#define SRC rescaled(BOUND(x,0), BOUND(y,1))
#define X     x
#define Y     y
#define HFLIP (winsize[0]+outerpadding[0]*2 - x - 1)
#define VFLIP (winsize[1]+outerpadding[1]*2 - y - 1)
void extractChips() {
  if (verbose >= 2) {    
    cout << "Extracting windows from " << infile;
    if (outfile.size()) {
      cout << ", writing chips to " << outdir 
           << ", and recording their names to " << outfile << "." << endl;
    } else {
      cout << " and writing chips to " << outdir << "." << endl;
    }
  }

  FILE   *lst = NULL;
  string  reldir = ".";
  if (outfile.size()) {
    if (!(lst = fopen(outfile.c_str(), append ? "a" : "w"))) {
      ThrowException("Could not open " << outfile << " for writing.");
    }
    //reldir = fs::complete(fs::path(outdir, fs::native), 
    //                      lear::dirname(outfile)).native_file_string();
    reldir = lear::reldir(outdir, lear::dirname(outfile));
  }

  safeMkdir(outdir);

  std::ifstream from(infile.c_str());
  if (!from) {
    ThrowException("Unable to open input file " << infile);
  }

  int winidx = 0;
  IndexType newsize;
  RealType prevscale=0;
  std::string prevfile="";
  RGBImage origimage, rescaled, chip(winsize + outerpadding*2);
  while (from) {
    std::string file;
    from >> file;
    IndexType lbound, extent;
    RealType scale, score;
    from >> lbound[0] >> lbound[1] >> extent[0] >> extent[1] >> scale >> score;
    if (!from) break;
    if (file.empty()) continue;
    
    bool newimage;
    if (prevfile != file){
      lear::ImageIO::read(file,origimage);
      newimage = true;
      prevfile = file;
      winidx   = 0;
    } else {
      newimage = false;
    }
    if (verbose >= 4) {
      cout << "Processing file " << file << " at " 
           << "(" << right << setw(4) << lbound[0] << "," 
           << left << setw(4) << lbound[1] << ") + " 
           << "(" << right << setw(4) << extent[0] << "x" 
           << left << setw(4) << extent[1] << "), " 
           << "with scale " << scale << " and score " << score
           << endl;
    }
    IndexType start(lbound/scale);
    //if (hasTopLeft) {
    //    start += topleft;
    //} else 
    {
      // extent difference, can have one-two pixel error
      IndexType exdiff = extent/scale - winsize; 
      exdiff = max(exdiff,0); // in case, there is small error due 
      // to rounding off while writing extent
      start += exdiff/2;
    }

    bool reprocess = false;
    if (newimage || std::abs(scale - prevscale) > 1e-3) {
      prevscale = scale;
      reprocess = true;
    }
    IndexType newsize(ceil(origimage.extent()/scale));
    if (reprocess) {
      rescaled.resize(newsize);
      rescaled = lear::rescale(origimage, newsize);
    }

    if (prunepartials && 
        (blitz::any(start < 0) || blitz::any(start+winsize > newsize))) {
      // skip windows that overlap the image boundary
    } else {
      if (hflip) {
        if (vflip) {
          for (int x=0; x<winsize[0]+outerpadding[0]*2; x++) {
            for (int y=0; y<winsize[1]+outerpadding[1]*2; y++) {
              chip(HFLIP,VFLIP) = SRC;
            }
          }   
        } else {
          for (int x=0; x<winsize[0]+outerpadding[0]*2; x++) {
            for (int y=0; y<winsize[1]+outerpadding[1]*2; y++) {
              chip(HFLIP,Y) = SRC;
            }
          }   
        } 
      } else {
        if (vflip) {
          for (int x=0; x<winsize[0]+outerpadding[0]*2; x++) {
            for (int y=0; y<winsize[1]+outerpadding[1]*2; y++) {
              chip(X,VFLIP) = SRC;
            }
          }   
        } else {
          for (int x=0; x<winsize[0]+outerpadding[0]*2; x++) {
            for (int y=0; y<winsize[1]+outerpadding[1]*2; y++) {
              chip(X,Y) = SRC;
            }
          }   
        }
      }
      
      char winidxStr[128];
      sprintf(winidxStr, "%04d", winidx);
      string const basename = lear::basename(file);
      string const ext      = "." + (outext.size() ? 
                                     outext : lear::extension(file));
      string outname = (basename + "_" + winidxStr);
      if (hflip || vflip) outname += "_";
      if (hflip) outname += "h";
      if (vflip) outname += "v";
      outname += ext;
      
      if (verbose >= 7) {
        cout << "  Writing " << reldir << "/" << outname << endl;
      }
      
      lear::ImageIO::write(outdir + "/" + outname, chip);
      if (lst) {
        fprintf(lst, "%s/%s\n", reldir.c_str(), outname.c_str());
        fflush(lst);
      }
      
      winidx++;
    }
  }
    
  if (lst) fclose(lst);
}
#undef VFLIP
#undef HFLIP
#undef Y
#undef X
#undef SRC
#undef BOUND


int main(int argc, char **argv) {
#ifndef _DEBUG
  try {
#endif
    parseCmdline(argc, argv);

    extractChips();

#ifndef _DEBUG
  } catch(lear::Exception const &e) {
    cerr << "Caught " << e.what() << endl;
    return 1;
  } catch(std::exception const &e) {
    cerr << "Caught " << (e.what() ? e.what() : "UNKNOWN ERROR") << endl;
    return 1;
  } catch(...) {
    cerr << "Caught unknown exception" << endl;
    return 1;
  }
#endif
}
