/*
 * ============================================================================
 *
 *       Filename:  windetectmain.cpp
 *
 *    Description:  Provides implementation of windetectmain.h
 *
 *        Version:  1.0
 *        Created:  Sunday 03 June 2007 10:56:03  IST IST
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Navneet Dalal
 *        Company:  Lear / INRIA
 *
 * ============================================================================
 */

#include <blitz/array.h>
#include <blitz/tinyvec.h> 

#include <lear/util/fileutil.h>
#include <lear/image/imageio.h>

#include <boost/filesystem/path.hpp>
#include <boost/filesystem/operations.hpp>

#include <lear/option.h>

#include "../lib/cfg.h"
#include "windetectmain.h"
#ifdef HAVE_TRACKER
#include <tracker/JTracker.h>
#endif

//## common main options

typedef blitz::TinyVector<RealType,3>       PixelType;
typedef blitz::TinyVector<int,3>            RGBType;
typedef blitz::Array<PixelType,2>           ImageType;

using namespace std;
using namespace lear;

void WinDetectMain::fill(WinDetect* param) const
{
  param->size_x = size[0];
  param->size_y = size[1];

  param->winstride_x = winstride[0];
  param->winstride_y = winstride[1];

  param->finestride_x = finestride[0];
  param->finestride_y = finestride[1];

  param->topleft_x  = topleft[0];
  param->topleft_y  = topleft[1];

  param->fullsize_x  = fullsize[0];
  param->fullsize_y  = fullsize[1];

}
void WinDetectDumpMain::fill(WinDetectDump* param) const
{
  WinDetectMain::fill(param);

  // WinDetectDump parameters
  param->jitterstep_x  = jitterstep[0];
  param->jitterstep_y  = jitterstep[1];
}
void WinDetectClassifyMain::fill(WinDetectClassify* param) const
{
  WinDetectMain::fill(param);

  // WinDetectClassify parameters
  param->margin_x = margin[0];
  param->margin_y = margin[1];

  param->initialsize_x = initialsize[0];
  param->initialsize_y = initialsize[1];

  param->avsize_x = avsize[0];
  param->avsize_y = avsize[1];

  param->alignmargin_x = alignmargin[0];
  param->alignmargin_y = alignmargin[1];

  param->fullstride_x = fullstride[0];
  param->fullstride_y = fullstride[1];

  param->nonmaxsigma_x = nonmaxsigma[0];
  param->nonmaxsigma_y = nonmaxsigma[1];
  param->nonmaxsigma_scale = nonmaxsigma[2];

  param->score2prob_a = score2prob[0];
  param->score2prob_b = score2prob[1];

  param->rodrigues   = rodrigues;
  param->translation = translation;
  param->focalLength = focalLength;
}

// Cmdline parsing 
void WinDetectMain::setCommonMainParam(lear::Cmdline& cmdline, WinDetect* param) 
{
  IndexOpt::seperator(",x");
  MultIndexOpt::seperator(":^");

#ifdef __RELEASE__
  std::string s=" (Release date ";
  s+=__DATE__;
  s+=")";
  cmdline.appendBrief(s);
#endif
  using namespace lear;
  std::ostringstream usageIssues ;
  usageIssues <<
    "options 'stride', 'startscale', 'scaleratio' and 'endscale' "
    " are valid iff 'pyramid' is true'\n"
    "  'infile'         can be an image file, a directory, or a list file.\n"
    "  'imageext'       empty string implies all listed files, otherwise images "
    "with given extension are selected\n"
    ;
  cmdline.appendUsageIssues(usageIssues.str());


  cmdline.addOption()
#ifdef HAVE_CUDA
    ("cuda",bool_option(&param->useCuda)->defaultValue(true),
     "Use the graphics processing unit to accelerate key algorithms "
     "Requires a nVidia-based graphics card with at least CUDA "
     "compute capability 1.2.  Enabling CUDA support may limit "
     "the values supported for other parameters.")
#endif
    ("window,W",option<IndexOpt>(&(size))
     ->defaultValue(IndexOpt(64,128))->minValue(3),
     "Canonical detector window width,height.  The classifier and "
     "training programs must all use the same window size.  This helps "
     "decide the size of the feature vector.")
    ("winstride",option<IndexOpt>(&(winstride))
     ->defaultValue(8),
     "window stride along x-y.  Descriptors are sampled from a uniform "
     "rectangular grid with these edge lengths.")
    ("finestride",option<IndexOpt>(&(finestride))
     ->defaultValue(-1),
     "determines how often (in a spatial sense) block descriptors are "
     "precomputed when using the CUDA implementation.  As of svn revision "
     "780, the fine stride must be the same as winstride when using a "
     "perspective slider, and it must be 1 for a pyramid slider.  Using "
     "a value of -1 causes the winstride to be used as the finestride.")
    ("quantizepos",bool_option(&(param->quantizepos))
     ->defaultValue(false),
     "When true, the window descriptor implementation is allowed to round "
     "detector window locations to nearby grid locations to increase speed "
     "and/or reduce memory requirements.")

    ("scaleratio",option<RealType>(&(param->scaleratio))
     ->defaultValue(1.05f)->minValue(1),
     "scale ratio.  The start scale is multiplied by this number until "
     "the endscale is reached.  This produces the set of scales that "
     "are examined in the image.")
    ("endscale",option<RealType>(&(param->endscale)),
     "scale-space pyramid end scale\n"
     "  (Default: till image size >= window size)")
    ("startscale",option<RealType>(&(param->startscale))
     ->defaultValue(1)->minValue(1/32768.0f),
     "start scale.  Larger scales correspond to trying to detect "
     "larger pedestrians.")

    ("topleft",option<IndexOpt>(&(this->topleft)),
     "top-left X,Y-coordinate of the (first) descriptor window. "
     "(Default window is at image center)")
    /* It's not clear what fullsize really does...looking at the code, it doesn't
       seem to quite do what my expanded comment suggests.  I'm removing the
       option until I can figure it out.
       //#ifndef __RELEASE__
       ("fullsize",option<IndexOpt>(&(this->fullsize)),
       "actual window width,height from which the data are sampled.  If "
       "the fullsize is not specified, the 'window' parameter's value is "
       "used.  Use iff top-left is valid, not used for hard examples.")
       //#endif
       */
    ("cachesize",option<int>(&(param->cachesize))
     ->defaultValue(128)->minValue(4)
     ->maxValue((int)(std::numeric_limits<int>::max()/1024/1024*sizeof(float))), // conservative max value.  Cannot be made size_t because blitz++ uses int indices instead of size_t indices.
     "descriptor cache size (im MB) to use\n"
     "  NOTE: small cache size can drastically reduce speed")
    ("verbose,v",option<int>(&(param->verbose))
     ->defaultValue(0)->minValue(0)->maxValue(9),
     "verbose level")

    ("testlocs",option<std::string>(&(testlocs))
     ->defaultValue(""),
     "filename where to write all test or feature dump locations\n"
     "  (Format: imagename X Y Width Height)")

    ("txtimgnames",bool_option(&(writeListFilename))->defaultValue(true),
     "Include the image file names in the outfile.")

    ("imageext,x",option<std::string>(&(imageext))
     ->defaultValue(""),
     "image extension (valid only if 'infile' = dir or list)")
    ;

  cmdline.addArgument()
    ("infile",option<std::string>(&(infile)),"input image file/dir/list")
    ;
}

void WinDetectDumpMain::setDumpParam( lear::Cmdline& cmdline, WinDetectDump* param) 
{ 
  using namespace lear;
  //#ifndef __RELEASE__
  cmdline.appendUsageIssues(
                            "   'dumpdetail'    dump window detail in the output file"
                            "(currently does not work for hard examples or with pyramid)\n"
                            "                   0 - Write only descriptor\n"
                            "                   1 - Write descriptor top-left, extent, scale\n"
                            "                   2 - Write filename information also\n"
                            );
  //#endif

  cmdline.addOption()
    //#ifndef __RELEASE__
    ("dumpdetail,f",option<int>(&(param->dumpfulldetail))
     ->minValue(0)->maxValue(3)->defaultValue(0),
     "dump extra information about the features in the file")
    //#endif
    ("samples,s",option<size_t>(&(param->samples))
     ->defaultValue(1)->minValue(0),
     "Number of random sampling windows over each image OR  "
     "if hard is true, number of samples to skip")
    ("random,r",bool_option(&(param->randomSeed)),
     "Initialize random generator with current time")
    ("hardscore",option<RealType>(&(param->hardscore))
     ->defaultValue(0),
     "Use only points with score greater than hardscore")

    ("poscases",option<int>(&(param->poscases)),
     "(Estimated) number of positive cases.  Used to help limit overall "
     "memory usage if memorylimit is enabled.")
    ("negcases",option<int>(&(param->negcases)),
     "(Estimated) number of negative cases.  Used to help limit overall "
     "memory usage if memorylimit is enabled.")
    ("memorylimit",option<int>(&(param->memorylimit)),
     "Upper memory limit on training examples.  "
     "Ignored if hard is false.")
    //#ifndef __RELEASE__
    ("jitterwinstep",option<int>(&(param->jitterwinstep))
     ->defaultValue(0)->minValue(0),
     "Jitter window steps")
    ("jitterstep",option<IndexOpt>(&jitterstep)
     ->defaultValue(1)->minValue(1),
     "jitter step size")
    //#endif
    ("label,l",option<char>(&(param->label))
     ->defaultValue(WinDetect::DefaultLabel)->minValue('N')->maxValue('P'),
     "Hard file labels (Default is ignore label)")
   
    ("pyramid,p",bool_option(&(param->pyramid)),
     "Compute descriptors over scale-space pyramid")


    ("outfile",option<std::string>(&(outfile)),
     "output text file containing one line per detection")
    ;
}


void WinDetectClassifyMain::setClassifyParam( lear::Cmdline& cmdline, WinDetectClassify* param) {
  using namespace lear;
  cmdline.addOption()
    ("initialsize",option<IndexOpt>(&initialsize)
     ->defaultValue(IndexOpt(-1,-1)),
     "If supplied, input images will first be rescaled to this size before "
     "further processing.  This option is especially useful when a whole "
     "collection of images are the same size and aspect ratio correction "
     "is needed.  A value of -1 means no extra resizing will be done in "
     "that dimension.")
    ("maximages",option(&param->maximages)
     ->defaultValue(-1),
     "If supplied, only the first [maximages] files will actually be "
     "processed.  This option is useful for having quick tests with "
     "input videos stop after a few frames.")

    ("svmthreshold,m",option<RealType>(&(param->lightthreshold))
     ->defaultValue(0),
     "threshold on svm score.  Only those locations which are above "
     "this value, are passed to non-maximum suppression stage.")
    ("threshold,t",option<RealType>(&(param->threshold))
     ->defaultValue(0.1f),
     "threshold to classify.  Final threshold after non-maximum "
     "threshold.")
    ("no_nonmax",bool_option(&(param->no_nonmax)),
     "do not perform non maximum suppression")
    ("avsize",option<IndexOpt>(&avsize)
     ->minValue(0)->defaultValue(IndexOpt(0,96)),
     "compute image margin to add using average size. "
     "0 imply do not use average size.")
    ("margin",option<IndexOpt>(&margin)
     ->minValue(0)->defaultValue(4),
     "pixels margin at each side in normalized image. "
     "0 imply do not use margin size.")
    //#ifndef __RELEASE__
    ("fullstride",option<IndexOpt>(&fullstride),
     "full reference stride\n"
     "  Use iff working with parts & top-left is valid")
    //#endif
    ("slider",option<int>(&(param->sliderType))
     ->defaultValue(1)->minValue(0)->maxValue(3),
     "how detector windows are scanned across the images.  "
     "Values (0=centered,1=pyramid,2=perspective). "
     "NOTE: centered runs a single detector window, at the canonical "
     "  scale, centered in the image.  "
     "NOTE: pyramid uses an exhaustive scale space pyramid to scan "
     "  the whole image.  "
     "NOTE: perspective works like pyramid, but min and max scales "
     "  are scaled by the perspective distortion parameters.  ")
    ("borderpixels", option<int>(&(param->borderPixels))
     ->defaultValue(4)->minValue(0),
     "normally the detector windows are designed to cover a pedestrian "
     "plus some extra contextual pixels.  If a detector window is H "
     "pixels tall in total, the pedestrians are assumed to be "
     "H - 2*borderpixels tall.  "
     "This option is only used when the perspective slider is being "
     "used (slider=2).")
    ("pedestrianHeight", option<RealType>(&(param->pedestrianHeight))
     ->defaultValue(5 + 9.f/12),//1.77f),
     "canonical height of a pedestrian, in world coordinates.  "
     "This option is only used when the perspective slider is being "
     "used (slider=2).")
    ("scaleprecision", option<RealType>(&(param->scalePrecision))
     ->defaultValue(0.1f)->minValue(0),
     "precision to use when quantizing window scales to speed up the "
     "runtime performance.  If a scale of 1.2452 is computed and "
     "scaleprecision=0.01, then a scale of 1.25 will actually be "
     "used.  "
     "This option is only used when the perspective slider is being "
     "used (slider=2).")
    ("rodrigues", option<RodriguesOptType>(&rodrigues)
     ->defaultValue(RodriguesOptType(1.555496860243484f, -1.561712756714797f, 0.911809104345828f)),
     "Rodrigues world-to-camera rotation vector.  "
     "This corresponds to the OM vector returned by the Matlab "
     "Camera Calibration Toolbox.  "
     "This option is only used when the perspective slider is being "
     "used (slider=2).")
    ("translation", option<WorldVectorOptType>(&translation)
     ->defaultValue(WorldVectorOptType(62.9644f, 56.0368f, 81.7144f)),
     "Post-rotation translation to use in world-to-camera transformations, "
     "in world spatial coordinates.  "
     "This corresponds to the T vector returned by the Matlab "
     "Camera Calibration Toolbox.  "
     "This option is only used when the perspective slider is being "
     "used (slider=2).")
    ("focalLength", option<FocalLengthOptType>(&focalLength)
     ->defaultValue(FocalLengthOptType(2379.161604170919f, 2379.161604170919f)),
     "Camera focal length, in pixels (fx,fy).  "
     "This corresponds to the F vector returned by the Matlab "
     "Camera Calibration Toolbox.  "
     "This option is only used when the perspective slider is being "
     "used (slider=2).")
    ("showscore", bool_option(&(param->showscore))
     ->defaultValue(param->showscore),
     "show final classifier scores for detections.  "
     "Ignored if outimage is not supplied.") 
    ("showbestscores", bool_option(&(param->showbestscores))
     ->defaultValue(param->showbestscores),
     "for every pixel location, show the best final classifier scores for "
     "detections.  This can be useful for making per-pixel ROC curves with "
     "external tools.  Ignored if outimage is not supplied.")
    ("softmax",option<int>(&(param->softmax))
     ->defaultValue(0)->minValue(0)->maxValue(3),
     "transformation function method used  to convert scores to "
     "probabilities. "
     "Values (0=hardclip,1=sigmoid,2=softclip,3=none). "
     "NOTE 1: score2prob provides A,B parameters for sigmoid"
     " and softclip method. " 
     "NOTE 2: (score2prob[0],svmthreshold) provides A,B values"
     " for hardclip methods."
     )
    ("score2prob,p",option<SigmaOptType>(&score2prob)
     ->defaultValue(SigmaOptType(1,0)),
     "(A,B) parameters of sigmoid function (1+exp(A*score+B))^-1"
     " or soft clip log(1+exp(A*(score-B)")
    ("nonmaxsigma,z",option<NonmaxOptType>(&nonmaxsigma)
     ->defaultValue(NonmaxOptType(8,16,1.3f))
     ->minValue(NonmaxOptType(0,0,1)),
     "smooth sigma for non-max suppression (x,y, scale)")
    
    ("outimage,i",option<std::string>(&outimage),
     "align input image to max of classifier\n"
     "  (valid iff no_nonmax=1 && alignin=1)\n"
     " ELSE\n"
     "  mark classification result on input image\n"
     "Will not work properly if 'label' parameter is not also specified.")
    ("alignin",bool_option(&(param->aligninimage)),
     "align input image (outimage must be specified)")
    ("alignmargin",option<IndexOpt>(&alignmargin)
     ->minValue(0)->defaultValue(4),
     "margin to add to aligned image")
    ("outhist,c",option<std::string>(&outhist),
     "write score histogram (for roc curve)")
    ("label,l",option<char>(&(param->label))
     ->defaultValue(WinDetect::DefaultLabel)->minValue('N')->maxValue('P'),
     "All examples are assumed to have this label (N=negative, "
     "P=positive).  label must be explicitly specified for the falsetxt "
     "parameter to work (Default is to not output hard example windows "
     "file).  If imageout is given, this is used to indicate correct "
     "vs. incorrect detections (Default is invalid in this case).  ")
    ("falsetxt,d",option<std::string>(&falsetxt)
     ->defaultValue(""),
     "output filename for image file + descriptor windows.  The file is "
     "primarily for hard retraining to prevent false positives.  Must be"
     "used with the 'label' parameter.")

#ifdef HAVE_TRACKER
    ("trkfile",option(&param->trkfile),
     "If supplied, the input images are assumed to come from a video and "
     "a constant-velocity Kalman tracker is used to associate detections "
     "across time.  The output is written to this file.")
    ("trkformat",option<std::string>(&param->trkformat)
     ->defaultValue("josh"),
     "Specifies whether the [trkfile] use Josh Migdal's format (\"josh\") or "
     "Chris Stauffer's (\"stauffer\").  See the tracker library for details.")
    ("fps",option(&param->fps)->defaultValue(24/1.001),
     "Number of frames per second.  This affects how the velocity parameters "
     "([ivdxy], [maxvdxy], and [minvdxy]) are interpreted.")
    ("maxtrk",option(&param->maxtrk)->defaultValue(5*30*1024),
     "Maximum number of active tracks.  This allows the user to cap the "
     "memory usage and runtime when there are too many false positive "
     "detections and/or the data association is failing.  If more active "
     "tracks than this are needed, the application will exit.  If too many "
     "tracks are being created, try increasing [maxtrk] and/or reducing the "
     "[threshold], [svmthreshold], [searchtime] and/or widen the allowed "
     "tracker variance values.")
    ("maxdet",option(&param->maxcc)->defaultValue(1024),
     "Maximum number of detections allowed in any single image.  See [maxtrk] "
     "for details.")
    ("ivxy",option(&param->ivxy)->defaultValue(DEFAULT_INITIAL_VXY),
     "Gaussian variance on the position of a detection when a new track is "
     "initialized.  This value should usually be much closer to than [maxvxy] "
     "than it is to [minvxy] (often [ivxy]=[maxvxy] works very well).  "
     "Units: pixels^2.")
    ("ivdxy",option(&param->ivdxy)->defaultValue(DEFAULT_INITIAL_VDXY),
     "Gaussian variance on the velocity of a detection when a new track is "
     "initialized.  The initial velocity is always assumed to be 0.  See "
     "[ivxy] for additional hints.")
    ("iva",option(&param->iva)->defaultValue(DEFAULT_INITIAL_VA),
     "Gaussian variance on the area of a detection window when a new track is "
     "initialized.  Units: ((pixels)^2)^2.  "
     "See [ivxy] for additional hints.")
    ("maxvxy",option(&param->maxvxy)->defaultValue(DEFAULT_MAX_VXY),
     "The Gaussian variance on the position of a track is capped to be no "
     "greater than this value.  Use smaller values to prevent spurious data "
     "associations with far-away detections.  Use larger values to compensate "
     "for poor detector window localization.  The maximum detector window "
     "localization error should typically be O(sqrt([maxvxy])).  Because the "
     "initial velocity of a new track is assumed to be 0, [maxvxy] should "
     "also typically be large enough to account for the maximum typical speed "
     "of an object.  Units are (pixels/sec)^2."
     "See [ivxy] for additional hints.")
    ("maxvdxy",option(&param->maxvdxy)->defaultValue(DEFAULT_MAX_VDXY),
     "Maximum Gaussian variance on the velocity of a track.  Use smaller "
     "values to prevent spurious data associations when objects move with "
     "(near) constant velocity.  Use larger values when objects tends to "
     "change speed and direction quickly.  "
     "See [maxvxy] and [ivxy] for additional hints.")
    ("maxva",option(&param->maxva)->defaultValue(DEFAULT_MAX_VA),
     "Maximum Gaussian variance on the detector window area.  "
     "See [iva], [maxvxy], and [ivxy] for additional hints.")
    ("minvxy",option(&param->minvxy)->defaultValue(DEFAULT_MIN_VXY),
     "Minimum Gaussian variance on the position of a track.  Use smaller "
     "values to allow the Kalman tracker to learn a very precise position "
     "model and avoid false long-distance data associations in cluttered "
     "situations.  Use larger values when the detector window accuracy can "
     "be high for a long period of time then suddenly it becomes inaccurate "
     "and associations are being missed.  "
     "See [maxvxy] and [ivxy] for additional hints.")
    ("minvdxy",option(&param->minvdxy)->defaultValue(DEFAULT_MIN_VDXY),
     "Minimum Gaussian variance on the velocity of a track.")
    ("minva",option(&param->minva)->defaultValue(DEFAULT_MIN_VA),
     "Minimum Gaussian variance on the area of a track.")
    ("matchdist",option(&param->addThreshold)->defaultValue(DEFAULT_ADD_THRESHOLD),
     "A detection must have a Mahalanobis distance to an existing track "
     "of [matchdist] or less to be considered a match.  Use smaller values "
     "when new objects get associated with existing tracks too often.  Use "
     "larger values when tracks are being broken up too often.  The current "
     "system (as of Spring 2009) uses greedy matching.")
    ("deldist",option(&param->delThreshold)->defaultValue(DEFAULT_DEL_THRESHOLD),
     "An existing track is immediately terminated if it is not matched to "
     "any detections and the minimum Mahalanobis distance of any detection "
     "in the current frame is above this value.  When [deldist]<[matchdist], "
     "this can still happen if other nearby tracks won the data association "
     "contest.  Use a very large value if missed detections are common and "
     "you would like [minunseensecs] to be the primary mechanism for "
     "terminating old tracks.")
    ("maxunseenms",option(&param->searchTime)->defaultValue(DEFAULT_SEARCH_TIME),
     "Maximum number of milliseconds that a track stays active since its most "
     "recent association with a detection.  Tracks that have not been "
     "matched within this time window are terminated.")
    ("prate",option(&param->pLearningRate)->defaultValue(DEFAULT_POS_LEARNING_RATE),
     "Learning rate for the position of a track.  This is also called the "
     "\"Kalman gain.\"  Use smaller values when the measurements and/or "
     "motions are noisy, to slow down how quickly the Kalman filter model "
     "adapts.  Use larger values when the constant velocity assumption "
     "is good and quicker adaptation is desired.  Units: 1/frames (I think).")
    ("vrate",option(&param->vLearningRate)->defaultValue(DEFAULT_VEL_LEARNING_RATE),
     "Learning rate for the velocity of a track.  See [prate] for hints.")
    ("arate",option(&param->aLearningRate)->defaultValue(DEFAULT_AREA_LEARNING_RATE),
     "Learning rate for the area of a track's detections.  "
     "See [prate] for hints.")
#endif

    ("maskfile",option(&maskfile),
     "Detections that land on black pixels in this image are ignored.  "
     "Currently, this option assumes that all images in the dataset have "
     "the same mask.") 
    ("outfile",option<std::string>(&(outfile)),
     "output text file containing one line per detection")
    ;
  cmdline.addArgument()
    ("modelfile",option<std::string>(&modelfile),
     "learned SVM light model file.  NOTE: If value is equal to "
     "'defaultperson', it loads default person detector.")
    ;
}

size_t RHOGDenseMain::size() const
{
  vector<size_t> svec;
#ifdef WIN32
  // 64-bit warning is incorrectly firing with VS2005 (svec now contains size_t)
#pragma warning(push)
#pragma warning(disable:4267)
#endif
  svec.push_back(iprocessor.size());
  svec.push_back(normalizer.size());
  svec.push_back(cellsize.size());
  svec.push_back(numcell.size());
  svec.push_back(descstride.size());
  svec.push_back(orientbin.size());
  svec.push_back(wtscale.size());
  svec.push_back(fullcirc.size());
  svec.push_back(gscale.size());
  svec.push_back(epsilon.size());
  svec.push_back(maxvalue.size());
#ifdef WIN32
#pragma warning(pop)
#endif

  return *std::max_element(svec.begin(),svec.end());
}
// descriptor specific parameters
void RHOGDenseMain::setRHOGDenseParam(lear::Cmdline& cmdline) 
{
  using namespace lear;

  cmdline.appendUsageIssues( iprocessoropt_.usage());
  cmdline.appendUsageIssues( normalizeropt_.usage());

  cmdline.addOption()
    ("proc",option<MultStrOpt>(&(iprocessor))
     ->defaultValue(iprocessoropt_.defaultOption()),
     "Image preprocessor and gradient computation method to use")
    ("norm",option<MultStrOpt>(&(normalizer))
     ->defaultValue(normalizeropt_.defaultOption()),
     "Descriptor normalizing method to use")

    ("cellsize,C",option<MultIndexOpt>(&(cellsize))
     ->defaultValue(8)->minValue(1),
     "cell size, position invariance tolerance in each cell")
    ("numcell,N",option<MultIndexOpt>(&(numcell))
     ->defaultValue(2)->minValue(1),
     "number of cells in each block.  Each descriptor block is cellsize*numcell pixels.")
    ("descstride,G",option<MultIndexOpt>(&(descstride))
     ->defaultValue(8)->minValue(1),
     "descriptor stride in window.  Distance between each descriptor block.")
    ("orientbin,B",option<MultIntOpt>(&(orientbin))
     ->defaultValue(9)->minValue(1),
     "number of orientation bin")
    ("wtscale",option<MultRealOpt>(&(wtscale))
     ->defaultValue(2)->minValue(0),
     "magnitude weighting scale")
    ("fullcirc,O",option<MultBoolOpt>(&(fullcirc)),
     "take orientations in range (0-360)")
    ("gscale,S",option<MultRealOpt>(&(gscale))
     ->defaultValue(0)->minValue(0),
     "gradient computation scale")
    ("epsilon",option<MultRealOpt>(&(epsilon))
     ->defaultValue(1)->minValue(0),
     "epsilon to add while normalizing")
    ("maxvalue",option<MultRealOpt>(&(maxvalue))
     ->defaultValue(0.2f)->minValue(0)->maxValue(1),
     "chop feature vector values > maxvalue after normalizing\n"
     "  Used iff normalizing method == NormL2Hys")
    ;
}

std::vector<RHOGDenseParam const *> RHOGDenseMain::fill() const
{
  size_t numdesc = size();

  if (numdesc != 1) {
    ThrowException("GED: I think that the following code is broken when numdesc != 1.");
  }

  std::vector<RHOGDenseParam const *>    desclist(numdesc);
  for (unsigned i= 0; i< numdesc; ++i) {
    RHOGDenseParam* desc = new RHOGDenseParam();

    // TODO: This code seems like it can only work if each element has exactly 0 or 1 elements.
    if (iprocessor.size())
      desc->preprocessing2use = (RHOGDenseParam::PreprocessorFlags)iprocessoropt_.check(iprocessor.at(i));

    if (normalizer.size())
      desc->norm2use = (RHOGDenseParam::NormalizerFlags)normalizeropt_.check(normalizer.at(i));

    if (cellsize.size()) {
      desc->cellsize_x = cellsize.at(i)[0];
      desc->cellsize_y = cellsize.at(i)[1];
    }

    if (numcell.size()) {
      desc->numcell_x = numcell.at(i)[0];
      desc->numcell_y = numcell.at(i)[1];
    }

    if (descstride.size()) {
      desc->descstride_x = descstride.at(i)[0];
      desc->descstride_y = descstride.at(i)[1];
    }

    if (orientbin.size()) 
      desc->orientbin = orientbin.at(i);

    if (fullcirc.size()) 
      desc->semicirc = !fullcirc.at(i);

    if (wtscale.size()) 
      desc->wtscale = wtscale.at(i);

    if (gscale.size()) 
      desc->gscale = gscale.at(i);

    if (epsilon.size()) 
      desc->epsilon = epsilon.at(i);
    if (maxvalue.size()) 
      desc->maxvalue = maxvalue.at(i);

    desclist[i] = desc;
  }
  return desclist;
}

