/**
 * @file
 * @brief Compute Mean Shift's for general arrays
 */

#ifndef _LEAR_MEANSHIFT_H_
#define _LEAR_MEANSHIFT_H_

//#define CONVERGETOMODE_DEBUG

#include <list>
#include <algorithm>
#include <functional>
#include <vector>
#include <lear/util/functional.h>

namespace lear {

  template<class RealType_>
  struct Basic1DKernel {
    typedef RealType_                       RealType;
    typedef RealType                        PointType;

    Basic1DKernel(const PointType sigma): sigma(sigma) {}

    // 1st moments of the weighted points 
    template<class Pt, class Wt, class Ref>
    inline PointType nvalue(Ref rf, Pt first, Pt last, Wt wt) const {
      using namespace std;
      PointType numer(.0);
      PointType denom(.0);

      for (;first != last; ++first, ++wt) {
        PointType x = (*first - rf)/sigma;
        RealType  w = (*wt)*exp(-x*x/2);

        numer += w*(*first);
        denom += w;
        //std::cout << "Loc " << *first << " | ref " << rf << " | sigma " << sigma << " | wt " << *wt << " | w " << w << std::endl;
      }
      numer /= denom;
      //std::cout << "Final value " << numer << std::endl;
      return numer;
    } 

    // 0th moment of the weighted points 
    template<class Pt, class Wt, class Ref>
    inline RealType fvalue(Ref rf, Pt first, Pt last, Wt wt) const {
      RealType numer=0;

      for (;first != last; ++first, ++wt) {
        PointType x = (*first - rf)/sigma;
        numer+=(*wt)*std::exp(-x*x/2)/(2*3.1415926536*sigma); 
      }
      return numer;
    } 

    RealType distsq(PointType a, PointType b) const {
      b-=a;
      b/=sigma;
      return b*b;
    }

  protected:
    PointType sigma;
  };



  // Mean Shift class

  /**
   * Compute Mean shift vectors for an image.
   * weight vector and position vector must be of equal size
   */


  /**
   * Implements basic mean shift operations without weight function
   */
  template<class Kernel_, class PtContType_>
  class BasicMeanshift{
    // Meanshift typedefs
  public:
    typedef Kernel_                                 Kernel;
    typedef typename Kernel::RealType               RealType;
    typedef typename Kernel::PointType              PointType;

    typedef PtContType_                             PtContType;

    typedef BasicMeanshift<Kernel, PtContType>      ThisType;

  protected:
    /// reference to position vector
    PtContType const &at;

    /// smooth kernel to use
    Kernel const &kernel;

    RealType const modeEpsilon;

    const int           maxIterations;

  public:
    // Constructors
    BasicMeanshift(PtContType const &at_, Kernel const &kernel_,
                   RealType const modeEpsilon_ = 1e-4,
                   int const maxIterations_ = 20)
      : at(at_), kernel(kernel_), modeEpsilon(modeEpsilon_), 
        maxIterations (maxIterations_)
    {}


    inline PointType shiftToMode(PointType I) {
      int count = 0;

      // store previous value 
      PointType i;
      do {
        i = I;
        I = ThisType::nvalue(i);
        ++count;
      } while (kernel.distsq(I,i) > modeEpsilon && count < maxIterations) ;
#ifdef CONVERGETOMODE_DEBUG
      std::cout << "< " <<std::setw(5) << count << " > ";
#endif
      return I;
    }


    /**
     * Compute the new location of the vector. 
     * To compute mean shift, subtract position vector pt.
     */
    inline PointType nvalue(PointType pt) {
      return kernel.nvalue(pt, at.begin(), at.end()); 
    }
    inline RealType fvalue(PointType pt) {
      return kernel.fvalue(pt, at.begin(), at.end()); 
    }
  };



  /**
   * Implements basic mean shift operations with weight function
   */
  template<class Kernel_, class PtContType_, class WtContType_>
  class BasicWtMeanshift {
  public:
    typedef Kernel_                                 Kernel;
    typedef typename Kernel::RealType               RealType;
    typedef typename Kernel::PointType              PointType;

    typedef PtContType_                             PtContType;
    typedef WtContType_                             WtContType;

    typedef BasicWtMeanshift<Kernel, PtContType, WtContType>  ThisType;

    // variables private
  protected:
    /// reference to weight vector 
    WtContType const &wt;

    /// reference to position vector
    PtContType const &at;

    /// smooth kernel to use
    Kernel const &kernel;

    RealType const modeEpsilon;

    const int           maxIterations;

  public:
    // Constructors
    BasicWtMeanshift(WtContType const &wt_, PtContType const &at_,
                     Kernel const &kernel_, RealType const modeEpsilon_ = 1e-4,
                     const int maxIterations_ = 20)
      : wt(wt_), at(at_), kernel(kernel_), modeEpsilon(modeEpsilon_), 
        maxIterations (maxIterations_)
    {}

    inline PointType shiftToMode(PointType I) {
      int count = 0;

      // store previous value 
      PointType i;
      do {
        i = I;
        I = ThisType::nvalue(i);
        ++count;
#ifdef CONVERGETOMODE_DEBUG
        std::cout << "< " <<std::setw(5) << count << ": from " << i << " to " << I << " > " << std::endl;
#endif
      } while ( kernel.distsq(I,i) > modeEpsilon 
                && count < maxIterations) ;
#ifdef CONVERGETOMODE_DEBUG
      std::cout << "< " <<std::setw(5) << count << " > " << std::endl;
#endif
      return I;
    }


    /**
     * Compute the new location of the vector. 
     * To compute mean shift, subtract position vector pt.
     */
    inline PointType nvalue(PointType pt) {
      return kernel.nvalue(pt, at.begin(),at.end(),wt.begin()); 
    }
    inline RealType fvalue(PointType pt) {
      return kernel.fvalue(pt, at.begin(),at.end(),wt.begin()); 
    }
  };


  // General-purpose Meanshift, with nvalue and fvalue methods
  // delegated to an implementation class (so that we can allow for
  // different optimizations).  Normally in C++, you'd declare the
  // fvalue and nvalue methods as pure virtual, then implement them in
  // a subclass.  The mechanism used here avoids the virtual method
  // lookup and allows for inlining.  fvalue and nvalue are inner-loop
  // calculations.
  template<class Kernel_, class PtContType_, class WtContType_,
           class ValueImpl_>
  class GenericMeanshift {
  public:
    typedef Kernel_                                 Kernel;
    typedef typename Kernel::RealType               RealType;
    typedef typename Kernel::PointType              PointType;
    typedef ValueImpl_                              ValueImpl;

    typedef PtContType_                             PtContType;
    typedef WtContType_                             WtContType;

    typedef GenericMeanshift<Kernel, PtContType, 
                             WtContType, ValueImpl> ThisType;
    // Mode type
    typedef std::list<PointType>                    Cluster;
    typedef std::list<Cluster>                      ClusterList;

  protected:
    /// reference to weight vector 
    WtContType const &wt;
    /// reference to position vector
    PtContType const &at;

    /// smooth kernel to use
    Kernel const &kernel;

    /// number of elements in at vector
    int numelem;

    /// stores the meanshift vector;
    std::vector<PointType> ms; 
    /// stores the distance vector to mode for each point
    std::vector<PointType> tomode;
    RealType const modeEpsilon;
    int      const maxIterations;

  public:
    GenericMeanshift(WtContType const &wt_, PtContType const &at_,
                     Kernel const &kernel_, RealType const modeEpsilon_ = 1e-4,
                     const int maxIterations_ = 20)
      : wt(wt_),at(at_), kernel(kernel_), numelem((int)at_.size()),
        ms(numelem), tomode(numelem), modeEpsilon(modeEpsilon_), 
        maxIterations(maxIterations_)
    {
      // compute meansift at all position vectors
      std::transform(at.begin(), at.end(), ms.begin(), 
                     op_mem_fun(this,(&ThisType::nvalue)));

      // shift to Mode i.e. distance vector to mode
      std::transform(ms.begin(), ms.end(), tomode.begin(),
                     op_mem_fun(this,(&ThisType::shiftToMode)));

      // compute the shifted amount
      std::transform(ms.begin(), ms.end(), at.begin(),
                     ms.begin(), std::minus<PointType>());
    }


    template<class ModeCont>
    inline void getModes(ModeCont& mode, RealType const epsilon) {
      // fill up clusters
      for (typename std::vector<PointType>::const_iterator pt = tomode.begin();
           pt != tomode.end(); ++pt) {
        if (std::find_if(mode.begin(),mode.end(), 
                         FindMode(kernel, *pt, epsilon)) == mode.end()) {
          mode.push_back(*pt);
        }
      }
    }


    template<class ModeCont, class ValueCont>
    void getModes(ModeCont& mode, ValueCont& value, RealType const epsilon) {
      getModes(mode,epsilon);
      value.resize(mode.size());
      std::transform(mode.begin(), mode.end(), value.begin(), 
                     op_mem_fun(static_cast<ValueImpl*>(this),
                                (&ValueImpl::fvalue)));
    }


    //      
    //      inline PtContType const &meanshift() const {
    //          return ms;
    //      }

    //      inline PtContType const &convergedpoints() const {
    //          return tomode;
    //      }
    //      


    inline PointType shiftToMode(PointType I) {
      int count = 0;

      // store previous value 
      PointType i;
      do {
        i = I;
        I = nvalue(i);
        ++count;
      } while ( kernel.distsq(I,i) > modeEpsilon && count < maxIterations) ;
#ifdef CONVERGETOMODE_DEBUG
      std::cout << "< " <<std::setw(5) << count << " > ";
#endif
      return I;
    }

    inline PointType nvalue(PointType pt) {
      return static_cast<ValueImpl&>(*this).nvalue(pt);
    }

    inline RealType fvalue(PointType pt) {
      return static_cast<ValueImpl&>(*this).fvalue(pt);
    }

  private:
    struct FindMode {
      FindMode(Kernel    const &kernel, 
               PointType const  p, 
               RealType  const  epsilon) :
        kernel(kernel), p(p), epsilon(epsilon*epsilon)
      {}

      bool operator()(const PointType m) const {
        return kernel.distsq(p,m) < epsilon;
      }
      Kernel    const &kernel;  // distance kernel
      PointType const  p;       // mode location
      RealType  const  epsilon; // max dist to be considered a match
    };
  };


  template<class Kernel_, class PtContType_, class WtContType_>
  class Meanshift : 
    public GenericMeanshift<Kernel_, PtContType_, WtContType_,
                            Meanshift<Kernel_, PtContType_, WtContType_> >
  {
  public:
    typedef Meanshift<Kernel_, PtContType_, WtContType_>              This;
    typedef GenericMeanshift<Kernel_, PtContType_, WtContType_, This> Parent;
    typedef Kernel_                                                   Kernel;
    typedef typename Kernel::RealType                                 RealType;
    typedef typename Kernel::PointType                                PointType;

    Meanshift(WtContType_ const &wt_, PtContType_ const &at_,
              Kernel_ const &kernel_, RealType const modeEpsilon_=1e-4,
              int const maxIterations_=20)
      : Parent(wt_, at_, kernel_, modeEpsilon_, maxIterations_) {}

    /**
     * Compute the new location of the vector. 
     * To compute mean shift, subtract position vector pt.
     */
    inline PointType nvalue(PointType pt) {
      return Parent::kernel.nvalue(pt, Parent::at.begin(), Parent::at.end(), 
                                   Parent::wt.begin());
    }

    inline RealType fvalue(PointType pt) {
      return Parent::kernel.fvalue(pt, Parent::at.begin(), Parent::at.end(), 
                                   Parent::wt.begin());
    }
  };

  // Tweak on Meanshift that assumes that the wt and at containers
  // have begin and end methods that take a point.  This allows those
  // containers to select a subset of their data to be evaluated
  // (making the computation faster).
  template<class Kernel_, class PtContType_, class WtContType_>
  class WindowedMeanshift : 
    public GenericMeanshift<Kernel_, PtContType_, WtContType_,
                            WindowedMeanshift<Kernel_, PtContType_, WtContType_> >
  {
  public:
    typedef WindowedMeanshift<Kernel_, PtContType_, WtContType_>      This;
    typedef GenericMeanshift<Kernel_, PtContType_, WtContType_, This> Parent;
    typedef Kernel_                                                   Kernel;
    typedef typename Kernel::RealType                                 RealType;
    typedef typename Kernel::PointType                                PointType;

    WindowedMeanshift(WtContType_ const &wt_, PtContType_ const &at_,
                      Kernel_ const &kernel_, RealType const modeEpsilon_=1e-4,
                      int const maxIterations_=20)
      : Parent(wt_, at_, kernel_, modeEpsilon_, maxIterations_) {}
    
    inline PointType nvalue(PointType pt) {
      // Expects an enhanced PtContType_ that gives all data that have
      // a non-negligible effect on pt.
      return Parent::kernel.nvalue(pt, Parent::at.begin(pt), 
                                   Parent::at.end(pt), Parent::wt.begin(pt));
    }
    
    inline RealType fvalue(PointType pt) {
      // Expects an enhanced PtContType_ that gives all data that have
      // a non-negligible effect on pt.
      return Parent::kernel.fvalue(pt, Parent::at.begin(pt), 
                                   Parent::at.end(pt), Parent::wt.begin(pt));
    }
  };
}

#endif // #ifndef _LEAR_MEANSHIFT_H_


