#ifndef LEAR_PERSPECTIVE_H
#define LEAR_PERSPECTIVE_H

#include <vector>
#include <string>

#include <blitz/tinyvec.h>
#include <blitz/tinymat.h>

#include <lear/cvision/iprocessor.h>
#include <lear/interface/windetect.h>

namespace lear {
  /** Utility class for handling perspective projection math as it relates to
  *  pedestrian detection.  
  *
  *  Parts of this code assumes that the camera left vector is roughly 
  *  parallel to the ground plane.  worldZ=0 is the ground plane.  We 
  *  assume a right-handed coordinate system for the camera and for the 
  *  world.  The optical axis of the camera points down the positive Z 
  *  axis.
  */
  class PerspectiveProjector {
  public:
    typedef lear::IProcessor::RealType        RealType;
    typedef blitz::TinyVector<int,2>          IndexType;
    typedef lear::IProcessor::GrayImage       GrayImage;   // Array<float,2>

    // Dimension indices for various matrices and vectors
    static int const X = 0;
    static int const Y = 1;
    static int const Z = 2;
    static int const W = 3; // homogenous dim for 3d points
    static int const S = 2; // homogenous dim for 2d points

  public:
    PerspectiveProjector(
      IndexType                       const imageExtent, 
      IndexType                       const winsize, 
      int                             const borderPixels,     
      RealType                        const pedestrianHeight, 
      blitz::TinyVector<RealType,3>   const rodrigues, 
      blitz::TinyVector<RealType,3>   const translation, 
      blitz::TinyVector<RealType,2>   const focalLength);

    PerspectiveProjector(std::string const &calibFile,
      IndexType                      const winsize,
      int                            const borderPixels,     
      RealType                       const pedestrianHeight);

    // Takes the top row of a detector window (in original image coordinates) 
    // and returns the image scale (new/orig) at which a detector window
    // should be ideally sized to detect a person along that image row.
    RealType scaleFromTopRow(RealType topRow, float baseScale) const; 

    // Takes the location of the top of a person's head and computes the
    // corresponding detector window (x,y,width,height,scale).  
    DetectedRegion detectedRegionFromHeadTip(int x, int y, float baseScale) const;

    // Finds all detector windows in a labeled image corresponding to non-zero
    // pixel locations in (headtips & mask).
    void detectedRegionsFromHeadTips(std::vector<DetectedRegion> &dets,
                                     GrayImage const &headtips,
                                     GrayImage const &mask) const;
    void detectedRegionsFromHeadTips(std::vector<DetectedRegion> &dets,
                                     GrayImage const &headtips) const;

    IndexType imageExtent() const { return imageExtent_; }
    IndexType winsize()     const { return winsize_; }

  private:
    IndexType                       imageExtent_;      // Original image size, on which the calibration was made
    IndexType                       winsize_;          // winsize[Y] = borderPixels*2 + person height in pixels
    int                             borderPixels;      // winsize[Y] = borderPixels*2 + person height in pixels
    RealType                        pedestrianHeight; 
    blitz::TinyVector<RealType,2>   focalLength;

    blitz::TinyMatrix<RealType,3,4> proj;  // full projective transform (world to image)
    blitz::TinyMatrix<RealType,4,4> w2c;   // just world to camera (no perspective, no image plane xform)
    blitz::TinyMatrix<RealType,4,4> c2w;   // camera to world
    blitz::TinyMatrix<RealType,3,3> Rinv;  // for rotating camera rays to be world rays
    blitz::TinyVector<RealType,3>   camPosWorld;

    void init(
      IndexType                       const imageExtent, 
      IndexType                       const winsize, 
      int                             const borderPixels,     
      RealType                        const pedestrianHeight, 
      blitz::TinyVector<RealType,3>   const rodrigues, 
      blitz::TinyVector<RealType,3>   const translation, 
      blitz::TinyVector<RealType,2>   const focalLength);
  };

};

#endif
