/**
 * @file
 * @brief Precision Histogram utility
 */


#ifndef _LEAR_PRECISION_HISTOGRAM_H_
#define _LEAR_PRECISION_HISTOGRAM_H_

#include <lear/cvision/histogram.h>
#include <lear/cvision/interpolate.h>

namespace lear {

  /**
   * Create precision histogram i.e. use linear interpolation between
   * histogram bins. This implies that template HType must be real 
   * typed value, e.g. float or double.
   */
  template<class HType, class VType, int VDim = 1> 
  class PrecisionHistogram : public Histogram< 
    PrecisionHistogram<HType, VType, VDim>, HType, VType, VDim > 
  {
  public:
    typedef Histogram< 
      PrecisionHistogram<HType, VType, VDim>, HType, VType, VDim > 
    Parent;
    typedef typename Parent::HistElementType    HistElementType;
    typedef typename Parent::IndexType          IndexType;
    typedef typename Parent::ValueType          ValueType;
    typedef typename Parent::FloatType          FloatType;
    typedef typename Parent::PrecisionType      PrecisionType;
    typedef typename Parent::HistType           HistType; 
    typedef typename Parent::RealType           RealType;
    typedef typename Parent::DistType           DistType;
    typedef typename HistType::iterator         iterator;
    typedef typename HistType::const_iterator   const_iterator;
    typedef typename Parent::ImplementorType    HistogramType;
    typedef typename Parent::ImplementorType    ImplementorType;
    
    typedef blitz::TinyVector<bool, VDim>           BoolType;
    typedef blitz::TinyVector<int, VDim>            TVIndexType;
    typedef blitz::TinyVector<FloatType, VDim>      TVPrecisionType;
        
    inline PrecisionHistogram(ValueType const &lbound_,
                              ValueType const &extent_,
                              ValueType const &bandwidth_,
                              BoolType const &warp_ = false)
      :   
      Parent(lbound_,extent_,bandwidth_),
      warp_(warp_), onehalf_(0.5), tvbin_(Parent::bin_)
    {}

    template <class HTypeIndexType>
    inline PrecisionHistogram(ValueType const &lbound_,
                              ValueType const &extent_,
                              ValueType const &bandwidth_,
                              BoolType const &warp_,
                              const HTypeIndexType hextent_)
      :   
      Parent(lbound_,extent_,bandwidth_,hextent_),
      warp_(warp_), onehalf_(0.5), tvbin_(Parent::bin_)
    {}
    /**
     * Makes a copy of the histogram element. Note the objects is deep copied,
     * i.e. its the array type is not shared.
     */
    PrecisionHistogram(PrecisionHistogram const &o)/// Copy Construcotr
      : Parent(o), warp_(o.warp_), onehalf_(o.onehalf_), tvbin_(o.tvbin_)
    {}

    inline PrecisionHistogram(std::string const &filename) {
      lear::BiIStream from(filename.c_str());
      if (!from) {
        ThrowException("Unable to open histogram file " + filename);
      }
      bload(from);
    }

    /// Precision fill histogram at value with weight (default 1)
    void push(const ValueType value, 
              const HistElementType weight = 1) 
    {
      using namespace blitz;
      using namespace std;
      // GED: I've rewritten the loops because the original implementation 
      // used extremely inefficient memory access and allocation patterns.
      // The old code dominated the entire classify_rhog execution time 
      // according to one test.  The new code is about 15x faster.
      // Note: static vars are *NOT* threadsafe.  We'll need to use 
      // thread-local storage if we ever go multithreaded.
      PrecisionType   static pI; // precision index
      TVIndexType     static fI; // floor index
      TVIndexType     static cI; // ceil index
      TVPrecisionType static d;  // delta
      BoolType        static lowervalid; 
      BoolType        static uppervalid;
      for (int i=0; i<VDim; i++) {
        FloatType const pi = pI[i] = Parent::halfBin_[i] - onehalf_ +  
          (value[i] - Parent::cenBound_[i])/Parent::bandwidth_[i];
        // GED: the original code omitted the floor.  This causes the
        // static_cast to perform a ceiling operation for negative
        // inputs (which do occur).
        int fi = static_cast<int>(floor(pi));
        int ci = static_cast<int>(ceil(pi));
        d[i] = pi-fi;

        if (warp_[i]) {
          cI[i] = ci % tvbin_[i];
          fI[i] = fi % tvbin_[i];
          if (fI[i] < 0) fI[i] += tvbin_[i];
          if (cI[i] < 0) cI[i] += tvbin_[i];
          uppervalid[i] = lowervalid[i] = true;
        } else {
          cI[i] = ci; 
          fI[i] = fi;
          uppervalid[i] = (ci<tvbin_[i]-onehalf_ && ci>=-onehalf_);
          lowervalid[i] = (fi>=-onehalf_ && fi<tvbin_[i]-onehalf_);
        }
      }
        
      Interpolate<VDim>::linear(Parent::histogram_,weight,d,fI,cI,
                                lowervalid,uppervalid);
    }

    /// Convert value to histogram index
    inline PrecisionType toindex(ValueType const &value) const {
      return Parent::halfBin_ - onehalf_ +  
        (value - Parent::cenBound_)/Parent::bandwidth_;
    }
    inline ValueType tovalue(PrecisionType const &index) const {
      return (index - Parent::halfBin_ + onehalf_)
        *Parent::bandwidth_ + Parent::cenBound_;
    }

    inline BiIStream& bload(BiIStream& in) {
      Parent::bload_common(FileHeader("PHistgrm",100),in);

      in >> warp_;
      onehalf_=0.5; 
      tvbin_=Parent::bin_;
      return in;
    }

    inline BiOStream& bsave(BiOStream& out) const {
      Parent::bsave_common(FileHeader("PHistgrm",100), out);
      out << warp_;
      return out;
    }
    void print(std::ostream& o ) const {
      o << "PrecisionHistogram::" << std::endl;
      Parent::print(o);
      o << "Warp " << warp_ << ", OneHalf " << onehalf_ << ", TinyVectorBin " << tvbin_ << std::endl;
    }
  protected:
    PrecisionHistogram& operator=(PrecisionHistogram const &o){
      Parent::operator=(o);
      if (this != & o) {
        warp_=o.warp_; 
        onehalf_=o.onehalf_;
        tvbin_=o.tvbin_;
      }
      return *this;
    }


    BoolType  warp_;
  private:
    FloatType     onehalf_;
    TVIndexType   tvbin_;
  };

  template<class HType, class VType, int VDim> 
  std::ostream& operator<<(std::ostream& o, const PrecisionHistogram<HType, VType, VDim>& h ) {
    h.print(o);
    return o;
  }

} // namespace lear

#endif // _LEAR_PRECISION_HISTOGRAM_H_

