/** 
* @file
* @brief Defines various image input/output operations.
*/

#ifndef _LEAR_IMAGEIO_H_
#define _LEAR_IMAGEIO_H_

#include <string>

#include <blitz/array.h>
#include <blitz/tinyvec.h>

#include <lear/exception.h>

/* standard headers */
#include <lear/windowsystem.h>

namespace lear {

  namespace detail {
    
    template<class T>
    struct Format {
      enum {depth = 8*4}; /// no. of pallet colors 
      typedef T Type;

      inline static void apply() {
        imlib_image_set_irrelevant_alpha(1);
        imlib_image_set_has_alpha(0);
        imlib_image_set_irrelevant_border(1);
        // do nothing
      }
    };

    template<>
    struct Format<bool> {
      enum {depth = 1};
      typedef bool Type;
      inline static void apply() {
#ifndef WIN32
        imlib_image_set_irrelevant_alpha(1);
        imlib_image_set_has_alpha(0);
        imlib_image_set_irrelevant_border(1);
#endif
      }
    };

    template<class T>
    struct Format<blitz::TinyVector<T,3> > {
      enum {depth = 8*3};
      typedef T Type;
      inline static void apply() {
#ifndef WIN32
        imlib_image_set_irrelevant_alpha(1);
        imlib_image_set_has_alpha(0);
        imlib_image_set_irrelevant_border(1);
#endif
      }
    };

    template<class T>
    struct Format<blitz::TinyVector<T,2> > {
      enum {depth = 8*2};
      typedef T Type;
      inline static void apply() {
#ifndef WIN32
        imlib_image_set_irrelevant_alpha(0);
        imlib_image_set_has_alpha(1);
        imlib_image_set_irrelevant_border(1);
#endif
      }
    };

    template<class T>
    struct Format<blitz::TinyVector<T,4> > {
      enum {depth = 8*4};
      typedef T Type;
      inline static void apply() {
#ifndef WIN32
        imlib_image_set_irrelevant_alpha(0);
        imlib_image_set_has_alpha(1);
        imlib_image_set_irrelevant_border(1);
#endif
      }
    };
  }

  /**
  * @brief Image Input/Output struct
  *
  * ImageIO uses imlib2 for image input output on linux and the WIN32
  * API on Windows.
  */
  struct ImageIO {

    static void setCache(int cacheSizeInMBs) {
#ifndef WIN32
      imlib_set_cache_size(cacheSizeInMBs*1024 * 1024);
#endif
    }

    
    /**
    * @defgroup read_Functions Overloaded ImageIO::read functions
    * @brief read image from file
    * 
    * Read image from file and returns bltiz::Array. read automatically reads
    * the data that best fits in blitz::Array element. For e.g.
    *
    * if blitz::Array element (i.e. template T) is
    *  - int A grayscale image is read
    *  - double A grayscale image is read
    *  - TinyVector<double/int,3> A color rgb image is read
    *  - TinyVector<double/int,4> A color rgba image is read
    */
    
    template<class T, int N>
    static inline blitz::Array<T,N> read(std::string const &imagename) {
#ifndef WIN32
      // Read a file into image object
      // an image handle 
      Imlib_Image image = imlib_load_image(imagename.c_str());;
      // if the load was successful
      if (image) {
        // set the image we loaded as the current context image to work on
        imlib_context_set_image(image);
      } else {
        ThrowException("Unable to read image: " + imagename);
      }
      detail::Format<T>::apply();
      return read<T,N>();
#else
      currImage.Destroy();
      CString n(imagename.c_str());
      HRESULT hr = currImage.Load(n);
      if (SUCCEEDED(hr)) {
        return read<T,N>();
      } else {
        ThrowException("Unable to read image: " + imagename);
      }
#endif
    }

    template<class T, int N>
    static inline void read(std::string const &imagename, blitz::Array<T,N>& imageData) {
      imageData.reference(read<T,N>(imagename));
    }
    
    

    
    /**
    * @brief write blitz::Array as image to file
    * 
    * Automatically writes an image in the format best described 
    * by blitz::Array element (provided the image format supports it).
    * For e.g.
    *
    * if blitz::Array element (i.e. template T) is
    *  - int A grayscale image is written
    *  - double A grayscale image is written
    *  - TinyVector<double/int,3> A color rgb image is written
    *  - TinyVector<double/int,4> A color rgba image is written
    *
    */
    template<class T> static inline
    void write(std::string const &imagename, blitz::Array<T,2> const &imageData) {
#ifndef WIN32
      int const columns = imageData.columns();
      int const rows    = imageData.rows();

      Imlib_Image image = imlib_create_image(rows, columns);
      if (image) {
        // set the image we loaded as the  current context image to work on
        imlib_context_set_image(image);
      } else {
        ThrowException("Unable to create image in memory");
      }

      DATA32* data = imlib_image_get_data();

      // copy data to array
      for (int i=0, k=imageData.lbound(0); i<rows; k++,i++) {
        for (int j=0, l=imageData.lbound(1); j<columns; l++,j++) {
          data[j*rows+i] = toColor(imageData(k,l));
        }
      }

      imlib_image_put_back_data(data);
      detail::Format<T>::apply();
#else
      int const columns = imageData.extent()[0];
      int const rows    = imageData.extent()[1];

      HRESULT hr = currImage.Create(columns, -rows, 32/*, createAlphaChannel*/);
      if (!SUCCEEDED(hr)) {
        ThrowException("Unable to create image in memory");
      }
      for (int yo=0, yi=imageData.lbound(1); yo<rows; yo++,yi++) {
        for (int xo=0, xi=imageData.lbound(0); xo<columns; xo++,xi++) {
          currImage.SetPixel(xo,yo, toColor(imageData(xi,yi)));
        }
      }
#endif

      save(imagename);
    }
    

#ifdef WIN32
    template<class T> static inline
    void write(std::string const &imagename, blitz::Array<T,3> const &imageData) {
      int const columns = imageData.extent()[0];
      int const rows    = imageData.extent()[1];

      HRESULT hr = currImage.Create(columns, -rows, 32/*, createAlphaChannel*/);
      if (!SUCCEEDED(hr)) {
        ThrowException("Unable to create image in memory");
      }
      for (int yo=0, yi=imageData.lbound(1); yo<rows; yo++,yi++) {
        for (int xo=0, xi=imageData.lbound(0); xo<columns; xo++,xi++) {
          currImage.SetPixelRGB(xo,yo, 
            imageData(xi,yi,0), imageData(xi,yi,1), imageData(xi,yi,2));
        }
      }
      // A blank image is always created (at least on Vista in 64-bit mode).  
      // Reason unknown.
      save(imagename);
    }
#endif

    /// Return the format of this image, currently uses image extension
    static std::string imageFormat(std::string const &imagename) ;

    /// Return the size of the image width,height
    static blitz::TinyVector<int,2> size(std::string const &imagename) ;

  protected:
    // Assumes BGR encoding on little-endian machines
    static inline DATA8 red(DATA32 c) {
      return static_cast<DATA8>((c & 0x00FF0000)>>16);
    }
    static inline DATA8 green(DATA32 c) {
      return static_cast<DATA8>((c & 0x0000FF00)>> 8);
    }
    static inline DATA8 blue(DATA32 c) {
      return static_cast<DATA8>((c & 0x000000FF));
    }
    static inline DATA8 gray(DATA32 c) {
      return static_cast<DATA8>((0.3*red(c) + 0.59*green(c) + 0.11*blue(c)));
    }
    static inline DATA8 alpha(DATA32 c) {
      return static_cast<DATA8>((c & 0xFF000000)>>24);
    }
    
    template<class T>
    static inline void fromColor(const DATA32 c, bool& val) {
      val = c ? true : false;
    }

    template<class T>
    static inline void fromColor(const DATA32 c, T& val) {
      val = gray(c);
    }

    // Gray and alpha channel
    template<class T>
    static inline void fromColor(const DATA32 c, blitz::TinyVector<T,2>& val) {
      val = gray(c), alpha(c);
    }

    template<class T>
    static inline void fromColor(const DATA32 c, blitz::TinyVector<T,3>& val) {
      val = red(c), green(c), blue(c);
    }

    template<class T>
    static inline void fromColor(const DATA32 c, blitz::TinyVector<T,4>& val) {
      val = red(c), green(c), blue(c), alpha(c);
    }
    

    
    static inline DATA32 toColor(bool const color) {
      return color ? 0xFFFFFF: 0;
    }

    template<class T>
    static inline DATA32 toColor(const T color) {
      const DATA8 c = static_cast<DATA8> (color); 
      DATA32 cc = 0;
      cc = 0xFF000000 | (c <<16) | (c << 8) | c;
      return cc;
    }

    // Gray and alpha channel
    template<class T>
    static inline DATA32 toColor(const blitz::TinyVector<T,2> color) {
      const DATA8 c = static_cast<DATA8> (color[0]); 
      const DATA8 a = static_cast<DATA8> (color[1]); 
      DATA32 cc = (a<<24) |  (c <<16) | (c << 8) | c;
      return cc;
    }

    template<class T>
    static inline DATA32 toColor(const blitz::TinyVector<T,3> color) {
      DATA32 cc = 0;
      cc =(0xFF000000) |
        (static_cast<DATA8> (color[0]) <<16) | 
        (static_cast<DATA8> (color[1]) << 8) | 
        (static_cast<DATA8> (color[2]));
      return cc;
    }

    template<class T>
    static inline DATA32 toColor(const blitz::TinyVector<T,4> color) {
      DATA32 cc = 0;
      cc =(static_cast<DATA8> (color[3]) <<24) | 
        (static_cast<DATA8> (color[0]) <<16) | 
        (static_cast<DATA8> (color[1]) << 8) | 
        (static_cast<DATA8> (color[2]));
      return cc;
    }
    

    
    /*! \brief read imlib2 image and return blitz array.
    */
    template<class T,int N>
    static inline blitz::Array<T,N> read() {
#ifndef WIN32
      int const width  = imlib_image_get_width();
      int const height = imlib_image_get_height();
#else
      int const width  = currImage.GetWidth();
      int const height = abs(currImage.GetHeight());
#endif 

      // resize data
      blitz::Array<T,N> imageData = 
        (N >= 3) ? 
        blitz::Array<T,N>(width, height, 3) :
        blitz::Array<T,N>(width,height);

#ifndef WIN32
      DATA32* data = imlib_image_get_data_for_reading_only();
      // copy data to array
      for (int i=0; i<width; i++) {
        for (int j=0; j<height; j++) {
          fromColor(data[j*width+i], imageData(i,j));
        }
      }
#else
      switch (currImage.GetBPP()) {
        case 24:
          for (int y=0; y<height; y++) {
            // For now, we get the pixel address for each row.  This makes it so
            // we don't have to worry about pitch and bottom-up vs. top-down.
            unsigned char const *px = 
              (unsigned char const *)currImage.GetPixelAddress(0,y);
            for (int x=0; x<width; x++, px+=3) {
              // BGR->RGB done here
              fromColor((px[2]<<16) | (px[1]<<8) | (px[0]), imageData(x,y));
            }
          }
          break;
  
        case 8:
          for (int y=0; y<height; y++) {
            // For now, we get the pixel address for each row.  This makes it so
            // we don't have to worry about pitch and bottom-up vs. top-down.
            unsigned char const *px = 
              (unsigned char const *)currImage.GetPixelAddress(0,y);
            for (int x=0; x<width; x++, px++) {
              imageData(x,y) = x;
            }
          }
          break;

        case 1:
          for (int y=0; y<height; y++) {
            // For now, we get the pixel address for each row.  This makes it so
            // we don't have to worry about pitch and bottom-up vs. top-down.
            unsigned char const *px = 
              (unsigned char const *)currImage.GetPixelAddress(0,y);
            for (int x=0; x<width; px++) {
              for (int bit=0; bit<8 && x<width; x++, bit++) {
                imageData(x,y) = ((*px >> bit) & 0x01);
              }
            }
          }
          break;

        default:
          ThrowException("Unsupported image bit depth: " << currImage.GetBPP());
      }
#endif

#ifndef WIN32
      imlib_free_image();
#else
      currImage.Destroy();
#endif
      return imageData;
    }

    template<class T, int N>
    static inline void read(blitz::Array<T,N>& imageData) {
      imageData.reference(read<T,N>());
    }

    static void save(std::string const &imagename);

#ifdef WIN32
    static ATL::CImage currImage;
#endif
  };

} // namespace lear

#ifdef None // defined in X11/X.h
#undef None
#endif
#ifdef GrayScale // defined in X11/X.h
#undef GrayScale
#endif

#endif // #ifdef _LEAR_IMAGEIO_H_

