/*
 * =====================================================================================
 * 
 *       Filename:  windetect.h
 * 
 *    Description:  Groups all common options for windows based object
 *    detection.
 * 
 *        Version:  1.0
 *        Created:  Sunday 03 June 2007 12:24:12  IST IST
 *       Revision:  none
 *       Compiler:  gcc
 * 
 *         Author:  Navneet Dalal
 *        Company:  Lear/ INRIA.
 * 
 * =====================================================================================
 */

#ifndef  _LEAR_WIN_DETECT_H_
#define  _LEAR_WIN_DETECT_H_

#include <list>
#include <vector>
#include <string>
#include <iostream>
#include <memory>

#include <boost/shared_ptr.hpp>

#include <lear/image/inimageiter.h>

#include <lear/cvision/classify.h>
#include <lear/cvision/windescriptor.h>

#include "../lib/cfg.h"
#ifdef HAVE_TRACKER
#include <tracker/JTracker.h>
#include <tracker/TrkOutputStream.h>
#include <lear/classifier/tracker_ppresult.h>
#endif

#ifdef WIN32
// suppress precision loss warnings for now
#pragma warning(push)
#pragma warning(disable:4305)
#endif

#define BUILD_APP

// Set required RHOG Dense parameters in an object of this class.
struct RHOGDenseParam {
  typedef float        RealType;

  enum NormalizerFlags {
    NormL2Hys=0,
    NormL1Sqrt,
    NormL2,
    NormL1,
    NoNorm
  };

  enum PreprocessorFlags {
    RGB_Grad=0, // Use gradient of RGB Image
    Lab_Grad, // Use gradient of Lab Image
    RGB_Sqrt_Grad, // Use gradient of Sqrt(RGB)
    Lab_Sqrt_Grad, // Use gradient of Sqrt(Lab)
    RGB_Log_Grad   // Use log(RGB+1)
  };

  int                 cellsize_x, cellsize_y;
  int                 numcell_x, numcell_y;
  int                 descstride_x, descstride_y;
  int                 orientbin;
  bool                semicirc; 
  RealType            wtscale;
  RealType            gscale;
  RealType            epsilon;
  RealType            maxvalue;

  /// Normalization to use
  PreprocessorFlags                   preprocessing2use;
  NormalizerFlags                     norm2use;

  RHOGDenseParam():
    cellsize_x(8), cellsize_y(8),
    numcell_x(2), numcell_y(2),
    descstride_x(8), descstride_y(8),
    orientbin(9), semicirc(true), wtscale(2), 
    gscale(0), epsilon(1), maxvalue(0.2),
    preprocessing2use(RGB_Sqrt_Grad), norm2use(NormL2Hys)
  { }
};

/**
 * This is the main class. It provides the initialization routine
 * which given an pointer to RHOGDenseParam, initializes the RHOG
 * Dense descriptor computation. The computefeature method can then be
 * used to compute HOG descriptors at any given location.
 *
 * Use it if you are interested in computing HOG features at specified
 * locations.
 */
struct WinDetect {
  // ===================================
  //------ Type definitions -------
  // ===================================

  // All calculations done in realtype precision.
  typedef RHOGDenseParam::RealType       RealType;

  typedef blitz::TinyVector<int,2>       IndexType;

  typedef blitz::TinyVector<RealType,3>  RodriguesVector;
  typedef blitz::TinyVector<RealType,3>  TranslationVector;
  typedef blitz::TinyVector<RealType,2>  FocalLength;

  // ===================================
  // ----- Function declarations -------
  // ===================================

  // No need to change any variable if you intend to call default
  // person detection.
  WinDetect() :
    endscale(0), scaleratio(1.05), startscale(1),
    size_x(64), size_y(128), winstride_x(8), winstride_y(8),
    finestride_x(-1), finestride_y(-1), quantizepos(false),
    topleft_x(-1), topleft_y(-1), fullsize_x(-1), fullsize_y(-1),
    label(DefaultLabel), verbose(0), cachesize(16)
#ifdef HAVE_CUDA
    , useCuda(true)
#endif
  { } 

  virtual ~WinDetect() {}

  /**
   * Creates and initializes rhogdense descriptors from provided parameters
   *
   * The RHOGDenseParam* may be destroyed immediately after this call.
   */
  virtual void init(RHOGDenseParam const *param);
  /**
   * Same as above. We use std::vector as the code supports
   * simultaneous computation of multiple RHOGDense descriptors each
   * with different parameter configurations, i.e. one can compute
   * different descriptors on the same test location.
   *
   * The vector may be destroyed immediately after this call.
   *
   * NOTE: The initialization must be called only once. Currently there is
   * no support for dynamically changing the RHOG parameters.
   */
  virtual void init(const std::vector<RHOGDenseParam const *>& param);

  /**
   * Preprocesses an image and computes the HOG descriptor. The descriptor
   * is set in the 'result' pointer.
   *
   * Return value is 'true' if descriptor computation was successful, else
   * its false.
   *
   * NOTE 1: User is expected to pre-allocate the 'result' pointer with its 
   * length at least as long as the output of featurelength() 
   *
   * NOTE 2: This is an inefficient way to compute descriptors. If you need
   * to compute features at several location at once, use the vector version
   *
   * Image data is assumed to be in row first format via unsigned
   * char* pointer.  The data must be RGB format.
   *
   * The i,j th pixel is read as
   * unsigned char *pixel = image+(j*step + i)*3*sizeof(unsigned char);
   * unsigned char R = pixel[0];
   * unsigned char G = pixel[1];
   * unsigned char B = pixel[2];
   *
   * Step of 0 implies use same value as width.
   */
  bool computefeature(float* result, const int xloc, const int yloc,
                      const unsigned char* image, int width, int height, int step=0) const;

  /**
   * Return value is a std::vector of same size as xlocs and ylocs,
   * with value set to true if descriptor computation succeeded else
   * it is set to false.
   *
   * NOTE 1: xloc size and yloc size must be same.
   *
   * NOTE 2: result vector must be preallocated with its size same as xloc.
   * Each element in the vector must also be preallocated with at least as much
   * memory as returned by the featurelength
   *
   * xloc and yloc specify the top-left corner. If location is found to be
   * out of bounds, then return value is false.
   *
   * Step of 0 implies use same value as width.
   */
  std::vector<bool> computefeature(std::vector<float*>& result, 
                                   std::vector<int> const &xloc, std::vector<int> const &yloc,
                                   const unsigned char* image, int width, int height, int step=0) const;


  size_t featurelength() const;

  virtual void print(std::ostream& o) const ;
  // ========================================================================
  // --- Variable declarations. Common to both feature dump and classify ----
  // ========================================================================

  // All variables are public on purpose. Feel free to change as you feel like.
  // However refrain from changing variables after the init calls.

  // Maximum number of input images to be processed? (use -1 for no limit)
  int maximages;

  /// Scale space parameters
  RealType scalePrecision, endscale, scaleratio, startscale;

  // window size. for e.g. in person detection I use 64x128 pixels
  int size_x, size_y;

  // HOG blocks are stride width apart.
  int winstride_x, winstride_y;

  // Spacing for where HOG blocks are calculated.  Eventually,
  // anything that's a factor of GCD(winstride,blockstride) should be
  // fine.  For now (svn revision 780), it must be the same as
  // winstride when a perspective slider is used, and it must be 1
  // when a pyramid slider is used.
  int finestride_x, finestride_y;

  // When true, allows the WinDescriptor to round detector window
  // locations to nearby grid locations to increase speed and/or
  // reduce memory requirements.
  bool quantizepos;

  // Ignore these. For internal use
  int topleft_x, topleft_y;

  // Ignore these. For internal use
  int fullsize_x, fullsize_y; 

  // Ignore these. For internal use
  char label; 

  /// verbose level. Between [0-9]. 9 is most verbose.
  int verbose;

  /// Cache size. Reducing cache size can significantly hurt the
  /// run-time. Keep it to at least its default value. Size in MBs.
  /// Cannot be a size_t unless blitz is changed to use size_t
  /// indices/array lengths.
  int cachesize;

  // Params used for perspective sliders only      
  int                 borderPixels;
  RealType            pedestrianHeight;
  RodriguesVector     rodrigues;
  TranslationVector   translation;
  FocalLength         focalLength;

#ifdef HAVE_CUDA
  bool useCuda;
#endif

  std::vector<boost::shared_ptr<RHOGDenseParam> > params;

  static const char DefaultLabel;
};

#ifdef BUILD_APP
class JitterWindow;
#ifdef HAVE_CUDA
#  include <pdcuda/imresize.h>
#  include <pdcuda/arrays.h>
#endif

/**
 * This class is for internal use. It is used to dump HOG blocks to a binary
 * file which can then be converted via dump4svmlearn to binary svm_learn
 * format.
 *
 * Need not bother about this class if in the library version of classify_rhog
 */
struct WinDetectDump : public WinDetect {
  // ===================================
  // ----- Function declarations -------
  // ===================================

  // No need to change any variable if you intend to call default person detection.
  WinDetectDump() :
    WinDetect(), 

    // dump feature options
    pyramid(false), 
    samples(1), 
    randomSeed(false),
    dumpfulldetail(0),

    hardscore(0),
    memorylimit(0),
    poscases(0),
    negcases(0),
    jitterstep_x(1), jitterstep_y(1), 
    jitterwinstep(0)

  {
  }

  void writeWinDesc(lear::InputImageIterator &infiles,// filenames to compute descriptors on in each line
                    std::string const &outfile,// Output file to dump
                    std::string const &testlocs// if specified, dump test locations to this file. 
                    );

  void writePresetExamples(std::string const &hardfile,// hard examples file
                           std::string const &outfile, // output file to dump
                           bool suppressNonJittered);

  virtual void print(std::ostream& o) const ;
  // ===================================
  //------ Variable declarations. Used during feature dump -------
  // ===================================
  bool pyramid;
  size_t samples;
  bool randomSeed;
  int dumpfulldetail;
  RealType hardscore;
  int memorylimit;
  int poscases;
  int negcases;
  int jitterstep_x, jitterstep_y;
  int jitterwinstep;
};

inline std::ostream& operator<<(std::ostream& o, WinDetectDump const &windet) 
{ windet.print(o); return o; }

#endif

struct DetectedRegion {
  float score, scale;
  int x, y, width, height;

  DetectedRegion() : 
    score(0), scale(0), x(0), y(0), width(0), height(0)
  {}

  DetectedRegion(float const score, float const scale,
                 int const x, int const y, int const w, int const h)
    : 
    score(score), scale(scale), x(x), y(y), width(w), height(h) 
  {}

  void print(std::ostream& o) const;
};
std::ostream& operator<<(std::ostream& o, DetectedRegion const &region);

/**
 * Detect objects in test images.
 */
struct WinDetectClassify : public WinDetect
{
  WinDetectClassify() :
    WinDetect(), 
    initialsize_x(-1), initialsize_y(-1),
    avsize_x(0), avsize_y(96),
    margin_x(4), margin_y(4),
    alignmargin_x(4), alignmargin_y(4),
    fullstride_x(-1), fullstride_y(-1),
    sliderType(1), no_nonmax(false),
    aligninimage(false), showscore(true), showbestscores(false),
    softmax(0), threshold(0.1), lightthreshold(0),
    nonmaxsigma_x(8), nonmaxsigma_y(16), nonmaxsigma_scale(1.3),
    score2prob_a(1), score2prob_b(0)
  {}

  inline virtual void init(RHOGDenseParam const *param) {
    WinDetect::init(param);// this one calls below function. which in turns calls initclassifier.
  }

  inline virtual void init(const std::vector<RHOGDenseParam const *>& param) {
    WinDetect::init(param);
    initclassifier();
  }

  /**
   * Run any test only after calling init.
   *
   * NOTE 1: 'detections' is cleared in each call. So it contains only fresh
   * detections on this image.
   *
   * NOTE 2: The detected region bounds may go out of image bounds.
   *
   * NOTE 3: sliderType, no_nonmax, showscore suppression options have no 
   * effect. This code always run non-maximum suppression on scale-space.
   *
   * Step of 0 implies use same value as width.
   */
  void test(lear::LinearClassify const &classifier, 
            std::list<DetectedRegion>& detections,
            const unsigned char* imagedata, int width, int height, 
            int step=0) const;

#ifdef BUILD_APP
  /** 
   * This is for internal use. Binary application functionality is coded in this.
   */
  void runImageSlider(lear::LinearClassify const &classifier, 
                      lear::InputImageIterator   &infiles,
                      std::string          const &maskfile,
                      std::string          const &outfile,    
                      bool                 const  writeListFilename,
                      std::string          const &outimage,   
                      std::string          const &outhist, 
                      std::string          const &falsetxt,   
                      std::string          const &testlocs// if specified, dump test locations to this file. 
                      )  const;
#endif

  virtual void print(std::ostream& o) const ;

  // ===================================
  // ------- Other options used during classification or testing
  // ===================================

  int initialsize_x, initialsize_y;

  // Set the height/width of object on average.
  // For e.g. in person detection even though the window length is 128, average person is 
  // about 96 pixel tall. So its default value is 96.
  // It is used only to take care of boundary cases as so often objects are close to boundaries.
  // Average size and margin are used to expand input image and then we run classifier on it.
  int avsize_x, avsize_y;

  // Specify the margin size w.r.t. average size.
  // Formula of extension is max(margin_x/avsize_x*width, margin_y/avsize_y*height) on each side.
  int margin_x, margin_y;

  // Use to align input w.r.t. classifier max. Test code. Ignore it.
  int alignmargin_x, alignmargin_y;

  // Test code. Ignore it.
  int fullstride_x, fullstride_y;

  // controls the scale space pyramid type used (if any).  See
  // WinDetectClassifyMain::setClassifyParam for details.
  static int const NO_SLIDER          = 0;
  static int const PYRAMID_SLIDER     = 1;
  static int const PERSPECTIVE_SLIDER = 2;
  int sliderType;

  // if true, do not perform non-maximum suppression
  bool no_nonmax; 

  // test code. ignore it.
  bool aligninimage; 

  bool showscore;
  bool showbestscores;

  int softmax;

  // Final threshold after non-maximum threshold.
  RealType threshold;

  // Only those locations which are above this value, are passed to non-maximum suppression stage.
  RealType lightthreshold; 

  // sigma for non-maximum suppression. 
  float nonmaxsigma_x, nonmaxsigma_y, nonmaxsigma_scale;

  float score2prob_a, score2prob_b;

#if HAVE_TRACKER
  // Frame-to-frame tracking.  Requires the MIT CSAIL APP tracker
  // library.
  std::string trkfile;
  std::string trkformat;
  double fps;
  int maxtrk, maxcc;
  double ivxy, ivdxy, iva;
  double maxvxy, maxvdxy, maxva;
  double minvxy, minvdxy, minva;
  double addThreshold, delThreshold, searchTime;
  double pLearningRate, vLearningRate, aLearningRate;
#endif

protected:
  void initclassifier() ;

#ifdef HAVE_TRACKER
  lear::Tracker_PPResult *createTrackerPPResult(IndexType const &initialExt,
                                                double fpsFromVid) const;
#endif
};

inline std::ostream& operator<<(std::ostream& o, WinDetectClassify const &windet) 
{ windet.print(o); return o; }

#ifdef WIN32
#pragma warning(pop)
#endif

#endif // 

