#include <iomanip>
#include <iostream>
#include <lear/exception.h>
#include <lear/util/fileutil.h>
#include <lear/image/painter.h>

#include <boost/filesystem/path.hpp>
#include <boost/filesystem/operations.hpp>
namespace fs=boost::filesystem;

#include <lear/classifier/markimage.h>

using namespace std;

typedef lear::MarkImage::Color  Color;
typedef lear::MarkImage::Colors Colors;

static inline Colors defaultIfNeeded(Colors const &c) {
  if (c.size()) return c;
  Colors cc;
  cc.push_back(Color(255,255,0));
  cc.push_back(Color(255,0,255));
  return cc;
}

lear::MarkImage::MarkImage(std::string const &outimage_,
                           bool const outfileIsDir_,
                           bool const hidescore_,
                           Colors borderColors)
                           : PostProcessResult(),
                           outimage_(outimage_),
                           outfileIsDir_(outfileIsDir_),
                           hidescore_(hidescore_),
                           bordercolors_(defaultIfNeeded(borderColors))
{}

bool lear::MarkImage::verifyImageOut(std::string const outimage_,
                                     bool const outfileIsDir_) 
{
  bool doImageOut_ = (!outimage_.empty());
  using namespace std;
  // check if outimage is valid
  if (doImageOut_ && outfileIsDir_) {
    fs::path outdir(outimage_);
    if (fs::exists(outdir)) {
      // make sure that outfile is also a directory
      if (!fs::is_directory(outdir)) {
        cerr<< "WARNING:: 'infile' is a directory/list, "
          " thus 'outimage'=" << outimage_ 
          <<"' must also be a directory."
          << "Disabling image output." << endl;
        doImageOut_ = false;
      }
    } else {
      // it does not exist, create a directory
      fs::create_directory(outdir);
      if (!fs::exists(outdir)) {
        cerr<< "WARNING:: Unable to create output image directory "
          << outimage_ << "Disabling image output." << endl;
        doImageOut_ = false;
      } else {
        cerr<< "WARNING:: Directory " << outdir.string() << 
          " does not exist. Creating it." << endl;
      }
    }
  }
  return doImageOut_;
}

namespace somegarbagename {
  class DrawBoundingBox {
  public:
    typedef blitz::TinyVector<int,2>        IndexType;

    //DrawBoundingBox() {}
    DrawBoundingBox(std::string const &infile, std::string const &outfile) :
      painter(infile), outfile_(outfile)
    {}

    DrawBoundingBox(lear::IProcessor::RGBImage const &inimage, 
                    std::string                const &outfile) :
      painter(inimage), outfile_(outfile)
    {}

    void saveto(std::string const &outfile) {
      outfile_ = outfile;
    }

    // box draw
    void drawBox(IndexType const &lbound, IndexType const &extent,
      Colors const &c) {
        lear::Rectangle<int> r(lbound[0]+1, lbound[1]+1,
          extent[0]-2, extent[1]-2);

        painter.noBrush();
        for (int i=0; i<(int)c.size(); i++) {
          painter.setColor(c[i](0), c[i](1), c[i](2));
          painter.drawRect(r-i);
        }
    }

    // text draw
    void drawLabel(IndexType const &xy, std::string const &txt) {
      int const padding = 2; // text box background padding
      int width  = 54; 
      int height = 15;
      painter.getTextSize(txt, width, height);
      painter.setBrush(0, 0, 0);
      painter.drawRect(xy[0],xy[1], width+padding*2, height+padding*2);
      painter.noBrush();
      painter.setColor(255, 255, 255);
      painter.drawText(txt, xy[0]+padding, xy[1]+padding);
    }

    ~DrawBoundingBox() {
      if (!outfile_.empty()) {
        painter.save(outfile_);
      }
    }

  private:
    lear::Painter painter;
    std::string outfile_;
  };
}

void lear::MarkImage::write(const_iterator s, const_iterator e, 
                            lear::PostProcessResult::RGBImage const &inimage,
                            std::string                       const &infile)
{
  std::vector<const_iterator> sv; sv.push_back(s);
  std::vector<const_iterator> ev; ev.push_back(e);
  std::vector<Colors>         cv; cv.push_back(bordercolors_);
  write(sv, ev, cv, inimage, infile);
}

void lear::MarkImage::write(std::vector<const_iterator>       const &sv, 
                            std::vector<const_iterator>       const &ev,
                            std::vector<Colors>               const &cv, 
                            lear::PostProcessResult::RGBImage const &inimage,
                            std::string                       const &infile) 
{
  std::string outfile = outimage_;
  if (outfileIsDir_) {
    if (lear::extension(infile).find(":") != string::npos) {
      // Filenames from videos are currently tagged with a trailing
      // frame number (e.g. "foo.avi:123456789").  In the future, a
      // cleaner system for passing along frame numbers should be
      // used.
      string ext = lear::extension(infile);
      outfile += "/" + lear::basename(infile) + "." + 
        ext.substr(ext.find(":")+1) + ".jpg";
    } else {
      outfile += "/" + lear::basename(infile) +".jpg";
    }
  }
  somegarbagename::DrawBoundingBox marker(inimage, outfile);

  assert(sv.size() == ev.size());
  assert(sv.size() == cv.size());

  char scoretxt[24];
  for (size_t i=0; i<sv.size(); i++) {
    const_iterator s = sv[i];
    const_iterator e = ev[i];
    Colors  const &c = cv[i];
    while (s != e) {
      marker.drawBox(s->lbound, s->extent, c);
      if (!hidescore_) {
        sprintf(scoretxt,"%7.4f",s->score);
        marker.drawLabel(s->lbound, scoretxt);
      }
      ++s;
    }
  }
}

