#include <lear/classifier/ms_processresult.h>
#include <cmath>
#include <vector>
#include <cstdlib>
#include <iomanip>
#include <iostream>
#include <vector>
#include <math.h>
#include <lear/exception.h>
#include <lear/cvision/meanshift.h>
#include <lear/blitz/ext/domainiter.h>

#ifdef WIN32
# define isnan _isnan
#endif

using namespace std;

template<class RealType_,int N>
struct DensityKernel {
  typedef RealType_                       RealType;
  typedef blitz::TinyVector<RealType,N>   PointType;

  DensityKernel(const PointType sigma) : 
    sigma(sigma) 
    //norm (std::pow(2*lear::MathConst<RealType>::PI,
    //static_cast<RealType>(N/2.0)))
  {}

  template<class Pt, class Wt, class Ref>
  inline PointType nvalue(Ref rf, Pt first, Pt last, Wt wt) const {
    using namespace std;
    PointType numer(.0);
    PointType denom(.0);

    for (;first != last; ++first, ++wt) {
      PointType x = *first;
      PointType r = rf;
      PointType ns = sigma;
      ns[0] *= exp(x[2]); ns[1] *= exp(x[2]); // x[2] is scale

      x/=ns;
      r/=ns;
      // These are right equations. But I realised this mistake quite late.
      // Keep the old code as parameters have been tuned w.r.t. that.
      //RealType w = (*wt)*exp(-dot(x-r,x-r)/2)/product(ns);

      //numer += w*x/ns;
      //denom += w/(ns*ns);

      RealType w = (RealType)((*wt)*exp(-dot(x-r,x-r)/2)/sqrt(product(ns)));

      numer += w*x;
      denom += w/ns;
    }
    numer /= denom;
    return numer;
  } 

  template<class Pt, class Wt, class Ref>
  inline RealType fvalue(Ref rf, Pt first, Pt last, Wt wt) const {
    RealType numer=0;

    for (;first != last; ++first, ++wt) {
      PointType x = *first;
      PointType ns = sigma;
      ns[0] *= exp(x[2]); ns[1] *= exp(x[2]); // x[2] is scale

      x -= rf;
      x /= ns;

      //numer+=(*wt)*std::exp(-dot(x,x)/2)/product(ns);
      numer += (RealType)((*wt)*std::exp(-dot(x,x)/2)/std::sqrt(product(ns)));
    }
    return numer;
  } 

  RealType distsq(PointType a, PointType b) const {
    PointType ns = sigma;
    ns[0] *= exp(b[2]);
    ns[1] *= exp(b[2]);
    b -= a;
    b /= ns;
    return blitz::dot(b,b);
  }

protected:
  PointType sigma;
  //     RealType  norm;
};

lear::MeanShift_ProcessResult::
MeanShift_ProcessResult(IndexType const extent_, RealType const threshold_,
                        RealType const finalthreshold_, 
                        SigmoidType const score2prob_, PointType const sigma_,
                        int const transfunc, int const verbose)
  : Parent(threshold_), extent_(extent_), finalthreshold_(finalthreshold_),
    sigma_(sigma_), verbose(verbose)
{
  assert(!isnan(sigma_[0]) && !isnan(sigma_[1]) && !isnan(sigma_[2]));
  switch (transfunc) {
  case 1:
    sigmoid=new Sigmoid<RealType>(score2prob_[0],score2prob_[1]);
    break;
  case 2:
    sigmoid=new SoftMax<RealType>(score2prob_[0],score2prob_[1]);
    break;
  case 0:
  default:
    sigmoid=new HardMax<RealType>(score2prob_[0],threshold_);
    break;
  };
}


// Simple implementation of a container for usage with
// WindowedMeanshift.  All containers for a GenericMeanshift need to
// hold all data points.  Additionally, WindowedMeanshift allows the
// container to prune out data that is irrelevant to the DensityKernel
// computations.
// 
// This implementation grids up the input space into a set of
// partially-redundant bins.  For any query point, we can easily find
// its bin and that bin will contain all of the relevant input data.
//
// An alternative implementation might use a kd tree or an octree
// internal representation instead of a uniform grid.  This
// implementation is easier to write and it's fast enough for now
// (~10-20x faster than the original non-windowed version for scenes
// with hundreds of pedestrians).
template <typename L_, typename D_>
class WindowedData {
public:
  static int const N = L_::numElements;
  typedef D_                       Datum;
  typedef L_                       Loc; // range is [lbound,rbound]
  typedef blitz::TinyVector<int,N> Subscripts; // range is [0,nbins]

  typedef typename vector<Datum>::iterator        Iterator;
  typedef typename vector<Datum>::const_iterator  ConstIterator;
  
  // TODO: use blitz arrays or perhaps a shared_ptr / shared_array
  // instead of a raw vector to avoid copying data?
  WindowedData(WindowedData const &rhs) 
    : lbound(rhs.lbound), ubound(rhs.ubound), 
      bandwidthInBins(rhs.bandwidthInBins), nbins(rhs.nbins), 
      data(rhs.data), rawdata(rhs.rawdata) {}

  WindowedData(Loc lbound, Loc ubound, Loc bandwidth)
    : lbound(lbound), ubound(ubound), 
      nbins(blitz::max(1, Subscripts(blitz::floor(2*(ubound-lbound)/bandwidth)))),
      bandwidthInBins(bandwidth*nbins/(ubound-lbound)), data(sub2ind(nbins)+1) 
  {
    // nbins is chosen above based on some heuristics that try to
    // balance minimizing the amount of duplicated data while also
    // minimizing the amount of superfluous data in a bin.
    assert(blitz::sum(nbins >= 1) == N);
  }

  inline void insert(Loc loc, Datum datum) {
    if (blitz::sum(loc < lbound) || blitz::sum(loc > ubound)) {
      ThrowException("location out of range");
    }
    Loc        const bin = loc2bin(loc);
    Subscripts const l   = blitz::max(blitz::floor(bin - bandwidthInBins), Subscripts(0));
    Subscripts const u   = blitz::min(blitz::ceil(bin + bandwidthInBins), nbins);
    if (N != 3) {
      ThrowException("only 3D locations are supported right now");
    } 
    for (int x=l[0]; x<=u[0]; x++) {
      for (int y=l[1]; y<=u[1]; y++) {
        for (int s=l[2]; s<=u[2]; s++) {
          Subscripts const bin(x,y,s);
          data[sub2ind(bin)].push_back(datum); 
        }
      }
    }
    rawdata.push_back(datum);
  }

  size_t size() const { return rawdata.size(); }

  ConstIterator begin() const { return rawdata.begin(); }
  ConstIterator end()   const { return rawdata.end();   }
  Iterator      begin()       { return rawdata.begin(); }
  Iterator      end()         { return rawdata.end();   }

  ConstIterator begin(Loc loc) const { return getbin(loc).begin(); }
  ConstIterator end(Loc loc)   const { return getbin(loc).end();   }
  Iterator      begin(Loc loc)       { return getbin(loc).begin(); }
  Iterator      end(Loc loc)         { return getbin(loc).end();   }
  
private:
  inline vector<Datum> const &getbin(Loc loc) const {
    return data[sub2ind(blitz::floor(loc2bin(loc)))];
  }

  inline vector<Datum> &getbin(Loc loc) {
    return data[sub2ind(blitz::floor(loc2bin(loc)))];
  }

  inline Loc loc2bin(Loc const &loc) const { 
    return blitz::max(blitz::min((loc - lbound) / (ubound - lbound) * nbins,
                                 nbins-1), 0); 
  }

  inline int sub2ind(Subscripts const &idx) const {
    int ind    = 0;
    int stride = 1;
    for (int i=0; i<N; i++) {
      // column major
      ind    += idx[i]*stride; 
      stride *= nbins[i]; // we really have nbins+1 bins 
    }
    return ind;
  }

  Loc        const lbound;
  Loc        const ubound;
  Subscripts const nbins;
  Loc        const bandwidthInBins; // units are bins
  vector<vector<Datum> > data;
  vector<Datum> rawdata;
};


// FUTURE: allow the user to choose dynamically whether to use
// windowed meanshift or not.
#define USE_WINDOWED

void lear::MeanShift_ProcessResult::doit()
{
  using namespace lear;
  using namespace std;

  if (!detect_.empty()) {
    PointType nsigma = sigma_;
    nsigma[2] = log(sigma_[2]);

    typedef DensityKernel<RealType,3>        KernelType;

#ifdef USE_WINDOWED
    // find the bounding box that covers all data
    typedef blitz::TinyVector<RealType,3> Loc;
    Loc lbound( std::numeric_limits<RealType>::max());
    Loc ubound(-std::numeric_limits<RealType>::max());
    for (const_iterator i=detect_.begin(); i!=detect_.end(); ++i) {
      blitz::TinyVector<RealType,2> cen = i->lbound + i->extent/2.0f;
      for (int j=0; j<IndexType::numElements; j++) {
        if (cen[j] < lbound[j]) lbound[j] = cen[j];
        if (cen[j] > ubound[j]) ubound[j] = cen[j];
      }
      RealType const scale = std::log(i->scale);
      if (scale < lbound[2]) lbound[2] = scale;
      if (scale > ubound[2]) ubound[2] = scale;
    }

    // stuff data into a grid-based data structure
    WindowedData<Loc,PointType> at(lbound, ubound, nsigma*2.5f); // TODO: make configurable
    WindowedData<Loc,RealType>  wt(lbound, ubound, nsigma*2.5f); // TODO: make configurable
    for (const_iterator i=detect_.begin(); i != detect_.end(); ++i) {
      PointType const cen(i->lbound[0] + i->extent[0]/2.0f, 
                          i->lbound[1] + i->extent[1]/2.0f, 
                          std::log(i->scale));
      at.insert(cen, cen);
      wt.insert(cen, (*sigmoid)(i->score));
    }

    typedef lear::WindowedMeanshift< KernelType,
      WindowedData<Loc,PointType>, WindowedData<Loc,RealType> >  MeanShiftType;
#else
    vector<PointType> at(numdetection_); // (centerX, centrY, log(scale)) for each detection
    vector<RealType > wt(numdetection_); // score for each detection
    {
      unsigned index = 0;
      for (const_iterator i=detect_.begin(); i != detect_.end(); ++i,++index) {
        wt[index] = (*sigmoid)(i->score);
        blitz::TinyVector<RealType,2> cen = i->lbound + i->extent/2.0f;
        at[index] = PointType(cen[0],cen[1],std::log(i->scale));
      }
    }

    typedef lear::Meanshift< KernelType,
      vector<PointType>, vector<RealType> >  MeanShiftType;
#endif

    KernelType kernel(nsigma);
    MeanShiftType ms(wt, at, kernel,1e-5f,100);

    vector<PointType> mode; vector<RealType> value;
    ms.getModes(mode, value, 1);

    detect_.clear();

    for (unsigned i=0; i< mode.size(); ++i) {
      typedef blitz::TinyVector<RealType,2> Real2DType;

      RealType scale = exp(mode[i][2]);
      Real2DType c(mode[i][0],mode[i][1]);
      Real2DType s(extent_*scale);

      DetectInfo d(static_cast<float>(value[i]),
                   static_cast<float>(scale),
                   blitz::ceil(c - s/2), blitz::floor(s));

#ifdef CONVERGETOMODE_DEBUG
      cout << "Mode " << d << endl;
#endif
      if (value[i] > finalthreshold_) {
        detect_.push_back(d);
      }
    } 

    if (verbose >= 6) {
      cout << "  " << mode.size() << " detections survived non-maxima suppression" << endl;
    }
  }
}
