#include <iostream> 
#include <fstream> 
#include <sstream>
#include <list>
#include <lear/cmdline.h>

using namespace lear;
using namespace std;
using namespace boost::program_options;

lear::Cmdline::Cmdline(
        string configfile_,
        bool const helpOnNoArgs_):  
    desc("Allowed options",WARP_WIDTH),
    descCmdline("Command line options only",WARP_WIDTH),
    descConfigfile("Options common to config file and command line",WARP_WIDTH),
    descHidden(WARP_WIDTH), descVisual(WARP_WIDTH),
    argument("Required Arguments", WARP_WIDTH),
    helpOnNoArgs(helpOnNoArgs_),
    configfile(configfile_)
{
  descCmdline.add_options()
    ("help,h", "show help screen")
    ("version", "print version number and contact details")
    ;

  descConfigfile.add_options() // allow config files that include others
    ("configfile",option<string>(&configfile)
        ->defaultValue(configfile.c_str()), 
        "options configuration file name")
    ;
}



lear::Cmdline::status lear::Cmdline::parse(int argc, char** argv){
  if (commandName_.empty()) commandName_ = argv[0];
  constructUsageString(argv[0]);

  desc  // Options descriptions used for parsing
    .add(descConfigfile) // public configfile + command line args
    .add(descCmdline)    // public commandline-only args
    .add(descHidden)     // hidden args
    .add(argument.desc); // Positional arguments

  descVisual // Options descriptions used for printing usage (no descHidden)
    .add(descConfigfile)
    .add(descCmdline)
    .add(argument.desc);

  // Options descriptions for config files (note descCmdline and argument.desc 
  // are omitted)
  options_description descConfig(WARP_WIDTH);
  descConfig 
    .add(descConfigfile)
    .add(descHidden);

  if (helpOnNoArgs && argc==1 && configfile.empty()) {
    printUsage();
    return failed;
  }

  try {
    // Highest priority to command-line.
    store(command_line_parser(argc, argv)
      .options(desc).positional(argument.position)
      .run(), vm);

    // Recursively read config files
    // TODO: this doesn't seem to work right with the Debian boost 1.33.
    // I don't know yet whether it's a usage bug or a problem with that
    // version of boost. parse_config_file seems to be a nop.
    while (true) {
      // In the current implementation, each parsing has a chance to 
      // overwrite the configfile variable.  We keep checking it until 
      // we come up empty.
      string cfg = configfile;
      if (vm.count("configfile")) {
        cfg = vm["configfile"].as<string>();
        vm.erase("configfile");
      }
      configfile = "";

      // Parse the file now
      if (cfg.empty()) {
        // Done with loop if no config file specified
        break;
      } else {
        ifstream from(cfg.c_str());
        if (from) {
          // Add new ones to the end of the list... first parsed = top priority
          store(parse_config_file(from, descConfig), vm);
        } else {
          // unable to open config file. Show user a message.
          cerr << "Warning: Unable to open configuration file: "
               << configfile << endl;
          break;
        }
      }
    }

    // notify results
    notify(vm);


    if (vm.count("help")) {
      printUsage();
      return info;
    }

    if (vm.count("version")) {
      printVersion();
      return info;
    }

    // flag error for missing arguments
    for (vector<string>::const_iterator ai = argument.args.begin();
      ai != argument.args.end(); ++ai) 
    {
      if (vm.count(*ai)< 1) {
        throw validation_error(string("missing required argument ") 
          + argument.desc.find(*ai, false).format_name());
      }
    }
#ifdef __MY_BOOST__
  } catch (validation_error& e) {
    string mesg(e.what());
#else
  } catch (validation_error&) {
    string mesg("during validation");
#endif 
    mesg.append("\n\tplease check values and bounds of all options\n\t");
    printError(mesg);
    printUsage(cout);
    return failed;
  } catch (error& e) {
    printError(e.what());
    printUsage(cout);
    return failed;
  }
  return ok;
}



void lear::Cmdline::constructUsageString(string const &name) {
    usageString_ = "Usage: ";
    usageString_.append(name).append(" [options] ");

    for (vector<string>::const_iterator ai = argument.args.begin();
            ai != argument.args.end(); ++ai) 
    {
        usageString_.append(" ").append(*ai).append(" ");
    }
    usageString_.append("\n");
    warptext(usageString_, WARP_WIDTH);
}



void lear::Cmdline::printUsage(ostream& o) const {
    o << commandName_ << '\n';
    const size_t length = commandName_.length();

    for (size_t i=0; i<length; ++i) o << "=";
    o << '\n';

    if (!brief_.empty()) {
        o << "\t" << brief_ << "\n\n";
    } else {
        o << '\n';
    }

    if (!description_.empty()) {
      o << "Description: " << "\n" << description_ << "\n\n";
    }

    printOptions(o);
}

void lear::Cmdline::printOptions(ostream& o) const {
  if (!usageString_.empty()) o << usageString_;

  output(o,descVisual);
  if (!usageIssues_.empty() ) {
    o << "\nUsage Issues:\n" << usageIssues_ << '\n';
  }
  if (!configfile.empty()) {
    string s= 
"\n"
"This command also accepts configuration file. " 
"First, all options on command line are read, then all options present in "
"configuration file are read.  Configuration files may include other "
"configuration files.  If there are still some undefined options, the "
"program substitutes their default values.  Command-line parameters override "
"those in config files.  For recursive config files, the first config file "
"loaded that sets a value takes precedence over others.  You can also specify "
"configuration file name on command line using 'configfile' option.  At this "
"time, a config file can at most include a single other config file.\n";
    o << warptext(s,WARP_WIDTH);
  }

}

void lear::Cmdline::printVersion(ostream& o) const {
    o << version() << '\n';
}

void lear::Cmdline::printError(string const &mesg, ostream& o) const {
    o << "Error "  << mesg << "\n\n";
}

//  namespace {
//      void format_one(
//              std::ostream& os,
//              option_description const &opt, 
//              unsigned first_column_width) 
//      {
//          std::ostringstream ss;
//          ss << "  " << opt.format_name() << ' ' << opt.format_parameter();
//      
//          os << ss.str();

//          if (!opt.description().empty()) {
//              for(int pad = first_column_width - ss.str().size(); pad > 0; --pad) {
//                  os.put(' ');
//              }
//              os << " : " << opt.description();
//          }
//  //          boost::shared_ptr<TypedOption> sem = 
//  //              dynamic_pointer_cast<const TypedOption> (opt.semantic());

//  //          if (!sem.defaultValue().empty()) 
//  //              os << " ( Default: " << opt.default_value() << " )";
//  //          

//  //          if (optdesc) {
//  //              if (!optdesc->minValue().empty() && !optdesc->maxValue().empty()) 
//  //                  os << " ( Min: " << optdesc->minValue() 
//  //                     << ", Max: " << optdesc->maxValue() << " )";
//  //              else if (!optdesc->minValue().empty()) 
//  //                  os << " ( Min: " << optdesc->minValue() << " )";
//  //              else if (!optdesc->maxValue().empty()) 
//  //                  os << " ( Max: " << optdesc->maxValue() << " )";

//  //              if (optdesc->isRequired() && optdesc->isInternal() ) {
//  //                  os << " ( Required, Internal ) " ;
//  //              } else if (optdesc->isRequired() ) {
//  //                  os << " ( Required ) " ;
//  //              } else if (optdesc->isInternal() ) {
//  //                  os << " ( Internal ) " ;
//  //              }
//  //          }
//      }
//  }

/*
    void format_one(std::ostream& os, option_description const &opt, 
                    unsigned first_column_width, unsigned line_length)
    {
        stringstream ss;
        ss << "  " << opt.format_name();
        
        // Don't use ss.rdbuf() since g++ 2.96 is buggy on it.
        os << ss.str();

        if (!opt.description().empty()) {
            for(unsigned pad = first_column_width - ss.str().size(); 
                pad > 0; 
                --pad)
            {
                os.put(' ');
            }
        
            format_description(os, opt.description(),
                                first_column_width, line_length);
        } 
        string format_parameter = opt.format_parameter();
        if (format_parameter()) {
            os << "\n";
            for(unsigned pad = first_column_width; pad > 0; --pad) {
                os.put(' ');
            }
            os << ' ' << format_parameter;
        }
    }

    void options_description::print(std::ostream& os) const
    {
        if (!m_caption.empty())
            os << m_caption << ":\n";

        // Find the maximum width of the option column //
        unsigned width(16);
        unsigned i; // vc6 has broken for loop scoping
        for (i = 0; i < m_options.size(); ++i) {
            option_description const &opt = *m_options[i];
            stringstream ss;
            ss << "  " << opt.format_name() ;//<< ' ' << opt.format_parameter();
            width = (max)(width, static_cast<unsigned>(ss.str().size()));            
        }
        
        // add an additional space to improve readability //
        ++width;
            
        // The options formatting style is stolen from Subversion. //
        for (i = 0; i < m_options.size(); ++i) {
            if (belong_to_group[i])
                continue;

            option_description const &opt = *m_options[i];

            format_one(os, opt, width, m_line_length);

            os << "\n";
        }

        for (unsigned j = 0; j < groups.size(); ++j) {            
            os << "\n" << *groups[j];
        }
    }

 * */


void lear::Cmdline::output(ostream& os, options_description const &d) const {
    // for time being, call desc output.
    d.print(os);
//      if (!d.caption().empty())
//          os << d.caption() << ":\n";

//      set<string> key = d.primary_keys();
//      unsigned width(24);

//      /* Find the maximum width of the option column */
//      for (set<string>::const_iterator ki=key.begin(); ki != key.end(); ++ki) {
//          option_description const &opt = d.find(*ki);

//          stringstream ss;
//          ss << "  " << opt.format_name() << ' ' << opt.format_parameter();
//          width = max(width, static_cast<unsigned>(ss.str().size()));            
//      }

//      for (set<string>::const_iterator ki=key.begin(); ki != key.end(); ++ki) {
//          option_description const &opt = d.find(*ki);

//          format_one(os, opt, width);

//          os << "\n";
//      }
}



/* Function used to check that 'opt1' and 'opt2' are not specified
at the same time. */
void Cmdline::conflicting_options(char const *opt1, char const *opt2) const {
    if (vm.count(opt1) && !vm[opt1].defaulted() && 
        vm.count(opt2) && !vm[opt2].defaulted())
            throw logic_error(string("conflicting options: '") 
                        + opt1 + "' and '" + opt2 + "'.");
}

/* Function used to check that of 'for_what' is specified, then
'required_option' is specified too. */
void Cmdline::option_dependency(
        char const *for_what, 
        char const *required_option) const
{
    if (vm.count(for_what) && !vm[for_what].defaulted())
        if (vm.count(required_option) == 0 || 
            vm[required_option].defaulted())
                throw logic_error(string("Option '") + for_what 
                        + "' requires option '" + required_option + "'.");
}

ostream& lear::operator<<(ostream& os, lear::Cmdline const &desc) {
    desc.printUsage(os);
    return os;
}

lear::TypedOption<bool>* lear::bool_option() {
    return bool_option(false);
}
lear::TypedOption<bool>* lear::bool_option(bool* v) {
    TypedOption<bool>* r = new TypedOption<bool>(v);
    r->defaultValue(false);
    //r->implicit();

    return r;
}

#ifdef WIN32
// Just assume 80 on Windows, for now
const unsigned lear::Cmdline::WARP_WIDTH = 80;
#else
// Try to get the real info on linux
// Adapted from:
//   http://julipedia.blogspot.com/2005/03/how-to-get-window-size.html
#include <sys/ioctl.h>
unsigned getTtyWidth() {
  struct winsize ws;
  unsigned short height, width;  

  if ((ioctl(STDOUT_FILENO, TIOCGWINSZ, &ws) == -1 &&
       ioctl(STDERR_FILENO, TIOCGWINSZ, &ws) == -1 &&
       ioctl(STDIN_FILENO, TIOCGWINSZ, &ws) == -1) ||
      ws.ws_col == 0) {
        height = 25;
        width = 80;
  } else {
    height = ws.ws_row;
    width = ws.ws_col;
  }
  return width;
}

const unsigned lear::Cmdline::WARP_WIDTH = getTtyWidth();
#endif
