#include <stdio.h>
#include <exception>
#include <stdexcept>

#include <cutil/cutil_inline.h>
#include <cutil/cutil.h>

#include <pdcuda/gradients.h>
#include <pdcuda/math.cuh>
#include <pdcuda/imtransfer.h>

using namespace std;
using namespace pdcuda;

// Here we have several implementations of non-smoothing spatial
// gradient calculators.  Performance numbers are embedded in comments
// in computeSpatialGradients().

#define PI 3.14159265358979323846264338327950288419716939937510

// Assumptions:
//   semicircle orientation is desired
__device__ inline
float oriFromDxDy(float dx, float dy) 
{
  float tmp = atan2f(dy, dx) * 180 / PI;
  // We want semicircle orientations (they are more robust to contrast
  // flips in the imagery, e.g. dark pants on a light background look
  // like light pants on a dark background if we use semicircle
  // orientations).
  if (tmp < 0) tmp += 180;
  if (tmp >= 180) tmp = 0; // handle roundoff errors
  return tmp;
}

// The most straight-forward implementation.  
__global__
void ComputeSpatialGradients_Kernel_Simple
(float *dMag, float *dOri, float const *dRgb, int w, int h)
{
  int const x = (threadIdx.x + blockDim.x*blockIdx.x);
  int const y = (threadIdx.y + blockDim.y*blockIdx.y);

  if (x >= w || y >= h) return;

  float bestDx   =  0;
  float bestDy   =  0;
  float bestMag2 = -1;

  for (int c=0; c<3; c++) {
    // The conditionals here are to handle boundary effects.  To match
    // Dalal's implementation, we use pixel replication.  This should
    // incur a performance cost only for the left and right boundaries
    // due to thread divergence on the GPU.
    float dx = (x > 0)   ? -sqrtf(dRgb[c+(x-1 + y  *w)*3]) : -sqrtf(dRgb[c+(x + y*w)*3]);
    dx +=      (x < w-1) ?  sqrtf(dRgb[c+(x+1 + y  *w)*3]) :  sqrtf(dRgb[c+(x + y*w)*3]);
    float dy = (y > 0)   ? -sqrtf(dRgb[c+(x + (y-1)*w)*3]) : -sqrtf(dRgb[c+(x + y*w)*3]);
    dy +=      (y < h-1) ?  sqrtf(dRgb[c+(x + (y+1)*w)*3]) :  sqrtf(dRgb[c+(x + y*w)*3]);
    float const mag2 = dx*dx + dy*dy;

    if (mag2 > bestMag2) {
      bestMag2 = mag2;
      bestDx   = dx;
      bestDy   = dy;
    }    
  }
  
  dMag[x + y*w] = sqrtf(bestMag2);
  dOri[x + y*w] = oriFromDxDy(bestDx, bestDy);
}

static texture<float, 1, cudaReadModeElementType> texLin;

// Slight adaptation of the simple kernel.  Here we use the texLin
// texture to take advantage of its cache.
__global__
void ComputeSpatialGradients_Kernel_TexLinear
(float *dMag, float *dOri, int w, int h)
{
  int const x = (threadIdx.x + blockDim.x*blockIdx.x);
  int const y = (threadIdx.y + blockDim.y*blockIdx.y);

  if (x >= w || y >= h) return;

  float bestDx  =  0;
  float bestDy  =  0;
  float bestMag = -1;

  for (int c=0; c<3; c++) {
    // The min/max here are to handle boundary effects.
    float const dx = 
      sqrtf(tex1Dfetch(texLin, c+(min(w-1,x+1) + y  *w)*3)) -
      sqrtf(tex1Dfetch(texLin, c+(max(0,  x-1) + y  *w)*3));
    float const dy = 
      sqrtf(tex1Dfetch(texLin, c+(x + min(h-1,y+1)*w)*3)) -
      sqrtf(tex1Dfetch(texLin, c+(x + max(0,  y-1)*w)*3));

    float const mag = sqrtf(dx*dx + dy*dy);

    if (mag > bestMag) {
      bestMag = mag;
      bestDx  = dx;
      bestDy  = dy;
    }    
  }
  
  dMag[x + y*w] = bestMag;
  dOri[x + y*w] = oriFromDxDy(bestDx, bestDy);
}

static texture<float4, 2, cudaReadModeElementType> texCuda;

// Here is a 2D CUDA texture-based implementation.  Like the 1D
// texture implementation, we get to take advantage of the texture
// cache.  As a 2D texture, we get 2D locality and we can let the
// texture unit handle boundary conditions.  Both of these help
// noticeably.  Unfortunately, we also incur a very heavy cost to
// construct the RGBA CUDA array.  Since this is a one-use texture, it's
// not worth it.  See computeSpatialGradients for actual performance
// numbers.
//
// Surprisingly, moving the sqrtf gamma correction into the texture
// generation, we get no appreciable speedup.  This is probably
// because the compute speed on a GT200-class GPU is fast enough that
// we're memory bound.
__global__
void ComputeSpatialGradients_Kernel_Texture2D
(float *dMag, float *dOri, int w, int h)
{
  // With non-interpolated lookups, we don't have to worry about the
  // 0.5 offset.
  int const x = (threadIdx.x + blockDim.x*blockIdx.x);
  int const y = (threadIdx.y + blockDim.y*blockIdx.y);

  float4 const dx   = sqrtf(tex2D(texCuda, x+1, y)) - sqrtf(tex2D(texCuda, x-1, y));
  float4 const dy   = sqrtf(tex2D(texCuda, x, y+1)) - sqrtf(tex2D(texCuda, x, y-1));
  float4 const mag2 = dx*dx + dy*dy;

  float bestMag2;
  float bestDx;
  float bestDy;
  if (mag2.x > mag2.y) {
    if (mag2.x > mag2.z) { bestMag2 = mag2.x; bestDx = dx.x; bestDy = dy.x; }
    else                 { bestMag2 = mag2.z; bestDx = dx.z; bestDy = dy.z; }
  } else {
    if (mag2.y > mag2.z) { bestMag2 = mag2.y; bestDx = dx.y; bestDy = dy.y; }
    else                 { bestMag2 = mag2.z; bestDx = dx.z; bestDy = dy.z; }
  }
  
  dMag[x + y*w] = sqrtf(bestMag2);
  dOri[x + y*w] = oriFromDxDy(bestDx, bestDy);
}

// Takes the RGB data in dRGB and transfers it to texCuda, returning
// the cudaArray to which the texture is bound.  The caller must
// unbind the texture and free the cudaArray when done with it.
__host__
static cudaArray *transferRgbToTex(float const *dRgb, int w, int h)
{
  cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<float4>();
  texCuda.addressMode[0] = cudaAddressModeClamp;
  texCuda.addressMode[1] = cudaAddressModeClamp;
  texCuda.filterMode = cudaFilterModePoint;
  texCuda.normalized = false;    // access with integer texture coordinates

  cudaArray *dCudaArray = rgbToCudaArray2d(dRgb, w, h, false, false);
  cutilSafeCall(cudaBindTextureToArray(texCuda, dCudaArray, channelDesc));

  return dCudaArray;
}

#define USE_GMEM             0
#define USE_CUDA_TEXTURES_2D 1
#define USE_LINEAR_TEXTURES  2

#define KERNEL USE_LINEAR_TEXTURES //USE_CUDA_TEXTURES_2D //USE_GMEM //

__host__
void pdcuda::computeSpatialGradients(float *dMag, float *dOri, 
                                     float const *dRgb, 
                                     int w, int h)
{
  if (!SemiCirc) {
    throw runtime_error("Full-circle gradients are not supported right now");
  }

  dim3 const block(16,4,1); // bigger=worse: texture cache pressure
  dim3 const grid((int)ceil(w/float(block.x)), (int)ceil(h/float(block.y)));

  // Timings below are given for a 3.2GHz Core i7 965 box with a GTX
  // 260+ graphics card, running on Debian Etch.  A 1920x1080 image
  // was used as the input data.  Results are for an average of 50
  // runs after an initial warmup run.
  //
  // On that machine, Dalal's code takes an average of 390.9 ms.
  if (KERNEL == USE_GMEM) {
    // 3.86 ms 
    //
    // NOTE: if we use the min/max trick from the linear texture code,
    // this slows down slightly to 3.93 ms.
    cudaGetLastError(); // reset error status
    ComputeSpatialGradients_Kernel_Simple<<<grid,block>>>(dMag,dOri, dRgb, w,h);
    cutilSafeCall(cudaGetLastError());
  } else if (KERNEL == USE_CUDA_TEXTURES_2D) {
    // 11.74 ms
    //  0.92 ms (if the texture transfer & cleanup are not counted)
    cudaArray *dRgba = transferRgbToTex(dRgb, w, h);
    cudaGetLastError(); // reset error status
    ComputeSpatialGradients_Kernel_Texture2D<<<grid,block>>>(dMag,dOri, w,h); 
    cutilSafeCall(cudaGetLastError());
    cutilSafeCall(cudaUnbindTexture(texCuda));
    cutilSafeCall(cudaFreeArray(dRgba));
  } else if (KERNEL == USE_LINEAR_TEXTURES) {
    // 1.56 ms
    // 1.56 ms (if the bind/unbind are not counted)
    //
    // Note: if we use the conditionals from the gmem approach instead
    // of min/max lookups, this slows down dramatically to 3.37 ms.
    cutilSafeCall(cudaBindTexture(0, texLin, dRgb, w*h*3*sizeof(float)));
    cudaGetLastError(); // reset error status
    ComputeSpatialGradients_Kernel_TexLinear<<<grid,block>>>(dMag,dOri, w,h);
    cutilSafeCall(cudaGetLastError());
    cutilSafeCall(cudaUnbindTexture(texLin));
  }
}
