#include <lear/cvision/perspective.h>
#include <blitz/tinyvec-et.h>
#include <lear/exception.h>
#include <fstream>

using namespace std;
using namespace blitz;
using namespace lear;

typedef PerspectiveProjector::RealType RealType;

template <typename T, int N>
static T l2norm(blitz::TinyVector<T,N> const &v) {
  T n = 0;
  for (int i=0; i<N; i++) {
    n += T(v[i] * v[i]);
  }
  return (T)sqrt(n);
}


PerspectiveProjector::PerspectiveProjector(
  IndexType                       const imageExtent, 
  IndexType                       const winsize, 
  int                             const borderPixels,     // winsize[Y] = borderPixels*2 + person height in pixels
  RealType                        const pedestrianHeight, 
  blitz::TinyVector<RealType,3>   const rodrigues, 
  blitz::TinyVector<RealType,3>   const translation, 
  blitz::TinyVector<RealType,2>   const focalLength)
{
  init(imageExtent, winsize, borderPixels, pedestrianHeight, 
       rodrigues, translation, focalLength);
}


PerspectiveProjector::PerspectiveProjector(std::string const &calibFile,
                                           IndexType   const winsize,
                                           int         const borderPixels,     
                                           RealType    const pedestrianHeight)
{
  RealType nan = numeric_limits<RealType>::quiet_NaN();
  IndexType                     imageExtent(-1,-1);
  blitz::TinyVector<RealType,3> rodrigues(nan,nan,nan);
  blitz::TinyVector<RealType,3> translation(nan,nan,nan);
  blitz::TinyVector<RealType,2> focalLength(nan,nan);

  ifstream in(calibFile.c_str());
  while (!in.eof()) {
    string key;
    in >> key;
    if      (key == "imagewidth")  { in >> imageExtent[0]; }
    else if (key == "imageheight") { in >> imageExtent[1]; }
    else if (key == "om")          { in >> rodrigues[0] >> rodrigues[1] >> rodrigues[2]; }
    else if (key == "T")           { in >> translation[0] >> translation[1] >> translation[2]; }
    else if (key == "f")           { in >> focalLength[0] >> focalLength[1]; }
    char dummy[2048];
    in.getline(dummy, sizeof(dummy));
    if (in.gcount() == 0) break; // getline doesn't trigger eof
  }

  init(imageExtent, winsize, borderPixels, pedestrianHeight, 
       rodrigues, translation, focalLength);
}


void PerspectiveProjector::init(
    IndexType                       const imageExtent, 
    IndexType                       const winsize, 
    int                             const borderPixels,     
    RealType                        const pedestrianHeight, 
    blitz::TinyVector<RealType,3>   const rodrigues, 
    blitz::TinyVector<RealType,3>   const translation, 
    blitz::TinyVector<RealType,2>   const focalLength)
{
  // Adapted from Pietro Perona's code found in the Matlab Camera 
  // Calibration Toolbox.

  this->imageExtent_     = imageExtent;
  this->winsize_         = winsize;
  this->borderPixels     = borderPixels;
  this->pedestrianHeight = pedestrianHeight;
  this->focalLength      = focalLength;

  /// Camera Rotation
  RealType const theta = l2norm(rodrigues);
  static RealType const EPS = (RealType)2e-16;
  if (theta < EPS) {
    w2c = 0;
    for (int i=0; i<4; i++) {
      w2c(i,i) = 1;
    }
  } else {
    RealType const alpha = (RealType)cos(theta);
    RealType const beta  = (RealType)sin(theta);
    RealType const gamma = 1 - alpha;
    blitz::TinyVector<RealType,3> const omega(rodrigues / theta);

    // clc;clear all;
    // syms o0 o1 o2  real;
    // syms alpha beta gamma;
    // omega = [o0;o1;o2];
    // omegav=[0 -omega(3) omega(2); omega(3) 0 -omega(1); -omega(2) omega(1) 0];
    // A = omega*omega';
    // R = eye(3)*alpha + omegav*beta + A*gamma;
    // disp(R);
    // 
    // [     alpha+gamma*o0^2, -beta*o2+gamma*o0*o1,  beta*o1+gamma*o0*o2]
    // [  beta*o2+gamma*o0*o1,     alpha+gamma*o1^2, -beta*o0+gamma*o1*o2]
    // [ -beta*o1+gamma*o0*o2,  beta*o0+gamma*o1*o2,     alpha+gamma*o2^2]
    w2c(0,0) = alpha +                  (omega[0]*omega[0])*gamma;
    w2c(0,1) =         -omega[2]*beta + (omega[1]*omega[0])*gamma;
    w2c(0,2) =          omega[1]*beta + (omega[2]*omega[0])*gamma;

    w2c(1,0) =          omega[2]*beta + (omega[0]*omega[1])*gamma;
    w2c(1,1) = alpha +                  (omega[1]*omega[1])*gamma;
    w2c(1,2) =         -omega[0]*beta + (omega[2]*omega[1])*gamma;

    w2c(2,0) =         -omega[1]*beta + (omega[0]*omega[2])*gamma;
    w2c(2,1) =          omega[0]*beta + (omega[1]*omega[2])*gamma;
    w2c(2,2) = alpha +                  (omega[2]*omega[2])*gamma;

    w2c(3,0) = w2c(3,1) = w2c(3,2) = 0; // no projective yet
    w2c(3,3) = 1; // no scaling
  }

  /// Camera Translation
  for (int r=0; r<3; r++) {
    w2c(r,W) = translation[r];
  }

  /// Simple method for inverting a rigid transform
  // http://www-graphics.stanford.edu/courses/cs248-98-fall/Final/q4.html
  // http://stackoverflow.com/questions/155670/invert-4x4-matrix-numerical-most-stable-solution-needed
  c2w = 0;
  for (int r=0; r<3; r++) {
    for (int c=0; c<3; c++) {
      c2w(r,c) = w2c(c,r);
      c2w(r,W) -= w2c(c,r) * w2c(c,3); // -vector_dot(R[0:3,:], T)
    }
  } 
  c2w(W,W) = 1;

  // camera-to-world rotation
  for (int r=0; r<3; r++) {
    for (int c=0; c<3; c++) {
      Rinv(r,c) = w2c(c,r);
    }
  } 

  /// Camera Perspective
  blitz::TinyMatrix<RealType,3,4> tmpProj;
  for (int r=0; r<3; r++) {
    for (int c=0; c<4; c++) {
      tmpProj(r,c) = w2c(r,c);
    }
  }

  /// Camera-to-image transformation
  blitz::TinyMatrix<RealType,3,3> cam2img;
  cam2img = 0;
  cam2img(X,X) = focalLength[X];
  cam2img(X,S) = imageExtent_[X] / (RealType)2;
  cam2img(Y,Y) = focalLength[Y];
  cam2img(Y,S) = imageExtent_[Y] / (RealType)2;
  cam2img(S,S) = 1;
  proj = blitz::product(cam2img, tmpProj);

  /// Camera position in world coordinates
  blitz::TinyVector<RealType,4> const camPosWorldHomo = blitz::product(
    c2w, blitz::TinyVector<RealType,4>(0,0,0,1));
  camPosWorld[0] = camPosWorldHomo[X] / camPosWorldHomo[W];
  camPosWorld[1] = camPosWorldHomo[Y] / camPosWorldHomo[W];
  camPosWorld[2] = camPosWorldHomo[Z] / camPosWorldHomo[W];
}


RealType PerspectiveProjector::scaleFromTopRow(
  RealType topRow, float baseScale) const
{
  blitz::TinyVector<RealType,3> const dWintopCam( // right-handed, y is down
    0, (topRow - imageExtent_[Y]/2) / focalLength[Y], 1);

  blitz::TinyVector<RealType,3> const dWintopWorld = 
    blitz::product(Rinv, dWintopCam); // ray from the camera focal point to the backprojected window top (in world coordinates)

  RealType const footToWinTopHeightWorld = 
    pedestrianHeight / baseScale *
    (winsize_[Y] - borderPixels) / (winsize_[Y] - 2 * borderPixels);
  // distance on the z axis from the focal point to the backprojected 
  // window top (in world coordinates).  
  RealType const dz = footToWinTopHeightWorld - camPosWorld[Z]; 

  blitz::TinyVector<RealType,4> const winbottomWorld(
    dWintopWorld[X] * dz / dWintopWorld[Z] + camPosWorld[X],
    dWintopWorld[Y] * dz / dWintopWorld[Z] + camPosWorld[Y],
    -pedestrianHeight * borderPixels / winsize_[Y],
    1);

  blitz::TinyVector<RealType,3> const winbottomImg = 
    blitz::product(proj, winbottomWorld);
  RealType const winbottomRow   = winbottomImg[Y] / winbottomImg[S];
  RealType const windowHeightPx = winbottomRow - topRow;

  RealType const scale = windowHeightPx / winsize_[Y];

  return scale;
}

DetectedRegion 
PerspectiveProjector::detectedRegionFromHeadTip(int x, int y, 
                                                float baseScale) const
{
  blitz::TinyVector<RealType,3> const dHeadtipCam( // right-handed, y is down
    (x - imageExtent_[X]/2) / focalLength[X], 
    (y - imageExtent_[Y]/2) / focalLength[Y], 
    1);

  blitz::TinyVector<RealType,3> const dHeadtipWorld = 
    blitz::product(Rinv, dHeadtipCam); // ray from the camera focal point to the backprojected window top (in world coordinates)

  RealType const dz = pedestrianHeight - camPosWorld[Z]; 

  blitz::TinyVector<RealType,4> const wintopWorld(
    dHeadtipWorld[X] * dz / dHeadtipWorld[Z] + camPosWorld[X],
    dHeadtipWorld[Y] * dz / dHeadtipWorld[Z] + camPosWorld[Y],
    pedestrianHeight * (winsize_[Y]-borderPixels) / (winsize_[Y]-2*borderPixels),
    1);
  blitz::TinyVector<RealType,3> const wintopImg = 
    blitz::product(proj, wintopWorld);
  RealType const wintopRow = wintopImg[Y] / wintopImg[S];
  RealType const wintopCol = wintopImg[X] / wintopImg[S];

  blitz::TinyVector<RealType,4> const winbottomWorld(
    dHeadtipWorld[X] * dz / dHeadtipWorld[Z] + camPosWorld[X],
    dHeadtipWorld[Y] * dz / dHeadtipWorld[Z] + camPosWorld[Y],
    -pedestrianHeight * borderPixels / winsize_[Y],
    1);
  blitz::TinyVector<RealType,3> const winbottomImg = 
    blitz::product(proj, winbottomWorld);
  RealType const winbottomRow = winbottomImg[Y] / winbottomImg[S];

  // Assumes right vector is (nearly) parallel to the ground plane
  RealType const windowHeightPx = baseScale * (winbottomRow - wintopRow);
  RealType const scale          = windowHeightPx / winsize_[Y];
  RealType const windowWidthPx  = winsize_[X] * scale;

  return DetectedRegion(0.f, scale, 
                        (int)floorf(wintopCol - windowWidthPx/2.f), 
                        (int)floorf(wintopRow),
                        (int)ceilf(windowWidthPx), (int)ceilf(windowHeightPx));
}

void PerspectiveProjector::detectedRegionsFromHeadTips(
  std::vector<DetectedRegion> &dets,
  GrayImage const &headtips, GrayImage const &mask) const
{
  if ((headtips.extent()[0] != mask.extent()[0]) ||
      (headtips.extent()[1] != mask.extent()[1]))
  {
    ThrowException("headtips and mask must be exactly the same size");
  }

  dets.clear();
  for (int x=0; x<headtips.extent()[0]; x++) {
    for (int y=0; y<headtips.extent()[1]; y++) {
      if (headtips(x,y) && mask(x,y)) {
        dets.push_back(detectedRegionFromHeadTip(x,y, 1.0f));
      }
    }
  }
}

void PerspectiveProjector::detectedRegionsFromHeadTips(
  std::vector<DetectedRegion> &dets,
  GrayImage const &headtips) const
{
  dets.clear();
  for (int x=0; x<headtips.extent()[0]; x++) {
    for (int y=0; y<headtips.extent()[1]; y++) {
      if (headtips(x,y)) {
        dets.push_back(detectedRegionFromHeadTip(x,y, 1.0f));
      }
    }
  }
}
