#ifndef PDCUDA_ARRAYS_H
#define PDCUDA_ARRAYS_H

#include <cutil/cutil.h>
#include <cutil/cutil_inline.h>

namespace pdcuda {

  template <typename T>
  inline T *cudaMallocT(size_t nElm) {
    T *dData = NULL;
    cutilSafeCall(cudaMalloc((void**)(void*)&dData, nElm*sizeof(T)));
    return dData;
  }

  // Code adapted from boost::scoped_array to wrap CUDA device arrays
  // instead of host arrays.  Some extra convenience methods have been
  // added as well.
  template <typename T>
  class ScopedDeviceArray {
  private:
    T *ptr;
    ScopedDeviceArray(ScopedDeviceArray const &); // prevent usage
    ScopedDeviceArray &operator=(ScopedDeviceArray const &); // prevent usage
    typedef ScopedDeviceArray<T> This;

  public:
    typedef T element_type;
    typedef T *This::*unspecified_bool_type;
    
    explicit ScopedDeviceArray(size_t nElm):ptr(cudaMallocT<T>(nElm)){}//throws
    explicit ScopedDeviceArray(T *p=0)     :ptr(p) {}
    ~ScopedDeviceArray()                   { if (ptr!=NULL) cudaFree(ptr); }
    
    void reset(T *p=0)                     { This(p).swap(*this); }
    
    T &operator[](ptrdiff_t i)       const { return ptr[i]; }
    T *get() const                         { return ptr; }
    T *release()                           { T *t=ptr; ptr=NULL; return t; }
    
    operator unspecified_bool_type() const { return (ptr==0)?0:&This::ptr;}
    bool operator!() const                 { return ptr == 0; }

    void swap(ScopedDeviceArray &b)        { T *t=b.ptr; b.ptr=ptr; ptr=t; }
  };

};

#endif
