#ifndef PDCUDA_DYNDISPATCH_H
#define PDCUDA_DYNDISPATCH_H

#include <cutil/exception.h>
#include <sstream>

#ifdef _WIN32
#  pragma warning (push)
// For now, I trust Microsoft to have properly implemented ostringstream,
// despite the VS2008 warnings on Microsoft's own headers.
#  pragma warning (disable: 4355) 
#endif

namespace pdcuda {

  // Contains templates for doing dynamic dispatch on multidimensional
  // positive integer spaces.  This is most useful when you have a
  // CUDA kernel that has integer template parameters and you want to
  // be able to choose the arguments at runtime.
  //
  // For example, suppose you have a templated kernel called Foo that
  // takes two integer template arguments and a float* function
  // argument.  Instead of a large switch block
  //
  //   float *farg  = ...;
  //   dim3   grid  = ...;
  //   dim3   block = ...;
  //
  //   switch (x) {
  //   case 1:
  //     switch (y) {
  //     case 1: Foo<1,1><<<grid,block>>>(farg); break;
  //     case 2: Foo<1,2><<<grid,block>>>(farg); break;
  //     case 3: Foo<1,3><<<grid,block>>>(farg); break;
  //     ...
  //     case MaxY: Foo<1,MaxY><<<grid,block>>>(farg); break;
  //     default: throw ...;
  //     };
  //     break;
  //   ...
  //   case MaxX:
  //     switch (y) {
  //     case 1: Foo<MaxX,1><<<grid,block>>>(farg); break;
  //     case 2: Foo<MaxX,2><<<grid,block>>>(farg); break;
  //     case 3: Foo<MaxX,3><<<grid,block>>>(farg); break;
  //     ...
  //     case MaxY: Foo<MaxX,MaxY><<<grid,block>>>(farg); break;
  //     default: throw ...;
  //     };
  //     break;
  //   default: throw ...;
  //   }
  //
  // you could use IntDispatcher2D:
  //
  //   struct FooFunctor {
  //     FooFunctor(float *farg) : farg(farg) {};
  //
  //     template <int X, int Y> 
  //     void dispatch() { 
  //       dim3 grid  = ...;
  //       dim3 block = ...;
  //       Foo<X,Y><<<grid,block>>>(farg); 
  //     }
  //
  //   private:
  //     float farg;
  //   };
  //   ...
  //   float     *farg = ...;
  //   FooFunctor f(farg);
  //   IntDispatch2D<MaxX,Max,FooFunctor>::dispatch(x,y,f);

  // FUTURE WORK:
  //   Figure out the syntax for calling a templated operator().  If we
  //   can do this, then we should be able to directly call templated
  //   functions that have no function parameters instead of forcing
  //   people to create pseudo-functor objects.
  //
  //   Create dispatchers for other dimensions (1D, 3D, 4D, ...)

  template <int X, int Y, typename F>
  struct IntDispatcher2D {
    static void dispatch(int x, int y, F &f) {
      if ((x == X) && (y == Y)) {
        f.template dispatch<X,Y>();
        return; // needed by the CUDA 2.1 compiler
      } else if (x < X) {
        IntDispatcher2D<X-1, Y, F>::dispatch(x,y,f);
      } else if (y < Y) {
        IntDispatcher2D<X, Y-1, F>::dispatch(x,y,f);
      } else {
        std::ostringstream msg;
        msg <<"x="<<x<<" (it must be positive and <= "<<X<<") and "
            <<"y="<<y<<" (it must be positive and <= "<<Y<<")";
        RANGE_EXCEPTION(msg.str());
      }    
    }
  };

  template<int X, typename F>
  struct IntDispatcher2D<X, 0, F> {
    static void dispatch(int x, int y, F &f) { 
      std::ostringstream msg;
      msg <<"y="<<y<<" is invalid";
      RANGE_EXCEPTION(msg.str());
    }
  };

  template<int Y, typename F>
  struct IntDispatcher2D<0, Y, F> {
    static void dispatch(int x, int y, F &f) { 
      std::ostringstream msg;
      msg <<"x="<<x<<" is invalid";
      RANGE_EXCEPTION(msg.str());
    }
  };

  template<typename F>
  struct IntDispatcher2D<0,0,F> {
    static void dispatch(int x, int y, F &f) { 
      std::ostringstream msg;
      msg <<"x="<<x<<" or y="<<y<<" is invalid";
      RANGE_EXCEPTION(msg.str());
    }
  };

};

#ifdef _WIN32
#  pragma warning (pop)
#endif

#endif
