#ifndef PDCUDA_IMRESIZE_H
#define PDCUDA_IMRESIZE_H

#include <cuda_runtime.h>
#include <cufft.h>

namespace pdcuda
{

  // Implementations (mostly possible future ones):
  //   upscaling:
  //     true resampling
  //       cufft
  //     linear interpolation
  //       tex2d
  //       tex3d
  //
  //   downscaling:
  //     true resampling
  //       cufft
  //     lanczos3
  //       separable convolution
  //       tex2d
  //       tex3d
  //     nearest neighbor
  //       tex2d
  //       tex3d
  //     box filter with interpolated edges & corners
  //       tex2d
  //       tex3d
  //     super sampling
  //       tex2d
  //       tex3d
  //     

  // Abstract base class: it's designed for resizing the same input
  // image many times.
  class ImageResizer {
  public:
    ImageResizer(int w, int h) : w(w), h(h) {};
    virtual ~ImageResizer() {};
    virtual float *resize(int w, int h) const = 0; // row-major, packed RGB, device 
    virtual float *croppedResize(int x0, int y0, // top-left coords for cropping
                                 int wc, int hc, // crop size
                                 int w,  int h) const = 0; // new size, after cropping then rescaling
    
  protected:
    int    w; // input image width
    int    h; // input image height
  };

  // ++==============++
  // || Timing data: ||
  // ++==============++
  // The following timing tests were performed with a Core i7 965 and
  // GTX 280 card on a 1920x1080 source image. The demoimresize app was
  // used to produce the timing information.
  //
  // 2x upsampling
  // =============
  //   Implementation           constructor      resize
  //   --------------           -----------     --------
  //   ImageResizerLanczos2D         15.5ms       58.7ms
  //   ImageResizerLanczos3D         21.5ms      162.1ms
  //   ImageResizerBilinear2D        15.5ms       11.2ms
  //
  // 2x downsampling
  // ===============
  //   Implementation           constructor      resize
  //   --------------           -----------     --------
  //   ImageResizerLanczos2D         15.5ms       20.8ms
  //   ImageResizerLanczos3D         21.5ms       39.1ms


  // Uses a Lanczos filter for near-optimal resampling (in a Fourier
  // sense).  A 2D texture is used to accelerate memory access.
  class ImageResizerLanczos2D : public ImageResizer {
  public:
    ImageResizerLanczos2D(float const *data, int w, int h,
                          int windowRadius,
                          bool dataOnHost, bool dataColumnMajor);
    virtual ~ImageResizerLanczos2D();
    virtual float *resize(int w, int h) const;
    virtual float *croppedResize(int x0, int y0,
                                 int wc, int hc,
                                 int w,  int h) const;
        
  private:
    int windowRadius;
    cudaArray *data;
  };


  // Uses a Lanczos filter for near-optimal resampling (in a Fourier
  // sense).  A 3D texture is used to accelerate memory access.
  //
  // When using a Core i7 965 and GTX 280 card on a 1920x1080 source
  // image, the 2D version is typically much faster.
  //
  //     Test         constructor      resize
  //  Conditions       2D     3D      2D     3D
  //  -------------  ------ ------  ------ -------
  //  2x upsample    15.5ms 21.5ms  58.7ms 162.1ms
  //  2x downsample  15.5ms 21.5ms  20.8ms  39.1ms 
  class ImageResizerLanczos3D : public ImageResizer {
  public:
    ImageResizerLanczos3D(float const *data, int w, int h,
                          int windowRadius,
                          bool dataOnHost, bool dataColumnMajor);
    virtual ~ImageResizerLanczos3D();
    virtual float *resize(int w, int h) const;
    virtual float *croppedResize(int x0, int y0,
                                 int wc, int hc,
                                 int w,  int h) const;
        
  private:
    int windowRadius;
    cudaArray *data;
  };

  // Does bilinear interpolation using device hardware.  This resizer
  // is not recommended for downsampling unless the input image has
  // been pre-smoothed.
  class ImageResizerBilinear2D : public ImageResizer {
  public:
    ImageResizerBilinear2D(float const *data, int w, int h,
                           bool dataOnHost, bool dataColumnMajor);
    virtual ~ImageResizerBilinear2D();
    virtual float *resize(int w, int h) const;
    virtual float *croppedResize(int x0, int y0,
                                 int wc, int hc,
                                 int w,  int h) const;
    
  private:
    cudaArray *data;
  };

};

#endif
