#ifndef PDCUDA_RHOGDENSE_H
#define PDCUDA_RHOGDENSE_H

#include "../lib/cfg.h"
#ifdef HAVE_CUDA
#include "arrays.h"
#endif

#include "gradients.h"

namespace pdcuda 
{

#ifdef __CUDACC__
#  define __pdcuda_sharedfcn__ __device__ __host__
#else
#  define __pdcuda_sharedfcn__
#endif

#ifndef MIN
#  define MIN(a,b) (((a)<(b)) ? (a) : (b))
#endif
#ifndef MAX
#  define MAX(a,b) (((a)>(b)) ? (a) : (b))
#endif

  //---------------------------------------------------------------------------
  // Important constants
  //---------------------------------------------------------------------------
  
  static float const WeightScale    = 2.0f;
  static float const DefaultEpsilon = 1.0f;
  static float const DefaultMaxVal  = 0.2f;
  static float const Epsilon2       = 0.01f;

  // Internal details of an RHOG descriptor cell  
  static int   const CellSizeX   = 8;
  static int   const CellSizeY   = 8;
  static int   const CellSize[2] = {CellSizeX, CellSizeY};
  static int   const OrientBin   = 9;
  
  // How cells are grouped into RHOG descriptors
  static int   const NumCellX       = 2;
  static int   const NumCellY       = 2;
  static int   const NumCell[2]     = {NumCellX, NumCellY};
  static int   const BlockSizeX     = CellSizeX*NumCellX;
  static int   const BlockSizeY     = CellSizeY*NumCellY;
  static int   const BlockSize[2]   = {BlockSizeX, BlockSizeY};
  static int   const BlockDescLen   = NumCellX * NumCellY * OrientBin;
  
  // How RHOG descriptors are spaced for a detector window
  static int   const NumBlockX      = 7;
  static int   const NumBlockY      = 15;
  static int   const NumBlock[2]    = {NumBlockX, NumBlockY}; 
  static int   const BlockStrideX   = CellSizeX;
  static int   const BlockStrideY   = CellSizeY;
  static int   const BlockStride[2] = {BlockStrideX, BlockStrideY};
  static int   const WinSizeX       = (NumBlockX-1)*BlockStrideX + BlockSizeX;
  static int   const WinSizeY       = (NumBlockY-1)*BlockStrideY + BlockSizeY;
  static int   const WinSize[2]     = {WinSizeX, WinSizeY};
  static int   const WinDescLen     = NumBlockX * NumBlockY * BlockDescLen;

  static int   const BorderPixels     = 16;
  static float const PedestrianHeight = (5 + 9.f/12);

  // How detector windows are scanned across the image
  // FUTURE: remove me
  static int   const WinStrideX     = BlockStrideX;
  static int   const WinStrideY     = BlockStrideY;
  static int   const WinStride[2]   = {WinStrideX, WinStrideY};

  // Limits for the resolution at which we need to produce
  // descriptors.  Larger values allow more flexibility, but cause
  // code bloat.
  static int   const MaxFineStrideX = MAX(WinStrideX, BlockStrideX); 
  static int   const MaxFineStrideY = MAX(WinStrideY, BlockStrideY); 
  static int   const MaxFineStride[2] = {MaxFineStrideX, MaxFineStrideY};

#ifdef HAVE_CUDA
  //---------------------------------------------------------------------------
  // Window and block size math
  //---------------------------------------------------------------------------

  // Number of block descriptors that can be created from a given
  // image size.
  __pdcuda_sharedfcn__ 
  inline int horizBlocks(int w, int fineStrideX) { 
    return 1 + (w - BlockSizeX) / fineStrideX; 
  }

  __pdcuda_sharedfcn__ 
  inline int vertBlocks(int h, int fineStrideY)  { 
    return 1 + (h - BlockSizeY) / fineStrideY; 
  }

  // Number of windows built upon regular tilings of block descriptors
  // that can be created from a given grid of blocks.
  __pdcuda_sharedfcn__
  inline int horizWindows(int bw, int fineStrideX) {
    return 1 + (bw - NumBlockX) / (WinStrideX / fineStrideX);
  }

  __pdcuda_sharedfcn__
  inline int vertWindows(int bh, int fineStrideY) {
    return 1 + (bh - NumBlockY) / (WinStrideY / fineStrideY);
  }

  //---------------------------------------------------------------------------
  // Workhorse functions
  //---------------------------------------------------------------------------

  // dBD: (cy,cx,bin,dy,dx): last dim is most contiguous
  //
  // NOT threadsafe
  void computeRhogDense(float *dBD, float const *dMag, float const *dOri,
                        int w, int h, int fineStrideX, int fineStrideY,
                        float epsilon, float maxVal, float epsilon2);

  // Convenience wrapper that automatically allocates device memory
  // for the return value.
  //
  // NOT threadsafe
  inline float *computeRhogDense(float const *dMag, float const *dOri,
                                 int w, int h, 
                                 int fineStrideX, int fineStrideY,
                                 float epsilon, float maxVal, float epsilon2)
  {
    int const bw = horizBlocks(w, fineStrideX);
    int const bh = vertBlocks(h, fineStrideY);
    ScopedDeviceArray<float> dBD(NumCell[1]*NumCell[0]*OrientBin *bw*bh);

    pdcuda::computeRhogDense(dBD.get(), dMag, dOri, w, h, 
                             fineStrideX, fineStrideY,
                             epsilon, maxVal, epsilon2);

    return dBD.release();
  }

  // Last dim is most contiguous
  //   dScores: (wx,wy)
  //       dBD: (cx,cy,bin,dy,dx)
  //
  // Requires that fineStride be a common divisor of WinStride and
  // BlockStride.
  //
  // NOT threadsafe
  void classifyWindows(float *dScores, float const *dBD,
                       int dw, int dh, int fineStrideX, int fineStrideY, 
                       float const *hClassifier, float bias);

#endif
};

#endif
