#ifndef PDCUDA_RHOGDENSE_GOLD_H
#define PDCUDA_RHOGDENSE_GOLD_H

#include <memory>

#include <lear/cvision/imageslider.h>
#include <lear/cvision/dnormalizer.h>

#include "types_gold.h"
#include "rhogdense.h"
#include "arrays.h"

namespace pdcuda {

  inline void computeRhogDenseCPU(BlockDescriptors  *hBD, 
                                  Magnitude   const *hMag, 
                                  Orientation const *hOri);
  inline void computeClassificationScoresCPU(ScoreMap *hScores, 
                                             BlockDescriptors const *hBD,
                                             FeatureVector    const *hClassifier,
                                             float classifierBias, 
                                             bool columnMajor);

  inline void computeRhogDenseGPU_b(BlockDescriptors *hBD, float **dBD,
                                    TMagnitude const *tMag,
                                    TOrientation const *tOri);
  inline void computeClassificationScoresGPU(TScoreMap **tScores, 
                                             float const *dBD, int bw, int bh,
                                             float const *hClassifier, 
                                             float classifierBias);

  //---------------------------------------------------------------------------

  static int const GoldFineStrideX = BlockStrideX;
  static int const GoldFineStrideY = BlockStrideX;
  static int const GoldFineStride[] = {GoldFineStrideX, GoldFineStrideY};

  inline void computeRhogDenseCPU(BlockDescriptors  *hBD, 
                                  Magnitude   const *hMag, 
                                  Orientation const *hOri) 
  {
    using namespace blitz;
    using namespace pdcuda;
    typedef lear::ImageSlider<2> SliderType;

    // Gaussian spatial weights (adapted from RHOGDense::RHOGDense)
    blitz::Array<RealType,2> weight((IndexType)BlockSize);
    {
      TinyVector<RealType,2> var2 = (IndexType)BlockSize / (2 * WeightScale);
      var2 *= var2*2;
      static TinyVector<int,2>      const BSM1(((IndexType)BlockSize) - 1);
      static TinyVector<RealType,2> const center(((IndexType)BlockSize)/2.0f);
      for (DomainIter<2> i(0,BSM1); i; ++i) {
        // NOTE: we shift to the center of the pixel, unlike Dalal, so
        // that the Gaussian mask is x- and y-symmetric about the true
        // center of the block.
        TinyVector<RealType,2> t = *i+0.5f - center;
        t *= t / var2;
        weight(*i) = exp(-blitz::sum(t));      
      }
    }

    // Temporary magnitude workspace
    Magnitude thMag(BlockSize);

    // Histogram calculator
    blitz::TinyVector<RealType,3> 
      histExt(BlockSize[0], BlockSize[1], SemiCirc ? 180.f : 360.f);
    blitz::TinyVector<RealType,3> 
      histBandwidth(CellSize[0], CellSize[1], 
                    (SemiCirc ? 180.f : 360.f) / OrientBin);
    Histogram hist(0.0f, histExt, histBandwidth, 
                   blitz::TinyVector<bool,3>(false,false,true));

    // Simple normalizer
    lear::L2HysNormalizer<RealType> normalizer(DefaultEpsilon, DefaultMaxVal);

    // Iterate over all block locations
    SliderType slider(hMag->extent(), BlockSize, GoldFineStride);
    hBD->resize(NumCell[1], NumCell[0], OrientBin,
                slider.elem_extent()[1], slider.elem_extent()[0]);
    for (SliderType::iterator siter = slider.begin(); 
         siter != slider.end(); ++siter) {

      hist.clear();
    
      // Code adapted from RHOGDense::operator()
      IndexType const point = *siter;
      RectDomain<2> span(point, point + (IndexType)BlockSize-1);
      Orientation tori = (*hOri)(span);
      thMag = weight * (*hMag)(span); 

      for (int i=0; i<BlockSize[0]; ++i) {
        for (int j=0; j<BlockSize[1]; ++j) { 
          TinyVector<RealType,3> at( i+0.5f, j+0.5f, tori(i,j) );
          hist(at, thMag(i,j));
        }
      }

      normalizer(hist.data());

      // Extract histogram data
      IndexType const blockIdx = (*siter) / (IndexType)GoldFineStride;
      for (int cy=0; cy<NumCell[1]; cy++) {
        for (int cx=0; cx<NumCell[0]; cx++) {
          for (int o=0; o<OrientBin; o++) {
            RealType const binVal = hist.data()(cx, cy, o);
            (*hBD)(cy,cx,o, blockIdx[1],blockIdx[0]) = binVal;
          }
        }
      }
    }
  }

#define CPU_OPTIMIZED_ORDERING 0
#define GPU_ORDERING           1
#define CLASSIFICATION_ORDERING CPU_OPTIMIZED_ORDERING//GPU_ORDERING

  inline void computeClassificationScoresCPU(ScoreMap *hScores, 
                                             BlockDescriptors const *hBD,
                                             FeatureVector    const *hClassifier,
                                             float classifierBias, 
                                             bool columnMajor)
  {
    // cBD: (cy,cx,bin,y,x): last dim is most contiguous

    using namespace blitz;
    using namespace pdcuda;

    int const DH = hBD->extent()[3]; // # block descriptors tiled horizontally on img
    int const DW = hBD->extent()[4]; // # block descriptors tiled vertically on img

    assert((WinStrideX   % GoldFineStrideX) == 0);
    assert((BlockStrideX % GoldFineStrideX) == 0);
    assert((WinStrideY   % GoldFineStrideY) == 0);
    assert((BlockStrideY % GoldFineStrideY) == 0);
    assert(DW >= NumBlockX);
    assert(DH >= NumBlockY);

    int const WW = horizWindows(DW, GoldFineStrideX);
    int const WH = vertWindows(DH, GoldFineStrideY);

    if (columnMajor) {
      hScores->resize(WH,WW); 
    } else {
      hScores->resize(WW,WH); 
    }

    for (int wx=0; wx<WW; wx+=(WinStrideX/GoldFineStrideX)) {
      for (int wy=0; wy<WH; wy+=(WinStrideY/GoldFineStrideY)) {

        float score = -classifierBias;
#if CLASSIFICATION_ORDERING == CPU_OPTIMIZED_ORDERING
        float const *clsfr = &(*hClassifier)(0);
        for (int bx=0; bx<NumBlockX; bx++) {
          for (int by=0; by<NumBlockY; by++) {

            for (int cx=0; cx<NumCellX; cx++) {
              for (int cy=0; cy<NumCellY; cy++) {
                for (int b=0; b<OrientBin; b++) {
                  score += ((*hBD)(cy,cx,b, wy+by,wx+bx) * (*clsfr++));
                }
              }
            }
          }
        }
#elif CLASSIFICATION_ORDERING == GPU_ORDERING
        for (int cy=0; cy<NumCellY; cy++) {
          for (int cx=0; cx<NumCellX; cx++) {
            for (int b=0; b<OrientBin; b++) {
            
              for (int bx=0; bx<NumBlockX; bx++) {
                for (int by=0; by<NumBlockY; by++) {                
                  score += ((*hBD)(cy,cx,b, wy+by,wx+bx) *
                            (*hClassifier)[bx*NumBlockY*BlockDescLen + 
                                           by*BlockDescLen + 
                                           cx*NumCellY*OrientBin +
                                           cy*OrientBin +
                                           b]);
                }
              }
            }
          }
        }
#else
#     error "BAD CLASSIFICATION_ORDERING"
#endif
      
        if (columnMajor) {
          (*hScores)(wy,wx) = score;
        } else {
          (*hScores)(wx,wy) = score;
        }
      }
    }
  }


  inline void computeRhogDenseGPU_b(BlockDescriptors *hBD, float **dBD,
                                    TMagnitude const *tMag,
                                    TOrientation const *tOri) 
  {
    // (*gBD)(cy,cx,bin,y,x): last dim is most contiguous

    using namespace blitz;
    typedef lear::ImageSlider<2> SliderType;
    SliderType slider(tMag->extent(), BlockSize, GoldFineStride);

    size_t const nBytes = 
      NumCellY * NumCellX * OrientBin * slider.elem_extent()[1] * 
      slider.elem_extent()[0] * sizeof(float);
    float *dBDtmp = NULL;
    cutilSafeCall(cudaMalloc((void**)(void*)&dBDtmp, nBytes));

    pdcuda::computeRhogDense(dBDtmp, tMag->devicePtr(), tOri->devicePtr(), 
                             tMag->extent()[0], tMag->extent()[1], 
                             GoldFineStrideX, GoldFineStrideY,
                             DefaultEpsilon, DefaultMaxVal, Epsilon2);

    if (hBD != NULL) {
      hBD->resize(NumCellY, NumCellX, OrientBin,
                  slider.elem_extent()[1], slider.elem_extent()[0]);
      cutilSafeCall(cudaMemcpy(&(*hBD)(0,0,0,0,0),
                               dBDtmp, nBytes, cudaMemcpyDeviceToHost));
    }

    if (dBD == NULL) {
      cutilSafeCall(cudaFree(dBDtmp));
    } else {
      *dBD = dBDtmp;
    }
    cutilSafeCall(cudaThreadSynchronize());
  }

  inline void computeClassificationScoresGPU(TScoreMap **tScores, 
                                             float const *dBD, int bw, int bh,
                                             float const *hClassifier, 
                                             float classifierBias)
  {
    int const ww = horizWindows(bw, GoldFineStrideX);
    int const wh = vertWindows(bh, GoldFineStrideY);
    TScoreMap *tScoresTmp = new TScoreMap(ww, wh);
    classifyWindows(tScoresTmp->devicePtr(), dBD, bw, bh, 
                    GoldFineStrideX, GoldFineStrideY, 
                    hClassifier, classifierBias);
    // For now, we don't plan on doing the non-maxima suppression on the
    // GPU, so force the memcpy.
    tScoresTmp->fromDevice();
    if (tScores) {
      *tScores = tScoresTmp;
    } else {
      delete tScoresTmp;
    }
    cutilSafeCall(cudaThreadSynchronize());
  }
};

#endif
