#include <stdio.h>
#include <string>
#include <iostream>
#include <iomanip>

#include <lear/cmdline.h>
#include <lear/vectoropt.h>
#include <lear/tinyvecopt.h>
#include <lear/util/customoption.h>

#include <lear/image/imageio.h>
#include <lear/cvision/iprocessor.h>

#include <pdcuda/imresize.h>
#include <pdcuda/transposedarrays.h>
#include <pdcuda/testutils.h>

using namespace std;
using namespace pdcuda;

//////////////////////////////////////////////////////////////////////////////

typedef enum {
  Bilinear,
  Lanczos2D,
  Lanczos3D
} EResizer;

typedef lear::TinyVecOpt<int,2>   IndexOpt;
typedef blitz::TinyVector<int,2>  IndexType;

int main(int argc, char **argv)
{
  //===========================================================================
  // Command line option parsing

  using namespace lear;
  IndexOpt::seperator(",x");

  int       verbose;
  IndexOpt  topleft;
  IndexOpt  croppedExt;
  IndexOpt  resizedExt;
  string    algorithm;
  int       windowRadius;
  int       nIter;

  string    infile;
  string    outfile;

  Cmdline cmdline;
  cmdline.commandName("demoimresize");
  cmdline.version("0.0.2", "Author: Gerald Dalley"
                  "(mailto: dalleyg@ieee.org)");
  cmdline.brief("crops and resizes images");
  cmdline.description("Reads an input image, crops it, resizes it on the GPU, "
                      "then writes out the results to a new image file.");

  CustomOption algorithmOpt("resizing algorithm", Lanczos2D);
  algorithmOpt
    .add("bilinear", Bilinear)
    .add("lanczos2d", Lanczos2D)
    .add("lanczos3d", Lanczos3D)
    ;
  cmdline.addOption()
    ("verbose,v",option<int>(&verbose)
     ->defaultValue(2)->minValue(0)->maxValue(9),
     "verbose level")
    ("topleft,t",option<IndexOpt>(&(topleft))
     ->defaultValue(IndexOpt(0,0))->minValue(0),
     "top left cropping corner of the original image")
    ("cropext,c",option<IndexOpt>(&croppedExt)
     ->defaultValue(-1)->minValue(-1),
     "(x,y) size of the original image to crop out.  Use -1 to extend to "
     "the lower-right corner of the original input image")
    ("ext,e",option<IndexOpt>(&(resizedExt))
     ->defaultValue(IndexOpt(320,240))->minValue(0),
     "size of the output image, after cropping then resizing the original")
    ("algorithm,a",option<string>(&algorithm)
     ->defaultValue(algorithmOpt.defaultOption()),
     "resizing algorithm to use.  Options include: bilinear, lanczos2d, "
     "lanczos3d.")
    ("windowradius,r",option<int>(&windowRadius)
     ->defaultValue(2)->minValue(0),
     "radius of the smoothing kernel (for Lanczos-based resizing only).  "
     "Larger kernels are more accurate, but they are slower, and they "
     "tend to produce more pronounced ringing artifacts.")
    ("iter,i",option<int>(&nIter)
     ->defaultValue(5)->minValue(1),
     "number of times to repeat the cropping and resizing operation "
     "(for producing more robust timing information)")
    ;
  cmdline.addArgument()
    ("infile",  option<string>(&infile),  "original input image filename")
    ("outfile", option<string>(&outfile), "output image filename")
    ;
  int const status = cmdline.parse(argc, argv);
  if (status != cmdline.ok) return status;

  //===========================================================================
  // Actual worker code
  try {
    TimerSet timer;

    timer.tic("#0 warm up GPU");
    {
      TransposedDeviceArray2D<blitz::TinyVector<float,3> > r(1,1);
    }

    timer.tic("#1 imread");
    RGBImage in;
    lear::ImageIO::read(infile, in);
    for (int i=0; i<2; i++) {
      if (croppedExt[i] == -1) {
        croppedExt[i] = in.extent()[i] - topleft[i];
      }
    }

    timer.tic("#2 resultAllocator");
    TransposedDeviceArray2D<blitz::TinyVector<float,3> > 
      r(resizedExt[0], resizedExt[1]);

    for (int iter=0; iter<nIter; iter++) {   
      timer.tic("#3 createResizer");
      auto_ptr<ImageResizer> resizer;
      switch (algorithmOpt.check(algorithm)) {
      case Bilinear:
        resizer.reset(new 
                      ImageResizerBilinear2D((float const *)in.data(),
                                             in.extent()[0], in.extent()[1], 
                                             true, true));
        break;
      case Lanczos2D:
        resizer.reset(new 
                      ImageResizerLanczos2D((float const *)in.data(),
                                            in.extent()[0], in.extent()[1], 
                                            windowRadius, 
                                            true, true));
        break;
      case Lanczos3D:
        resizer.reset(new 
                      ImageResizerLanczos3D((float const *)in.data(),
                                            in.extent()[0], in.extent()[1], 
                                            windowRadius, 
                                            true, true));
        break;
      default:
        ThrowException("Invalid resize algorithm");
      } 
        
      timer.tic("#4 resize");
      float *dResized = resizer->croppedResize(topleft[0],    topleft[1],
                                               croppedExt[0], croppedExt[1],
                                               resizedExt[0], resizedExt[1]);
      timer.tic("#5 resized->host");
      cutilSafeCall(cudaMemcpy(r.devicePtr(), dResized, 
                               resizedExt[0]*resizedExt[1]*3*sizeof(float), 
                               cudaMemcpyDeviceToDevice));
      cutilSafeCall(cudaFree(dResized));
      r.fromDevice();
    }
    
    //=========================================================================
    // Results summary
    timer.tic("#6 imwrite");
    lear::ImageIO::write(outfile, *r.hostArray());  

    if (verbose >= 1) {
      if (verbose >= 2) {
        cout << "----------------------------------------------------------" << endl;
        cout << "demoimresize" << endl; 
        cout << "  parameters: " << endl;
        cout << "    infile:       " << infile << endl;
        cout << "    input size:   " << in.extent()[0]<<"x"<<in.extent()[1] << endl;
        cout << "    top-left:     " << topleft[0]<<"x"<<topleft[1] << endl;
        cout << "    cropped size: " << croppedExt[0]<<"x"<<croppedExt[1] << endl;
        cout << "    outfile:      " << outfile << endl;
        cout << "    output size:  " << resizedExt[0]<<"x"<<resizedExt[1] << endl;
        cout << "    algorithm:    " << algorithm << endl;
        cout << "    windowRadius: " << windowRadius << endl;
        cout << "    nIter:        " << nIter << endl;
        cout << "  timing profile:" << endl;
      }
      timer.printAverageTimes();
      if (verbose >= 2) {
        cout << endl;
      }
    }
      
  } catch(lear::Exception& e) {
    cerr << "Caught " << e.what() << endl;
    return 1;
  } catch(std::exception& e) {
    cerr << "Caught " << (e.what() ? e.what() : "UNKNOWN ERROR") << endl;
    return 1;
  } catch(...) {
    cerr << "Caught unknown exception" << endl;
    return 1;
  }
  return 0;
}

// reset; (make -j16 || make) && (cd test ; ./demoimresize input.png resized.png -e 3840,2160) && display test/resized.png 
// reset; (make -j16 || make) && (cd test ; ./demoimresize resamplingTestPattern.png resized.png -e 2048,1280) && display test/resized.png 
// reset; (make -j16 || make) && (cd test ; ./demoimresize resamplingTestPattern.png resized.png -t 32,32 -c 256,128 -e 1024,512) && display test/resized.png 
