function varargout = bboxAnnotationGui(varargin)
% BBOXANNOTATIONGUI interface for interactive annotation of pedestrian
%   locations.
% 
% bboxAnnotationGui(imgname)
%   Create new annotations for the image in the given filename.
%
% bboxAnnotationGui(imgname,bboxes,headcenters)
% bboxAnnotationGui(imgname,bboxes,headcenters,hipys)
%   Edit exiting annotations for the image in the given filename.  
%
%   BBOXES is a 2D matrix where each column specifies a bounding box of the
%     form [xmin;ymin;xmax;ymax].  Coordinates a 1-index (TODO: verify)
%     from the top-left corner of the image and have units of pixels.
%
%   HEADCENTERS is a 2D matrix where each column specifies the estimated
%     [x;y] centroid of a person's head.
%
%   HIPYS is a 1D matrix where each column specifies the estimated
%     y-coordinate of a person's hips.  If HIPYS is omitted, (ymax+ymin)/2
%     is assumed.
%
%   There must be the same number BBOXES, HEADCENTERS, and HIPYS entries.
%
% bboxAnnotationGui(basedir, annotationsHier)
%   Edit a whole set of existing annotatations, where ANNOTATIONSHIER is an
%   array of structures.  
%
%   Each struct must have an 'imagename' field in which contains the name
%   of the annotated image.  If the name is a relative path, it is treated
%   as relative to BASEDIR, e.g. if the name is 'pos/img00001.png' and
%   BASEDIR is '/foo/bar', then the image should be found at
%   '/foo/bar/pos/img00001.png'. 
%  
%   Each struct must also have a 2D array BBOXES conforming to the
%   description above.  These are the bounding boxes for the corresponding
%   image only.
%
%   Each struct must also have either a 'headcenters' or 'facecenters'
%   field as well.  If 'headcenters' are present, they must conform to the
%   HEADCENTERS description above.  For 'facecenters', we assume that the
%   top of the bounding box is the top of the head, that people are
%   upright, and that the face center us 75% of the way down from the top
%   of the head, relative to the total head height.  We convert
%   'facecenters' into 'headcenters'.
%
%   If the 'hipys' field is omitted, the bounding box midline is used.
%    
%   Any extra fields are discarded.
%
% bboxAnnotationGui(annotationsFile)
%   Edits a whole set of existing anntations, where we load an
%   ANNOTATIONSHIER structure from a .MAT file.  The .MAT file's directory
%   is treated as the BASEDIR.
%
% bboxAnnotationGui(basedir, imagesSubDir, annoFilename)
%   Creates a new set of annotations using all files found through a
%   recursive search of basedir.  Prepares to save the results to
%   fullfile(basedir,annoFilename). 
%
%   Note: if the annotations file already exists, it is loaded instead of
%   being recreated
%
% [annotationsHier, basedir, annoFilename] = bboxAnnotationGui(...)
%   Returns the final accepted annotations.  If available, the base
%   directory and filename of the annotations are given too.  If BASEDIR is
%   present, all 'imagename' fields of ANNOTATIONSHIER and the ANNOFILENAME
%   are relative paths w.r.t. the BASEDIR.
%
%
% UI CONTROLS:
%   * Clicking on an existing annotation selects it.  
%
%   * Dragging a solid-colored part of annotation moves it.
%
%   * SHIFT-click and/or CTRL-click adds/removes items from the selection
%     set.
%
%   * Clicking on the image creates a new annotation.
%
%   * Q /q      quit       exits the GUI, returning any committed edits. 
%
%   * <ESC>     unselect   deselects all annotations.
%
%   * <CTRL>-z  rollback   undoes all uncommitted edits to the current
%                          image.  
%
%   * <ENTER>   commit     commits current edits. 
%
%   * <DEL>     delete     deletes the currently-selected annotations. 
%
%   * -> / <-   navigate   arrow keys go to the next/previous image in the
%                          annotation set.  Auto-committing.
%
%   * <HOME>    first      go to the first image in the annotation set.
%                          Auto-committing.
%
%   * <END>     last       go to the first image in the annotation set.
%                          Auto-committing.
%
%   * <PGUP>    forward    Bring selected items forward
%
%   * <PGDN>    backward   Send selected items backward
%
%   * A / a     saveas     Prompts for a new annotation file for saving
%
%   * G / g     goto       select a particular image to go to.  
%
%   * H / h     hide       toggles between hiding and showing unselected
%                          annotations.
%
%   * S / s     save       saves to the current annotation file (reverts to
%                          saveas if there isn't a current one)
%
%
% EXAMPLES:
%   bboxAnnotationGui(inriaDir('train/inria/annotations_20080211.mat'))
%   bboxAnnotationGui(inriaDir('test/inria/annotations_20080211.mat'))
%   bboxAnnotationGui(grandCentralDir('test/1159407741750'), 'annotations_20080221.mat')
%   bboxAnnotationGui(grandCentralDir('test/notregct02'), 'annotations_20080221.mat')
%
% TODO: prevent bboxes from going beyond the image boundaries
% TODO: make border thickness, mouse deadzones, near threshold be w.r.t.
%        screen coords, not image coords.
% TODO: debug picking closest
% TODO: allow loading by pressing 'L'/'l'.
% TODO: when selecting parts, give priority to finest parts first (if dist
%        is equal)
% TODO: bboxAnnotationGui(basedir, imagessubdir)
% TODO: make these docs better
% TODO: allow varargs for setting the UI theme values

% Declare non-UI data here so that they can be used in any functions in
% this GUI file.

% Master table of annotations
mBaseDir         = '';
mAnnoFilename    = '';
mAnnotationsHier = struct('imagename',{}, 'headcenters',[], 'bboxes',[]); % master annotations structure

% Image being editing right now
mCurrImgNum      = -1;         % which entry in mAnnotationsHier are we using?
mBboxes          = zeros(4,0); % Annotation bounding boxes
mHeadCenters     = zeros(2,0); % Head center keypoints
mHipYs           = zeros(1,0); % Hip y-coordinates
mImage           = [];         % Current image

% UI Theme
mNearTaggedPt2          = 1.^2;  % mouse deadzone for selection (squared screen pixels)
mNearEdge2              = 3.^2;  % mouse click must be within this radius of an edge to consider it a selection (squared screen pixels)
mNearCorner2            = 5.^2;  % ...of a corner
mNearHeadCenter2        = 5.^2;  % ...of a head center indicator
mNearHipY               = 5.^2;  % ...of a hip y indicator
mBorderThickenessMinScr = 1;     % min thickness for the bounding box border, in screen pixels
mInactiveColor          = [0 1 0];
mSelectedColor          = [1 0 0];
mTaggedColor            = [0 0 1];
mDefaultBboxHeight      = 128;  % Dalal & Triggs canonical detector height
mDefaultHeadOffs        =   9;  % from exptExploreProportions
mDefaultBboxHalfWidth   =  20;  % from exptExploreProportions

% Interaction states (see annotationGuiStateDiagram.vsd)
WAITING     = 1;
SELECTING   = 2;
UNSELECTING = 4;
MOVING      = 8;
STATES = {WAITING,SELECTING,UNSELECTING,MOVING}; %#ok<NASGU>

% Different parts of an annotation that can be tagged by the mouse
HEADCENTER =   1;
VERTEX_NW  =   2;
VERTEX_NE  =   4;
VERTEX_SE  =   8;
VERTEX_SW  =  16;
EDGE_N     =  32;
EDGE_E     =  64;
EDGE_S     = 128;
EDGE_W     = 256;
HIPY       = 512; 
VERTICES  = VERTEX_NW   + VERTEX_NE + VERTEX_SE + VERTEX_SW;
EDGES     = EDGE_N      + EDGE_E    + EDGE_S    + EDGE_W;
PARTS     = HEADCENTER  + VERTICES  + EDGES;
SEL_PARTS = PARTS + HIPY;

% Composite state machine state
mUiState = struct;
mUiState.state        = WAITING; % current state (member of STATES)
mUiState.hideInactive = false;   % when enabled, only show selected or tagged annotations (logical)
mUiState.shift        = false;   % is the <SHIFT> key pressed? (logical)
mUiState.ctrl         = false;   % is the <CTRL> key pressed? (logical)
mUiState.button       = false;   % is a mouse button down? (logical)
mUiState.prevCoords   = [-1;-1]; % scrX,scrY mouse coordinates from previous callback (x,y screen coordinates).  
mUiState.sel          = [];      % list of selected annotations (indices into mBboxes or mHeadCenters)
mUiState.tagged       = [];      % annotation tagged as active (index into mBboxes or mHeadCenters)
mUiState.taggedPart   = [];      % part of the tagged annotation that's tagged (member of PARTS)
mUiState.taggedCoords = [];      % scrX,scrY mouse coordinates when the tagging was done (x,y screen coordinates).  

% Create all the UI objects in this GUI here so that they can
% be used in any functions in this GUI
hMainFigure = figure(...
  'Units','characters',...
  'MenuBar','none',...
  'Toolbar','figure',...
  'WindowStyle', 'modal',...
  'visible','off',...
  'KeyPressFcn',           @hMainFigureKeyPressFcn,...
  'KeyReleaseFcn',         @hMainFigureKeyReleaseFcn,...
  'WindowButtonDownFcn',   @hMainFigureWindowButtonDownFcn,...
  'WindowButtonUpFcn',     @hMainFigureWindowButtonUpFcn,...
  'WindowButtonMotionFcn', @hMainFigureWindowButtonMotionFcn);
hImgAxes = axes(...
  'Parent',hMainFigure,...
  'vis','off',...
  'Units','normalized',...
  'Position',[0 0 1 1]);
hImg = []; 

% Make changes needed for proper look and feel and running on different
% platforms
prepareLayout(hMainFigure);

% Process the command line input arguments supplied when the GUI is
% invoked
processUserInputs(varargin{:});

% Put user-input variables into canonical form, where applicable
fixInvertedBboxes();

% Initialize the iconEditor using the defaults or custom data given through
% property/value pairs
localUpdateGui();

% Make the GUI on screen
set(hMainFigure,'visible', 'on');
maximize(hMainFigure);

% Make the GUI blocking
uiwait(hMainFigure);

% Return the final annotations if requested
outputArgs    = {};         
outputArgs{1} = mAnnotationsHier;
outputArgs{2} = mBaseDir;
outputArgs{3} = mAnnoFilename;
if nargout>0
  [varargout{1:nargout}] = outputArgs{:};
end

try close(hMainFigure); catch end

%------------------------------------------------------------------
  function localUpdateGui
    % helper function that updates the gui when the data changes
    %set(hImgAxes,'CData', mImage,'Visible','on');
    if isempty(hImg)
      hImg = imshow(mImage);
    else
      delete(setdiff(get(hImgAxes,'Children'), hImg));
    end
   
    [toShow,nShown] = getVisible();
    for i=1:size(mHeadCenters,2)
      if (toShow(i))
        % TODO: make a third color when near an inactive?
        parts = [];
        issel = 0;
        if i == mUiState.tagged
          color = mSelectedColor;
          parts = mUiState.taggedPart;
          issel = 1;
        elseif ismember(i,mUiState.sel)
          color = mSelectedColor;
          issel = 1;
        else
          color = mInactiveColor;
        end
        drawAnnotation(floor(mHeadCenters(:,i)), floor(mBboxes(:,i)), ...
          floor(mHipYs(:,i)), ...
          issel, borderThicknessImg(nShown(i)), color, parts);
      end
    end
  end

%------------------------------------------------------------------
  function drawAnnotation(head,bbox,hipY,issel,thickness,color,taggedParts)
    hold on;
    % TODO: do I need to bias by +1 everywhere?

    % bounding box
    borderedRect([bbox(1) bbox(2) bbox(3)-bbox(1) bbox(4)-bbox(2)], ...
      thickness, ...
      [colorPicker(color,taggedParts,EDGE_N);...
      colorPicker(color,taggedParts,EDGE_S);...
      colorPicker(color,taggedParts,EDGE_E);...
      colorPicker(color,taggedParts,EDGE_W)]);
    
    % draw hip line for selected
    if issel
      line(bbox([1 3]), [hipY hipY], 'LineStyle',':', 'Color',colorPicker(color,taggedParts,HIPY));
    end
    
    % head
    plot(head(1),head(2),'+','MarkerSize',7,'LineWidth',1,'Color',colorPicker(color,taggedParts,HEADCENTER));
    headHalfHeight = (head(2) - bbox(2));
    headHalfWidth  = headHalfHeight * 0.75;
    ellipse(head, [headHalfWidth headHalfHeight], 17, ':','Color', colorPicker(color,taggedParts,HEADCENTER));
    line([-headHalfWidth headHalfWidth]+head(1), head(2)+headHalfHeight+[0 0], 'LineStyle',':', 'Color',colorPicker(color,taggedParts,HEADCENTER));

    % corners
    if bitand(VERTEX_NW,taggedParts), plot(bbox(1),bbox(2), '.', 'MarkerSize',5, 'Color',mTaggedColor); end
    if bitand(VERTEX_NE,taggedParts), plot(bbox(3),bbox(2), '.', 'MarkerSize',5, 'Color',mTaggedColor); end
    if bitand(VERTEX_SW,taggedParts), plot(bbox(1),bbox(4), '.', 'MarkerSize',5, 'Color',mTaggedColor); end
    if bitand(VERTEX_SE,taggedParts), plot(bbox(3),bbox(4), '.', 'MarkerSize',5, 'Color',mTaggedColor); end
    
    hold off;
    
    function c = colorPicker(untaggedColor,taggedParts,part)
      if bitand(part,taggedParts)
        c = mTaggedColor;
      else
        c = untaggedColor;
      end
    end
  end

%------------------------------------------------------------------
  function [toShow,nShown] = getVisible
    if mUiState.hideInactive
      toShow = ismember(1:size(mHeadCenters,2), ...
                        unique([mUiState.sel mUiState.tagged]));
    else
      toShow = true(1,size(mHeadCenters,2));
    end
    nShown = cumsum(toShow);
  end

%------------------------------------------------------------------
  function s = getSelectedAndTagged
    s = false(1,size(mHeadCenters,2));
    s(mUiState.sel)    = true;
    s(mUiState.tagged) = true;
  end

%------------------------------------------------------------------
  function t = borderThicknessImg(annoIdx)
    t = borderThicknessScr(annoIdx)*getZoom;
  end

%------------------------------------------------------------------
  function t = borderThicknessScr(annoIdx)
    t = (mBorderThickenessMinScr+annoIdx);
  end

%------------------------------------------------------------------
  function z = getZoom
    % screen pixels * getZoom = img pixels
    oldUnits = get(hMainFigure, 'Units');
    set(hMainFigure, 'Units','pixels'); % TODO: can I just keep pixel units in the setup?
    pos = get(hMainFigure, 'Position');
    figWidth = pos(3);         figHeight = pos(4);
    imgWidth = size(mImage,2); imgHeight = size(mImage,1);
    figAspect = figWidth / figHeight;
    imgAspect = imgWidth / imgHeight;
    if figAspect > imgAspect % img touches top and bottom of figure
      z = imgHeight / figHeight;
    else % img touches left and right of figure
      z = imgWidth / figWidth;
    end
    set(hMainFigure, 'Units',oldUnits);
  end

%------------------------------------------------------------------
  function hMainFigureKeyPressFcn(hObject, eventdata) %#ok<INUSL>
    % Callback called when key is pressed
    mUiState.ctrl  = any(ismember(eventdata.Modifier, 'control'));
    mUiState.shift = any(ismember(eventdata.Modifier, 'shift'));
  end

%------------------------------------------------------------------
  function hMainFigureKeyReleaseFcn(hObject, eventdata) %#ok<INUSL>
    % Callback called when key is released
    mUiState.ctrl  = any(ismember(eventdata.Modifier, 'control'));
    mUiState.shift = any(ismember(eventdata.Modifier, 'shift'));    
    
    if strcmp(eventdata.Key, 'q') % quit GUI
      switch questdlg('Save before quitting?')
        case 'Cancel'
          % do nothing
        case 'Yes'
          if doSave
            uiresume;
          end
        case 'No'
          % just quit
          uiresume;
        otherwise
          error('unexpected return value');
      end
      
    elseif strcmp(eventdata.Key, 'escape') % deselect
      resetStateMachine;
      
    elseif strcmp(eventdata.Key, 'z') && mUiState.ctrl % rollback
      selectImage(mCurrImgNum);
      resetStateMachine;
      
    elseif strcmp(eventdata.Key, 'return') % commit
      commitChanges;
      
    elseif strcmp(eventdata.Key, 'delete') % delete selected
      toKeep       = setdiff(1:size(mBboxes,2), mUiState.sel);
      mBboxes      = mBboxes(:,toKeep);
      mHeadCenters = mHeadCenters(:,toKeep);
      mHipYs       = mHipYs(:,toKeep);
      resetStateMachine;
      
    elseif strcmp(eventdata.Key, 'rightarrow') % commit & next
      commitChanges;
      selectImage(mod(mCurrImgNum, length(mAnnotationsHier))+1);
      resetStateMachine;

    elseif strcmp(eventdata.Key, 'leftarrow') % commit & prev
      commitChanges;
      selectImage(mod(mCurrImgNum-2, length(mAnnotationsHier))+1);
      resetStateMachine;

    elseif strcmp(eventdata.Key, 'home') % commit & first
      commitChanges;
      selectImage(1);
      resetStateMachine;

    elseif strcmp(eventdata.Key, 'end') % commit & end
      commitChanges;
      selectImage(length(mAnnotationsHier));
      resetStateMachine;

    elseif strcmp(eventdata.Character, 'g') % goto
      newIdx = listdlg('ListString',{mAnnotationsHier.imagename}, ...
        'PromptString','Select an image', 'SelectionMode','single', ...
        'InitialValue',mCurrImgNum, 'ListSize',[200 400]);
      if ~isempty(newIdx)
        selectImage(newIdx);
        resetStateMachine;
      end
      
    elseif strcmp(eventdata.Character, 's') % save
      doSave;
      
    elseif strcmp(eventdata.Character, 'a') % saveas
      doSaveAs;
      
    elseif strcmp(eventdata.Character, 'h') % toggle inactive hiding
      mUiState.hideInactive = ~mUiState.hideInactive;
      localUpdateGui;
        
    elseif strcmp(eventdata.Key, 'pageup') % forward
      if ~isempty(mUiState.sel)
        if mUiState.ctrl
          shiftSelected(size(mBboxes,2));
        else
          shiftSelected(1.5); % extra 0.5 avoids collisions
        end
        localUpdateGui;
      end
      
    elseif strcmp(eventdata.Key, 'pagedown') % backward
      if ~isempty(mUiState.sel)
        if mUiState.ctrl
          shiftSelected(-size(mBboxes,2));
        else
          shiftSelected(-1.5); % extra 0.5 avoids collisions
        end
        localUpdateGui;
      end
      
    else
      % do nothing
    end
  end

%------------------------------------------------------------------
  function shiftSelected(shiftAmt)
    allIdcs   = 1:size(mBboxes,2);
    selIdcs   = mUiState.sel;
    priorityList = allIdcs;
    priorityList(selIdcs) = selIdcs + shiftAmt;
    [yy,ii] = sort(priorityList);
    [dummy,mUiState.sel] = ismember(mUiState.sel, ii);
    if ~isempty(mUiState.tagged)
      mUiState.tagged = find(ii==mUiState.tagged);
    end
    mBboxes      = mBboxes(:,ii);
    mHeadCenters = mHeadCenters(:,ii);
    mHipYs       = mHipYs(:,ii);
  end
    
%------------------------------------------------------------------
  function done = doSave
    if ~isempty(mAnnoFilename)
      commitChanges;
      annotationsHier = mAnnotationsHier; %#ok<NASGU>
      save(fullfile(mBaseDir, mAnnoFilename), 'annotationsHier');
      clear annotationsHier;
      done = 1;
    else
      done = doSaveAs;
    end
  end

%------------------------------------------------------------------
  function done = doSaveAs
    [fname,pathname] = uiputfile(fullfile(mBaseDir, '*.mat'), ...
      'Save annotations as', mAnnoFilename);
    if fname==0
      done = 0;
    else
      commitChanges;
      annotationsHier = mAnnotationsHier; %#ok<NASGU>
      save(fullfile(pathname,fname), 'annotationsHier');
      clear annotationsHier;
      mBaseDir      = pathname;
      mAnnoFilename = fname;
      done = 1;
    end
  end

%------------------------------------------------------------------
  function resetStateMachine
    mUiState.state        = WAITING;
    % leave shift&ctrl alone...it's annoying to the user if we don't
    % leave button alone...to be consistent with shift&ctrl
    % leave prevCoords alone...it won't hurt anything
    mUiState.sel          = [];
    mUiState.tagged       = [];
    mUiState.taggedPart   = [];
    mUiState.taggedCoords = [];
    localUpdateGui;
  end

%------------------------------------------------------------------
  function commitChanges
    mAnnotationsHier(mCurrImgNum).headcenters = floor(mHeadCenters);
    mAnnotationsHier(mCurrImgNum).bboxes      = floor(mBboxes);
    mAnnotationsHier(mCurrImgNum).hipys       = floor(mHipYs);
  end

%------------------------------------------------------------------
  function hMainFigureWindowButtonDownFcn(hObject, eventdata) %#ok<INUSD>
    % Callback called when mouse is released
    mUiState.button = 1;
    if (ancestor(gco,'axes') == hImgAxes)
      eventDispatcher(0);
    end
  end

%------------------------------------------------------------------
  function hMainFigureWindowButtonMotionFcn(hObject, eventdata) %#ok<INUSD>
    % Callback called when mouse is moving
    eventDispatcher(1);
  end

%------------------------------------------------------------------
  function hMainFigureWindowButtonUpFcn(hObject, eventdata) %#ok<INUSD>
    % Callback called when mouse is released
    mUiState.button = 0;
    eventDispatcher(0);
  end

%------------------------------------------------------------------
  function eventDispatcher(mouseMoved)
    % Dispatcher for mouse-based events.  It was easier to collect the
    % state machine into a single dispatcher module.  See
    % annotationGuiStateDiagram.vsd (Visio diagram) for the design.
    [imgC,imgR,scrX,scrY] = getCurrentPoint;

    updateWindowTitle;
    
    switch mUiState.state
      case WAITING
        if mouseMoved || ~mUiState.button % WAITING(M***),WAITING(CU**): moving the mouse or spurious mouse up event
          if clearTag
            localUpdateGui;
          end
          
        elseif tagNearestObject(scrX,scrY) % WAITING(CD*S),WAITING(CD*U): near an object
          if mUiState.shift || mUiState.ctrl % WAITING(CDSS),WAITING(CDSU): near object to be toggled
            if ismember(mUiState.tagged, mUiState.sel) % WAITING(CDSS) --> UNSELECTING
              % Don't deselect yet; we might need to convert to MOVING
              % mode if the mouse moves before the button is released.
              mUiState.state = UNSELECTING;
              
            else % WAITING(CDSU) --> SELECTING
              mUiState.sel   = unique([mUiState.sel, mUiState.tagged]); 
              mUiState.state = SELECTING;
              
            end
          else % WAITING(CD-*): near object and no key modifiers
            if ismember(mUiState.tagged, mUiState.sel) % WAITING(CD-S) ---> SELECTING
              mUiState.state = SELECTING;
              
            else % WAITING(CD-U) ---setsel---> SELECTING
              mUiState.sel = mUiState.tagged;
              mUiState.state = SELECTING;
            end            
          end
          localUpdateGui;
          
        else % WAITING: far from everything
          if mUiState.shift || mUiState.ctrl % WAITING(CDSF) --> clear selection
            mUiState.sel = [];

          else % WAITING(CD-F) --> add new annotation
            % +0.5 is to move them to the center of the pixel for UI
            % purposes
            mHeadCenters(:,end+1) = [imgC;imgR+mDefaultHeadOffs] + 0.5;
            imgW = size(mImage,2);
            imgH = size(mImage,1);
            mBboxes(:,end+1)      = [...
              max(1,      imgC-mDefaultBboxHalfWidth); ...
              max(1,      imgR); ...
              min(imgW-1, imgC+mDefaultBboxHalfWidth); ...
              min(imgH-1, imgR+mDefaultBboxHeight)] + 0.5; % see note above about +0.5
            mHipYs(:,end+1) = floor(mean(mBboxes([2 4],end)));
            mUiState.sel          = [ size(mBboxes,2) ]; %#ok<NBRAK>
          end
          localUpdateGui;
        end
        
        
      case SELECTING
        assert(~isempty(mUiState.tagged));
        if mouseMoved % SELECTING(M***)
          if isNearTaggedCoords(scrX,scrY) % SELECTING(M**T)
            % do nothing
            
          else % SELECTING(M**t)
            moveSelected(scrX,scrY);
            mUiState.state = MOVING;
            localUpdateGui;
          end
          
        elseif mUiState.button % SELECTING(CD**)
          % do nothing
          
        else % SELECTING(CU**) ---untag---> WAITING
          clearTag;
          mUiState.state = WAITING;
          localUpdateGui;
          
        end
        
        
      case UNSELECTING
        assert(~isempty(mUiState.tagged));
        if mouseMoved % UNSELECTING(M***)
          if isNearTaggedCoords(scrX,scrY) % UNSELECTING(M**T)
            % do nothing
            
          else % SELECTING(M**t)
            moveSelected(scrX,scrY);
            mUiState.state = MOVING;
            localUpdateGui;
          end
          
        elseif mUiState.button % UNSELECTING(CD**)
          % do nothing
          
        else % UNSELECTING(CU**) ---untag,-sel---> WAITING
          mUiState.sel = setdiff(mUiState.sel, mUiState.tagged);
          clearTag;
          mUiState.state = WAITING;
          localUpdateGui;
          
        end
        
        
      case MOVING
        if mouseMoved || mUiState.button % MOVING(M***), MOVING(CD**)
          moveSelected(scrX,scrY);
          localUpdateGui;
                    
        else % MOVING(CU**)
          clearTag;
          fixInvertedBboxes;
          mUiState.state = WAITING;
          localUpdateGui;
          
        end
        
        
      otherwise, 
        error('Illegal state: %d', mUiState.state);
    end
    
    mUiState.prevCoords = [scrX;scrY];
  end

%------------------------------------------------------------------
  function cleared = clearTag
    cleared = ~isempty(mUiState.tagged);
    mUiState.tagged       = [];
    mUiState.taggedPart   = [];
    mUiState.taggedCoords = [];
  end
    
%------------------------------------------------------------------
  function near = isNearTaggedCoords(x,y)
    % Is the current mouse point near the place where the user clicked when
    % tagging the most recent object?  Reason for this function: if the
    % user hasn't moved too far from the click, suppress the motion.
    assert(numel(mUiState.tagged)==1);
    assert(numel(mUiState.taggedPart)==1);
    assert(numel(mUiState.taggedCoords)==2);
    near = (sum(([x;y]-mUiState.taggedCoords).^2) <= mNearTaggedPt2);
  end

%------------------------------------------------------------------
  function wasTagged = tagNearestObject(x,y)
    clearTag;
    wasTagged = 0;
    bestDist2 = inf;
    toShow = getVisible();
    isSel  = getSelectedAndTagged();
    for annoIdx=1:size(mBboxes,2)
      if (toShow(annoIdx))
        if (isSel(annoIdx))
          [near,part,dist2] = findNearestAnnoPart(x,y, annoIdx, SEL_PARTS);
        else
          [near,part,dist2] = findNearestAnnoPart(x,y, annoIdx, PARTS);
        end
        if near && dist2 <= bestDist2
          mUiState.tagged       = annoIdx;
          mUiState.taggedPart   = part;
          mUiState.taggedCoords = [x;y];
          wasTagged             = 1;
          bestDist2             = dist2;
        end
      end
    end
  end

%------------------------------------------------------------------
  function moveSelected(x,y)
    assert(numel(mUiState.tagged)==1);
    assert(numel(mUiState.taggedPart)==1);
    assert(numel(mUiState.taggedCoords)==2);

    z = getZoom;
    dx = (x-mUiState.prevCoords(1)) * z;
    dy = (y-mUiState.prevCoords(2)) * z;
    
    switch mUiState.taggedPart
      case HEADCENTER
        mHeadCenters(1,mUiState.sel) = mHeadCenters(1,mUiState.sel) + dx;
        mHeadCenters(2,mUiState.sel) = mHeadCenters(2,mUiState.sel) + dy;
        
      case VERTEX_NW
        mBboxes(1,mUiState.sel) = mBboxes(1,mUiState.sel) + dx;
        mBboxes(2,mUiState.sel) = mBboxes(2,mUiState.sel) + dy;
      case VERTEX_NE
        mBboxes(3,mUiState.sel) = mBboxes(3,mUiState.sel) + dx;
        mBboxes(2,mUiState.sel) = mBboxes(2,mUiState.sel) + dy;
      case VERTEX_SE
        mBboxes(3,mUiState.sel) = mBboxes(3,mUiState.sel) + dx;
        mBboxes(4,mUiState.sel) = mBboxes(4,mUiState.sel) + dy;
      case VERTEX_SW
        mBboxes(1,mUiState.sel) = mBboxes(1,mUiState.sel) + dx;
        mBboxes(4,mUiState.sel) = mBboxes(4,mUiState.sel) + dy;
        
      case EDGE_N
        mBboxes(2,mUiState.sel) = mBboxes(2,mUiState.sel) + dy;
        mHipYs(mUiState.sel)    = mHipYs(mUiState.sel) + dy/2;
      case EDGE_S
        mBboxes(4,mUiState.sel) = mBboxes(4,mUiState.sel) + dy;
        mHipYs(mUiState.sel)    = mHipYs(mUiState.sel) + dy/2;
      case EDGE_E
        mBboxes(3,mUiState.sel) = mBboxes(3,mUiState.sel) + dx;
      case EDGE_W
        mBboxes(1,mUiState.sel) = mBboxes(1,mUiState.sel) + dx;
        
      case HIPY
        mHipYs(mUiState.sel)    = mHipYs(mUiState.sel) + dy;

      otherwise,
        error('Illegal tagged part');
    end
    
  end    

%------------------------------------------------------------------
  function fixInvertedBboxes
    xinv = (mBboxes(1,:) > mBboxes(3,:));
    yinv = (mBboxes(2,:) > mBboxes(4,:));
    
    tmp = mBboxes(1,xinv);
    mBboxes(1,xinv) = mBboxes(3,xinv);
    mBboxes(3,xinv) = tmp;
    
    tmp = mBboxes(2,yinv);
    mBboxes(2,yinv) = mBboxes(4,yinv);
    mBboxes(4,yinv) = tmp;
  end
 
%------------------------------------------------------------------
  function [near,part,dist2] = findNearestAnnoPart(x,y, annoIdx, allowedParts)
    bbox = screenCenteredFromImg(mBboxes(:,annoIdx))+0.5;       % -> screen coords
    head = screenCenteredFromImg(mHeadCenters(:,annoIdx))+0.5;  % -> screen coords
    hipy = screenCenteredFromImg(mHipYs(:,annoIdx))+0.5;        % -> screen coords

    [toShow,nShown] = getVisible();
    t  = borderThicknessScr(nShown(annoIdx));
    th = t / 2; % half thickness

    dHead = sum(([x;y]-head).^2);
    dNW   = sum(([x;y]-bbox([1 2])).^2);
    dNE   = sum(([x;y]-bbox([3 2])).^2);
    dSE   = sum(([x;y]-bbox([3 4])).^2);
    dSW   = sum(([x;y]-bbox([1 4])).^2);
    dN    = dist2toHorizSeg(x,y, t, bbox(1),bbox(3), bbox(2)-th);
    dS    = dist2toHorizSeg(x,y, t, bbox(1),bbox(3), bbox(4)+th);
    dE    = dist2toVertSeg( x,y, t, bbox(3)+th, bbox(2),bbox(4));
    dW    = dist2toVertSeg( x,y, t, bbox(1)-th, bbox(2),bbox(4));
    dHipY = dist2toHorizSeg(x,y, 1, bbox(1),bbox(3), hipy);
    
    near = 1;
    if     bitand(allowedParts,HEADCENTER) && (dHead<=mNearHeadCenter2), dist2=dHead; part = HEADCENTER;
    elseif bitand(allowedParts,VERTEX_NW)  && (dNW  <=mNearCorner2    ), dist2=dNW;   part = VERTEX_NW;
    elseif bitand(allowedParts,VERTEX_NE)  && (dNE  <=mNearCorner2    ), dist2=dNE;   part = VERTEX_NE;
    elseif bitand(allowedParts,VERTEX_SE)  && (dSE  <=mNearCorner2    ), dist2=dSE;   part = VERTEX_SE;
    elseif bitand(allowedParts,VERTEX_SW)  && (dSW  <=mNearCorner2    ), dist2=dSW;   part = VERTEX_SW;
    elseif bitand(allowedParts,EDGE_N)     && (dN   <=mNearEdge2      ), dist2=dN;    part = EDGE_N;
    elseif bitand(allowedParts,EDGE_S)     && (dS   <=mNearEdge2      ), dist2=dS;    part = EDGE_S;
    elseif bitand(allowedParts,EDGE_E)     && (dE   <=mNearEdge2      ), dist2=dE;    part = EDGE_E;
    elseif bitand(allowedParts,EDGE_W)     && (dW   <=mNearEdge2      ), dist2=dW;    part = EDGE_W;
    elseif bitand(allowedParts,HIPY)       && (dHipY<=mNearHipY       ), dist2=dHipY; part = HIPY;
    else
      near  = 0;
      part  = [];
      dist2 = 0;
    end
  end

%------------------------------------------------------------------
  function dist2 = dist2toHorizSeg(x,y, t, segxmin,segxmax,segy)
    assert(segxmin<=segxmax);
    if x<segxmin % left of line segment -- L2 distance to endpoint
      dist2 = sum(([x;y]-[segxmin;segy]).^2);
    elseif x>segxmax % right of line segment -- L2 distance to endpoint
      dist2 = sum(([x;y]-[segxmax;segy]).^2);
    else
      if y>segy-t/2 && y<segy+t/2 % inside band
        dist2 = 0;
      else
        dist2 = (abs(y-segy) - t/2).^2; % outside band
      end
    end
  end

%------------------------------------------------------------------
  function dist2 = dist2toVertSeg(x,y, t, segx, segymin,segymax)
    dist2 = dist2toHorizSeg(y,x, t, segymin,segymax, segx);
  end

%------------------------------------------------------------------
  function processUserInputs(varargin)
    % populates:
    %  mBaseDir, mAnnoFilename, mAnnotationsHier 
    %  mCurrImgNum, mBboxes, mHeadCenters, mImage
    
    if nargin == 0
      error('Invalid arguments.  Type HELP %s', mfilename);
      
    elseif nargin==1 && ischar(varargin{1})    
      fname = varargin{1};
      try img = imread(fname); catch img=[]; end
      if ~isempty(img)                              % bboxAnnotationGui(imagename)
        [basedir,ff,ext] = fileparts(fname);
        mBaseDir                        = basedir;
        mAnnoFilename                   = '';
        mAnnotationsHier(1).imagename   = [ff ext];
        mAnnotationsHier(1).headcenters = zeros(2,0);
        mAnnotationsHier(1).bboxes      = zeros(4,0);
        cleanUpAndValidateImportedAnnos;
        selectImage(1);
        
      else                                          % bboxAnnotationGui(annotationsFile)
        importAnnoFile(fname);
      end

    elseif nargin==2 && isstruct(varargin{2})       % bboxAnnotationGui(basedir,annotationsHier)
      mBaseDir         = varargin{1};
      mAnnoFilename    = '';
      mAnnotationsHier = varargin{2};
      cleanUpAndValidateImportedAnnos;
      selectImage(1);
      
    elseif nargin==2 && ischar(varargin{2})         % bboxAnnotationGui(basedir, annoFilename)
      mBaseDir         = varargin{1};
      mAnnoFilename    = varargin{2};
      if exist(fullfile(mBaseDir, mAnnoFilename), 'file')
        importAnnoFile(fullfile(mBaseDir, mAnnoFilename));
      else
        mAnnotationsHier = buildNewHierFromDirSearch(mBaseDir);
        cleanUpAndValidateImportedAnnos;
        selectImage(1);
      end
      
    elseif nargin==3 || nargin==4                  % bboxAnnotationGui(imgname,bboxes,headcenters)
      [basedir,ff,ext] = fileparts(varargin{1});   % bboxAnnotationGui(imgname,bboxes,headcenters,hipys)
      mBaseDir                        = basedir;
      mAnnoFilename                   = '';
      mAnnotationsHier(1).imagename   = [ff ext];
      mAnnotationsHier(1).bboxes      = varargin{2};
      mAnnotationsHier(1).headcenters = varargin{3};
      if nargin>3
        mAnnotationsHier(1).hipys     = varargin{4};
      end
      cleanUpAndValidateImportedAnnos;
      selectImage(1);
      
    else
      error('Invalid arguments.  Type HELP %s', mfilename);
    end
  end

%------------------------------------------------------------------
  function importAnnoFile(fname)
    % Load the file (no changes if the load fails)
    data = load(fname, 'annotationsHier');
    
    mAnnotationsHier = data.annotationsHier;
    [mBaseDir,ff,ext] = fileparts(fname);
    mAnnoFilename = [ff ext];
    
    cleanUpAndValidateImportedAnnos;
    selectImage(1);
  end

%------------------------------------------------------------------
  function hier = buildNewHierFromDirSearch(basedir)
    hier = struct;
    [dummy,dummy,fnames] = dirr(basedir, '\.png\>', 'name');
    for i=1:length(fnames)
      hier(i).imagename   = fnames{i}(length(basedir)+2:end);
      hier(i).headcenters = zeros(2,0);
      hier(i).bboxes      = zeros(4,0);
    end
  end

%------------------------------------------------------------------
  function cleanUpAndValidateImportedAnnos
    fixInvertedBboxes();

    convertCenters = isfield(mAnnotationsHier, 'facecenters') && ...
                    ~isfield(mAnnotationsHier, 'headcenters');
    addHipYs = ~isfield(mAnnotationsHier, 'hipys');
                  
    for i=1:length(mAnnotationsHier)
      % convert facecenters to headcenters
      if convertCenters
        fy   = mAnnotationsHier(i).facecenters(2,:);
        ymin = mAnnotationsHier(i).bboxes(2,:);
        mAnnotationsHier(i).headcenters = mAnnotationsHier(i).facecenters;
        mAnnotationsHier(i).headcenters(2,:) = (fy-ymin)*2/3 + ymin;
      end
      
      % Add hipys if necessary
      if addHipYs
        mAnnotationsHier(i).hipys = floor(mean(mAnnotationsHier(i).bboxes([2 4],:)));
      end
      
      % datatype and dimensions checks
      anno = mAnnotationsHier(i);
      assert(ischar(anno.imagename));
      assert(isnumeric(anno.headcenters) && ndims(anno.headcenters)==2 && size(anno.headcenters,1)==2);
      assert(isnumeric(anno.bboxes)      && ndims(anno.bboxes)==2      && size(anno.bboxes,1)==4);
    end
    
    % remove unrecognized fields (since we'd probably handle them
    % incorrectly) 
    mAnnotationsHier = rmfield(mAnnotationsHier, ...
      setdiff(fieldnames(mAnnotationsHier), ...
              {'imagename','headcenters','bboxes','hipys'}));
  end

%------------------------------------------------------------------
  function selectImage(num)
    assert(num>0 && num<=length(mAnnotationsHier));
    mCurrImgNum  = num;
    % +0.5 is to move them to the center of the pixel for UI purposes
    mBboxes      = mAnnotationsHier(num).bboxes + 0.5;
    mHeadCenters = mAnnotationsHier(num).headcenters + 0.5;
    mHipYs       = mAnnotationsHier(num).hipys + 0.5;
    mImage       = imread(fullfile(mBaseDir, mAnnotationsHier(num).imagename));
    hImg = [];

    updateWindowTitle;
  end

%------------------------------------------------------------------
  function updateWindowTitle
    [imgC,imgR,scrX,scrY,rawPt] = getCurrentPoint;
    
    coordsstr = sprintf('image raw (%f,%f),  image (%d,%d),  screen (%d,%d)',...
      rawPt(1),rawPt(2), imgC,imgR, scrX,scrY);

    if mAnnotationsHier(mCurrImgNum).imagename
      windowTitle = sprintf('%s (%d of %d) - %s - %s',...
        mAnnotationsHier(mCurrImgNum).imagename,...
        mCurrImgNum, length(mAnnotationsHier), mfilename, coordsstr);
    else
      windowTitle = [mfilename coordsstr];
    end
    
    set(hMainFigure, 'Name',windowTitle, 'NumberTitle','off');
  end

%------------------------------------------------------------------
  function [imgC,imgR,scrX,scrY,rawPt] = getCurrentPoint
    rawPt = get(hImgAxes,'currentpoint');
    rawPt = rawPt(1,1:2);
    
    imgC = floor(rawPt(1)+0.5);          
    imgR = floor(rawPt(2)+0.5);
    
    scrX = (rawPt(1)-0.5)/getZoom-0.5;   
    scrY = (rawPt(2)-0.5)/getZoom-0.5;
  end

%------------------------------------------------------------------
  function [imgC,imgR] = imgFromScreen(scrX,scrY)
    % image coordinates given screen coordinates.  Use floor on the results
    % to get image indices.
    imgC = ((scrX-0.5)*getZoom + 1);
    imgR = ((scrY-0.5)*getZoom + 1);
  end

%------------------------------------------------------------------
  function [screenCoords] = screenCenteredFromImg(imgCoords)
    % screen coordinates of the center of the given image pixel.  
    screenCoords = zeros(size(imgCoords));
    screenCoords(1:2:end,:) = (floor(imgCoords(1:2:end,:)) - 1 + 0.5) / getZoom - 0.5;
    screenCoords(2:2:end,:) = (floor(imgCoords(2:2:end,:)) - 1 + 0.5) / getZoom - 0.5;
  end
end

%=========================================================================
function prepareLayout(topContainer)
% This is a utility function that takes care of issues related to
% look&feel and running across multiple platforms. You can reuse
% this function in other GUIs or modify it to fit your needs.
allObjects = findall(topContainer);
warning off  %#ok<WNOFF> %Temporary presentation fix
try
  titles=get(allObjects(isprop(allObjects,'TitleHandle')), 'TitleHandle');
  allObjects(ismember(allObjects,[titles{:}])) = [];
catch
end
warning on %#ok<WNON>

% Use the name of this GUI file as the title of the figure
defaultColor = get(0, 'defaultuicontrolbackgroundcolor');
if isa(handle(topContainer),'figure')
  set(topContainer,'Name', mfilename, 'NumberTitle','off');
  % Make figure color matches that of GUI objects
  set(topContainer, 'Color',defaultColor);
end

% Make GUI objects available to callbacks so that they cannot
% be changes accidentally by other MATLAB commands
% TODO: uncomment when done debugging
%%set(allObjects(isprop(allObjects,'HandleVisibility')), 'HandleVisibility', 'Callback');

% Make the GUI run properly across multiple platforms by using
% the proper units
if strcmpi(get(topContainer, 'Resize'),'on')
  set(allObjects(isprop(allObjects,'Units')),'Units','Normalized');
else
  set(allObjects(isprop(allObjects,'Units')),'Units','Characters');
end

% You may want to change the default color of editbox,
% popupmenu, and listbox to white on Windows
if ispc
  candidates = [findobj(allObjects, 'Style','Popupmenu'),...
    findobj(allObjects, 'Style','Edit'),...
    findobj(allObjects, 'Style','Listbox')];
  set(findobj(candidates,'BackgroundColor', defaultColor), 'BackgroundColor','white');
end

end
