function extractPositiveExamples(basedir, annoset, w,fullh, fracTop, innerpadding,outerpadding, maxOccl, exclInnerCropped)
%extractPositiveExamples(basedir, annoset, w,fullh, fracTop, innerpadding,outerpadding, maxOccl, exclInnerCropped)
%   Takes a hierarchical annotations structure and writes out Dalal-sized
%   positive samples (64x128 images where the person is 32x96 and
%   centered).
%
%   Output will go to 
%     <basedir>/pos_<annoset>_<cropset>*
%   where <cropset> is of the form AAxBB_HCC, where AA is inner-padded
%   detector window width, BB is the inner-padded window height, and CC is
%   the fullh value.  The inner padding can be recovered by computing
%   (BB-CC)/2.  The outer padding can be recovered by taking 
%   (imageHeight-CC)/2.
%
%EXAMPLE:
%   doExtractPositiveExamples
%

if nargin<9
  exclInnerCropped = 0;
end

writeMarked = 1;

h = fullh * fracTop;
padding = innerpadding+outerpadding;

cropset = sprintf('w%dx%dc8x8', w+innerpadding*2, h+innerpadding*2);

load(fullfile(basedir, ['annotations_' annoset '.mat']), 'annotationsHier');
anno = annotationsHier;
outreldir = ['pos_' annoset '_' cropset];
outdir    = fullfile(basedir, outreldir);
markeddir = fullfile(basedir, [outreldir, '_marked']);
outlst    = fullfile(basedir, [outreldir '_all.lst']);
unoccllst = arrayfun(...
  @(mo) fullfile(basedir, [outreldir sprintf('_occl%03d.lst', floor(mo*100))]), ...
  maxOccl, 'UniformOutput',0);

warning off; %#ok<WNOFF>
mkdir(outdir);
if writeMarked, mkdir(markeddir); end
F  = fopen(outlst, 'w');
Fu = cellfun(@(l) fopen(l, 'w'), unoccllst);

% foreach image
for iimg = 1:length(anno)
  % get data
  N = size(anno(iimg).bboxes,2);
  if ~N, continue; end
  img = imread(fullfile(basedir, anno(iimg).imagename));
  bboxesOrig = anno(iimg).bboxes; bboxesCrop = bboxesOrig;
  bboxesCrop(4,:) = bboxesOrig(2,:) + (bboxesOrig(4,:) - bboxesOrig(2,:)) * fracTop;
  heads  = anno(iimg).headcenters;
  origHeights = bboxesCrop(4,:) - bboxesCrop(2,:);
  [dummy, imgname] = fileparts(anno(iimg).imagename);

  % foreach annotation
  for ia=1:size(heads,2)
    % rescale
    scale   = h / origHeights(ia);
    small   = imresize(img, scale, 'lanczos3');

    % crop
    xmin    = ceil((heads(1,ia)-1) * scale - w/2) - padding + 1;
    ymin    = ceil((bboxesCrop(2,ia)-1) * scale) - padding + 1;
    cropped = imcrop(small, [xmin ymin w+padding*2-1, h+padding*2-1]);

    % did the inner padding boundary extend beyond the image borders?
    pixelReplInsideDetector = ...
      (xmin+outerpadding<1) || (xmin+w+padding*2-outerpadding>size(img,2)) ||...
      (ymin+outerpadding<1) || (ymin+h+padding*2-outerpadding>size(img,1));

    % handle cases where outer padding boundary extended beyond the image
    % borders
    if ~all(size(cropped) == [h+padding*2, w+padding*2, 3])
      % we hit an image boundary... pixel replication as was done by Dalal
      % and Triggs
      planes = cell(1,3);
      for c=1:3
        planes{c} = cropped(:,:,c);
        
        if xmin<=0
          planes{c} = [repmat(planes{c}(:,1), [1, -xmin+1]) planes{c}];
        end
        
        rpad = w+padding*2 - size(planes{c},2);
        if rpad>0
          planes{c} = [planes{c} repmat(planes{c}(:,end), [1 rpad])];
        end
        
        if ymin<=0
          planes{c} = [repmat(planes{c}(1,:), [-ymin+1, 1]) ; planes{c}];
        end
        
        bpad = h+padding*2 - size(planes{c},1);
        if bpad>0
          planes{c} = [planes{c} ; repmat(planes{c}(end,:), [bpad 1])];
        end
      end
      
      cropped = planes{1};
      cropped(:,:,2) = planes{2};
      cropped(:,:,3) = planes{3};
    end
    assert(all(size(cropped) == [h+padding*2, w+padding*2, 3]));
    
    % mark
    marked = cropped;
    for c=1:3
      marked(outerpadding, outerpadding+(0:w+innerpadding*2),     c) = (c==1)*255;
      marked(outerpadding+(0:h+innerpadding*2), outerpadding,     c) = (c==1)*255;
      marked(end-outerpadding, outerpadding+(0:w+innerpadding*2), c) = (c==1)*255;
      marked(outerpadding+(0:h+innerpadding*2), end-outerpadding, c) = (c==1)*255;

      marked(padding, padding+(0:w),     c) = (c==1)*255;
      marked(padding+(0:h), padding,     c) = (c==1)*255;
      marked(end-padding, padding+(0:w), c) = (c==1)*255;
      marked(padding+(0:h), end-padding, c) = (c==1)*255;
    end
    
    % figure out occlusion amount -- use biggest bboxes for occluders, but
    % just the region of interest for the occludee
    occlfrac = occlFrac(bboxesOrig, ia, [...
      heads(1,ia)-w/2/scale ; bboxesCrop(2,ia) ; ...
      heads(1,ia)+w/2/scale ; bboxesCrop(4,ia)]);

    % output images
    if writeMarked
      imwriteIfNeeded(marked,          markeddir, sprintf('%s_%02d.png', imgname, ia-1));
    end
    imwriteIfNeeded(cropped,           outdir,    sprintf('%s_%02d.png', imgname, ia-1));
    imwriteIfNeeded(imfliplr(cropped), outdir,    sprintf('%s_%02d_flip.png', imgname, ia-1));
    
    % output to index files
    if ~exclInnerCropped || ~pixelReplInsideDetector
      fprintf(F, '%s/%s_%02d.png\n', outreldir, imgname, ia-1);
      fprintf(F, '%s/%s_%02d_flip.png\n', outreldir, imgname, ia-1);
      for io=1:length(maxOccl)
        if (occlfrac <= maxOccl(io))
          fprintf(Fu(io), '%s/%s_%02d.png\n', outreldir, imgname, ia-1);
          fprintf(Fu(io), '%s/%s_%02d_flip.png\n', outreldir, imgname, ia-1);
        end
      end
    end
  end
end
arrayfun(@(f) fclose(f), Fu);
fclose(F);

%-------------------------------
function imwriteIfNeeded(im,d,fname)
fname = fullfile(d,fname);
%if ~exist(fname, 'file')
  imwrite(im,fname);
%end

%-------------------------------
function occlfrac = occlFrac(bboxes,ia,iabbox)

% 0=no annotation, 1=potential occluder, 2=annotation of interest
H = max(bboxes(4,:));
W = max(bboxes(3,:));
mask = zeros(H, W);
for ii=1:size(bboxes,2)
  if ii ~= ia
    mask(max(1,bboxes(2,ii)):min(H,bboxes(4,ii)), ...
      max(1,bboxes(1,ii)):min(H,bboxes(3,ii))) = 1;
  else
    mask(max(1,iabbox(2)):min(H,iabbox(4)), ...
      max(1,iabbox(1)):min(H,iabbox(3))) = 2;
  end
end

iaArea     = (iabbox(4)-iabbox(2)+1) * (iabbox(3)-iabbox(1)+1);
unocclArea = sum(mask(:)==2);
occlfrac = (iaArea - unocclArea) / iaArea;
