function multisavefig(h, fname, formats, dims, res)
%multisavefig(h, fname, formats, dims, res)
%  Saves a figure in a collection of formats.
%    H        figure handle (a scalar)
%    FNAME    base filename for saving the figure (without extension) 
%    FORMATS  list of formats to which the saving will be done.  This
%             should be a comma, semicolon, colon, and/or
%             whitespace-delimited string.  Default: 'png'.
%    DIMS     saved [width height], in inches.  Default: the figure's
%             'papersize' and 'paperposition' properties are unmodified.
%    RES      resolution in dots per inch, for raster formats.  Default:
%             same default as PRINT.
%
%EXAMPLE:
%  multisavefig(gcf, 'myfig', 'png;fig;eps;emf', [3 2], 300)
%    Saves the current figure to myfig.png, myfig.fig, myfig.eps, and
%    myfig.emf.  The .png, .eps, and .emf files will be 3 inches wide by 2
%    inches tall.  myfig.png will be a 900x600 image.

% parse args
error(nargchk(2,5,nargin,'struct'))

if nargin<3
  formats = 'png';
end
if ~iscell(formats)
  formats = split(sprintf(',:;\t\n\r '), formats);
end

if nargin>=4
  oldps = get(h,'papersize');
  oldpp = get(h,'paperposition');
  set(h,'papersize',dims);
  set(h,'paperposition',[0 0 dims]);
end

if nargin>=5
  resStr = sprintf('-r%d', res);
else
  resStr = '';
end

% do the work
for i=1:length(formats)
  usesaveas = 0;
  fmt = formats{i};
  switch fmt
    case {'pdf'}
      device = '-dpdf';
    case {'wmf', 'emf'}
      device = '-dmeta';
    case 'bmp'
      device = '-dbitmap';
    case 'ps'
      device = '-dpsc2';
    case 'eps'
      device = '-depsc2';
    case 'ai'
      device = '-dill';
    case {'jpeg','jpg'}
      device = '-djpeg';
    case {'tiff','png','hpgl'}
      device = ['-d' fmt];
    case {'fig','m'}
      usesaveas = 1;
  end
  
  if usesaveas
    saveas(h, [fname '.' fmt]);    
  else
    print(h, device, resStr, [fname '.' fmt]);
  end
end

% clean up if we modified dims
if nargin>=4
  set(h,'papersize',oldps);
  set(h,'paperposition',oldpp);
end
