%% outputCompositeCurves

%% Setup

clc;
basedir = '../../experiments';
close all;
addpath ..
if ~exist(todaysResultsDir('pr'), 'dir')
  mkdir(todaysResultsDir('pr'));
end
if ~exist(todaysResultsDir('roc'), 'dir')
  mkdir(todaysResultsDir('roc'));
end
if ~exist(thesisDir('ped/hog/dtResults'), 'dir')
  mkdir(thesisDir('ped/hog/dtResults'));
end

%% Load data

fprintf('loading data...\n');
[e1] = loadDetOptErrorData(basedir);
[e2] = loadDalalHistData(basedir);

%% Merge data
% This section creates a unified data structure for the experiment results.
% It uses a collection of parallel arrays that have the form:
%
%    experiments.train.(trainCfgField).(testIdx)
%    experiments.test.(testCfgField).(testIdx)
%    experiments.results.(resultField).(testIdx)

fprintf('merging detection and classification experiments...\n');
N1 = length(e1.train.hogId);
N2 = length(e2.train.hogId);
experiments = struct;
assert(all(cellfun(@strcmp, fieldnames(e1), fieldnames(e2))));
for ttc = fieldnames(e1)'
  tt = ttc{1};
  for fnc = union(fieldnames(e1.(tt)), fieldnames(e2.(tt)))'
    fn = fnc{1};
    if isfield(e1.(tt), fn)
      if isfield(e2.(tt), fn)
        experiments.(tt).(fn) = cat(2, e1.(tt).(fn), e2.(tt).(fn));
      else
        experiments.(tt).(fn) = cat(2, e1.(tt).(fn), cell(1,N2));
      end
    else
      experiments.(tt).(fn) = cat(2, cell(1,N1), e2.(tt).(fn));
    end
  end
end

%% Fill in blank values for test params 
% We're grabbing data from multiple different types of tests that are
% parameterized different ways.  The above code just creates blanks for
% missing parameter settings.  We'll convert them to 'N/A' strings now.
% This makes the HTML and Javascript select boxes much easier to handle.    

fprintf('data structure tweaks...\n');
N = length(experiments.train.hogId);
for fnc = fieldnames(experiments.test)'
  fn = fnc{1};
  for i=1:N
    if isempty(experiments.test.(fn){i})
      experiments.test.(fn){i} = 'NA';
    end
  end
end

%% Interpolate precision-recall data and compute area above curve values. 

fprintf('curve interpolation...\n');

N = length(experiments.train.hogId);
experiments.results.aacRoc                    = cell(1,N);
experiments.results.aacPr                     = cell(1,N);
experiments.results.achievablePrecisionDense  = cell(1,N);
experiments.results.achievableRecallDense     = cell(1,N);
experiments.results.achievablePrecisionSparse = cell(1,N);
experiments.results.achievableRecallSparse    = cell(1,N);
for i=1:N
  ex = sliceExperiment(experiments, i, 0);
  r = ex.results;

  % TODO: use rocInterp

  % Compute ROC envelope and AAC
  fpr = r.falsePos ./ (r.falsePos + r.trueNeg);
  tpr = r.truePos  ./ (r.truePos  + r.falseNeg);
  [efpr,etpr] = rocEnvelope(fpr, tpr);
  experiments.results.aacRoc{i} = ...
    sum((efpr(2:end)-efpr(1:end-1)) .* ...
    (1-(etpr(2:end)+etpr(1:end-1))/2));
  
  % High-resolution ROC interpolation -> PR & AAC
  hfpr = interp1(efpr, linspace(1,length(efpr),2000), 'linear');
  htpr = interp1(etpr, linspace(1,length(etpr),2000), 'linear');
  recall = htpr;
  htp = htpr * r.pos;
  hfp = hfpr * r.neg;
  precision = htp ./ (htp + hfp);
  precision(isnan(precision)) = 1;
  experiments.results.achievablePrecisionDense{i} = precision;
  experiments.results.achievableRecallDense{i}    = recall;
  experiments.results.aacPr{i} = ...
    sum((recall(2:end)-recall(1:end-1)) .* ...
    (1-(precision(2:end)+precision(1:end-1))/2));
  
  % Low-resolution ROC interpolation -> PR
  lfpr = interp1(efpr, linspace(1,length(efpr),10), 'linear');
  ltpr = interp1(etpr, linspace(1,length(etpr),10), 'linear');
  recall = ltpr;
  ltp = ltpr * r.pos;
  lfp = lfpr * r.neg;
  precision = ltp ./ (ltp + lfp);
  precision(isnan(precision)) = 1;
  experiments.results.achievablePrecisionSparse{i} = precision;
  experiments.results.achievableRecallSparse{i}    = recall;
end

%% Create ROC and Precision-Recall Plots

fprintf('plots for the DHTML explorer application...\n');

fontsize = 10;

% Come up with a unified ranking.  This is just used so that we can have
% consistent plot styles for a given experiment across the different plot
% types. 
aacKey = 'aacPr';
[aac,i] = sort(cellfun(@(x)x, experiments.results.(aacKey)));
experiments.results.aacPrRank = {};
for ii=1:length(i)
  experiments.results.aacPrRank{ii} = i(ii);
end

% Create all the plots
plotCfgs = struct(...
  'axislimit', {[0 1 0 1],          [1e-3 5e-1 1e-3 5e-1]},...
  'aacKey',    {'aacPr',            'aacRoc'},...
  'curveName', {'Precision-Recall', 'ROC'},...
  'dir',       {'pr',               'roc'},...
  'plotfcn',   {@prPlot,            @detPlot},...
  'legIdx',    {3,                  1},...
  'legpos',    {'SouthWest',        'NorthEast'},...
  'aspect',    {0.5                 0.65});
for pci=1:length(plotCfgs)
  axislimit = plotCfgs(pci).axislimit;
  aacKey    = plotCfgs(pci).aacKey;
  curveName = plotCfgs(pci).curveName;
  dir       = plotCfgs(pci).dir;
  plotfcn   = plotCfgs(pci).plotfcn;
  legIdx    = plotCfgs(pci).legIdx;
  
  [aac,i] = sort(cellfun(@(x)x, experiments.results.(aacKey)));
  experiments = sliceExperiment(experiments, i);
  
  % Create the DHTML visualization applications
  if strcmp(dir,'pr')
    htmlRender(experiments, ...
      'pr', 'Precision-Recall Curves',...
      'The following results are the precision-recall curves for our pedestrian detection experiments, using a modified form of the PASCAL Visual Object Challenge protocol.');
  elseif strcmp(dir,'roc')
    htmlRender(experiments, ...
      'roc', 'ROC Curves',...
      'The following results are the ROC curves for our pedestrian detection experiments, using Dalal''s protocol.  This protocol tends to result in identical curves for various experiments.  Equivalent results are grouped for readability.');
  else
    error('unrecognized dir: %s',dir);
  end
  
  % Plot with all PR curves
  clf; set(gcf, 'visible', 'off');
  plotfcn(experiments.results, 'All', experiments.results.legend, axislimit, fontsize); %#ok<NOEFF>
  set(legend, 'fontsize',7);
  multisavefig(gcf, todaysResultsDir([dir '/all']), {'png','fig'}, [7.5 7.5]*2, 100);

  % AAC bar chart
  clf; set(gcf, 'visible', 'off');
  bar(1:length(aac),aac,1);
  axis tight;
  set(gca,'xtick',1:length(aac));
  xlabel('Rank');
  ylabel({'Area Above',curveName, 'Curve', '(lower is better)'});
  set(gca, 'YAxisLocation','right', 'FontSize',7);
  grid on;
  multisavefig(gcf, todaysResultsDir([dir '/aacByRank']), {'png','fig'}, [10 1.5], 100);

  % Write transparent PRs to use with the DHTML application
  fPlot   = createRenderContext(1000,500); set(gcf,'Color',[1 1 1]); set(gca, 'Position', [0.1300 0.1100 0.7750 0.8150]);
  fLegend = createRenderContext(108,10); set(gcf,'Color',[1 1 1]);
  if ~exist(todaysResultsDir([dir '/figs']), 'dir')
    mkdir(todaysResultsDir([dir '/figs']));
  end
  for rank=1:length(aac)
    figure(fPlot);
    ex = sliceExperiment(experiments, rank);
    plotfcn(ex.results, '', {}, axislimit, fontsize, experiments.results.aacPrRank{rank}-1); %#ok<NOEFF>
    clear ex;
    title(''); axis off; set(fPlot, 'visible','off');
    imwrite(renderFig(fPlot), todaysResultsDir(sprintf('%s/figs/rank%03d.png',dir,rank)), 'Transparency',[1 1 1]);

    hlines = arrayfilt(@(c) strcmp('line', get(c,'type')), get(gca,'Children'));
    hline = hlines(legIdx);
    figure(fLegend);
    legendArgs = fullcollapse(cellfun(@(p) {p,get(hline,p)}, ...
      {'LineStyle','LineWidth','Marker','MarkerSize','Color'}, ...
      'UniformOutput',0));
    plot([-10 0 10], [0 0 0], legendArgs{:}); axis([-1 1 -1 1]); axis off; set(fLegend, 'visible','off');
    imwrite(renderFig(fLegend), todaysResultsDir(sprintf('%s/figs/legend%03d.png',dir,rank)), 'Transparency',[1 1 1]);
  end

  figure(fPlot);
  delete(hlines);%get(gca,'Children'));
  title([curveName ' Curves']); axis on; set(fPlot, 'visible','off');
  imwrite(renderFig(fPlot), todaysResultsDir([dir '/figs/bg.png']));

  close(fPlot);
  close(fLegend);
end

%% Custom plots for the thesis

clc;
fprintf('custom plots for my thesis...\n');

extensions = {'pdf','eps','png'};%{'png'};%
res = 100;

for pci=1:length(plotCfgs)
  axislimit = plotCfgs(pci).axislimit;
  aacKey    = plotCfgs(pci).aacKey;
  curveName = plotCfgs(pci).curveName;
  dir       = plotCfgs(pci).dir;
  plotfcn   = plotCfgs(pci).plotfcn;
  legIdx    = plotCfgs(pci).legIdx;
  legpos    = plotCfgs(pci).legpos;
  aspect    = plotCfgs(pci).aspect;
  
  % Here we re-organize the experiments data structure to make it easier to
  % manipulate (we probably should have done this in the first place).
  [aac,i] = sort(cellfun(@(x)x, experiments.results.(aacKey)));
  experiments = sliceExperiment(experiments, i);

  % "Declare" the data structure.  There's probably some clever way to
  % efficiently fill it at the same time as creating the fields, but this
  % works well enough for now.
  expts = struct;
  for fn = fieldnames(experiments)'
    expts.(fn{1}) = struct;
    for f = fieldnames(experiments.(fn{1}))'
      expts.(fn{1}).(f{1}) = [];
    end
  end

  % Fill it with data
  expts = repmat(expts, [1 length(experiments.(fn{1}).(f{1}))]);
  for fn = fieldnames(experiments)'
    for f = fieldnames(experiments.(fn{1}))'
      for i=1:length(experiments.(fn{1}).(f{1}))
        expts(i).(fn{1}).(f{1}) = experiments.(fn{1}).(f{1}){i};
      end
    end
  end
  
  % CPU Classification results (training vs. test set effects)
  asCfgs = {
    'orig',    'INRIA-orig',   0;...
    '20080211','INRIA-new',    1;...
    '20090319','GrandCentral', 0};
  for asi=1:size(asCfgs,1)
    annoset  = asCfgs{asi,1};
    testname = asCfgs{asi,2};
    forceSW  = asCfgs{asi,3};

    ii = (...
      cellfun(@(impl)strcmp(impl,'cpu'), experiments.train.impl) & ...
      cellfun(@(as)strcmp(as,annoset), experiments.test.annoset));
    exs = sliceExperiment(experiments, ii);
    leg = thesisLegend(expts(ii),...
      {'train.imageset','train.annoset','train.pass'});
    plotfcn(exs.results, '', leg, axislimit, fontsize, [exs.results.aacPrRank{:}]-1); %#ok<NOEFF>
    title('');
    if forceSW
      set(legend,'Location',legpos);
    else
      set(legend,'Location','SouthWest');
    end
    basename = thesisDir('ped/hog/dtResults', join('-', 'classification', dir, testname, 'cpu'));
    multisavefig(gcf, basename, extensions, [6 6*aspect], res);
  end

  % CPU vs. GPU classification
  ii = (...
    cellfun(@(x)strcmp(x,'20080211'), experiments.train.annoset) & ...
    cellfun(@(x)strcmp(x,'20090319'), experiments.test.annoset));
  exs = sliceExperiment(experiments, ii);
  leg = thesisLegend(expts(ii),...
    {'train.imageset','train.annoset','train.pass', 'train.impl'});
  plotfcn(exs.results, '', leg, axislimit, fontsize, [exs.results.aacPrRank{:}]-1); %#ok<NOEFF>
  title('');
  if forceSW
    set(legend,'Location',legpos);
  else
    set(legend,'Location','SouthWest');
  end
  basename = thesisDir('ped/hog/dtResults', join('-', 'classification', dir, 'GrandCentral', 'impl'));
  multisavefig(gcf, basename, extensions, [6 6*aspect], res);

  % CPU detection experiments
  if plotfcn==@prPlot
    etCfgs = {
      'train';...
      'test'};
    for eti = 1:size(etCfgs,1)
      errorType = etCfgs{eti,1};
      ii = (...
        cellfun(@(impl)strcmp(impl,'cpu'), experiments.train.impl) & ...
        cellfun(@(as)strcmp(as,errorType), experiments.test.errorType));
      exs = sliceExperiment(experiments, ii);
      leg = thesisLegend(expts(ii),...
        {'train.imageset','train.annoset','train.pass'});
      plotfcn(exs.results, '', leg, axislimit, fontsize, [exs.results.aacPrRank{:}]-1); %#ok<NOEFF>
      title('');
      set(legend,'Location',legpos);
      basename = thesisDir('ped/hog/dtResults', join('-', 'detection', dir, errorType, 'GrandCentral', 'cpu'));
      multisavefig(gcf, basename, extensions, [6 6*aspect], res);
    end
  end
end
fprintf('done.\n');


