function [etpr,efpr,p,r,sp,sr,aacRoc,aacPr] = rocInterp(truePos, falsePos, falseNeg, trueNeg)
%[etpr,efpr,p,r,sp,sr,aacRoc,aacPr] = rocInterp(truePos, falsePos, falseNeg, trueNeg)
%  ROC and Precision-Recall Interpolation + Achievable Envelope 
%
%  All inputs must be vectors of the same length.
%
%  OUTPUTS:
%    etpr,efpr  true positive rate, false positive rate envelope curve
%    p,r        densely interpolated achievable precision and recall
%    sp,sr      sparsely subsampled p,r points
%    aacRoc     area above the ROC curve
%    aacPr      area above the precision-recall curve
%

pos = truePos + falseNeg;
neg = trueNeg + falsePos;
assert(all(pos(1)==pos), 'The interpolation assumes the number of positives is constant across all experiments.');
assert(all(neg(1)==neg), 'The interpolation assumes the number of negatives is constant across all experiments.');
pos = pos(1);
neg = neg(2);

% Compute ROC envelope and AAC
fpr = falsePos ./ (falsePos + trueNeg);
tpr = truePos  ./ (truePos  + falseNeg);
[efpr,etpr] = rocEnvelope(fpr(:), tpr(:));
aacRoc = ...
  sum((efpr(2:end)-efpr(1:end-1)) .* ...
  (1-(etpr(2:end)+etpr(1:end-1))/2));

% High-resolution ROC interpolation -> PR & AAC
hfpr = interp1(efpr, linspace(1,length(efpr),2000), 'linear');
htpr = interp1(etpr, linspace(1,length(etpr),2000), 'linear');
recall = htpr;
htp = htpr * pos;
hfp = hfpr * neg;
precision = htp ./ (htp + hfp);
precision(isnan(precision)) = 1;
p = precision;
r = recall;
aacPr = ...
  sum((recall(2:end)-recall(1:end-1)) .* ...
  (1-(precision(2:end)+precision(1:end-1))/2));

% Low-resolution ROC interpolation -> PR
lfpr = interp1(efpr, linspace(1,length(efpr),10), 'linear');
ltpr = interp1(etpr, linspace(1,length(etpr),10), 'linear');
recall = ltpr;
ltp = ltpr * pos;
lfp = lfpr * neg;
precision = ltp ./ (ltp + lfp);
precision(isnan(precision)) = 1;
sp = precision;
sr = recall;
